/* -*- Mode: C; c-file-style: "stroustrup" -*- */
/* bifuncs_prog.c -- implementation of the programing constructs of
   NumExp */
#include "nxp_kernel.h"
#include "functions.h"
#include "config.h"
#include "i18n.h"

static NxpElement* bifunc_map_elem(NxpElement    *function,
				   NxpIdentifier *id,
				   NxpElement    *element, gboolean eval, ETC)
{
    GError *err = NULL;
    NxpElement *result;
    NxpElement *evaluated_element;
    NxpDefinition *definition;

    /* 0. eval element */
    if (eval)
	evaluated_element = nxp_element_eval(element, context, &err);
    else
	evaluated_element = nxp_element_simplify(element, context, &err);

    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }

    /* 1. Define the id var with 'evaluated_element' value */
    nxp_user_variable_define(id, evaluated_element, context, &err);
    if (err) {
	g_propagate_error(error, err);
	NXP_ELEMENT_UNREF(evaluated_element);
	return NULL;
    }
    definition = nxp_name_space_lookup_definition(context->cwd,
						  nxp_identifier_to_string(id));

    /* 2. Evaluate 'function'*/
    if (eval)
	result = nxp_element_eval(function, context, &err);
    else
	result = nxp_element_simplify(function, context, &err);
    if (err) {
	g_propagate_error(error, err);
	NXP_ELEMENT_UNREF(evaluated_element);
	return NULL;
    }

    NXP_ELEMENT_UNREF(evaluated_element);

    /* 3. Undefine the id var */
    nxp_name_space_remove_definition(context->cwd,definition);
    
    return result;
}

static NxpElement* bifunc_eval_map_vector_(NxpElement       *function,
					   NxpIdentifier    *id,
					   NxpElementVector *vector,
					   gboolean	    eval, ETC)
{
    NxpElement *mapped;
    guint vsize = 0, i;

    vsize = NXP_ELEMENT_VECTOR_SIZE(vector);
    mapped = nxp_element_vector_new(vsize, NULL);
    for (i=0; i<vsize; i++) {
	NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(mapped), i) =
	    bifunc_map_elem(function, id, NXP_ELEMENT_VECTOR_ELEM(vector, i),
			    eval, named_args, context, error);
    }
    return mapped;
}

NxpElement* bifunc_eval_map_vector(NxpElement *function,
				   NxpElementVariable *variable,
				   NxpElementVector *vector, ETC) 
{
    return bifunc_eval_map_vector_(function, 
				   variable->ident, vector, TRUE,
				   named_args,
				   context, 
				   error);
}

NxpElement* bifunc_eval_map_single_vector(NxpElement *function,
					  NxpElementVector *vector, ETC)
{
    NxpIdentifier* id;
    NxpElement* result;
    id = nxp_identifier_new_simple("_");
    result = bifunc_eval_map_vector_(function, id, 
				     vector, TRUE, named_args, 
				     context, error);
    nxp_identifier_unref(id);
    return result;
}

NxpElement* bifunc_simp_map_vector(NxpElement *function,
				   NxpElementVariable *variable,
				   NxpElementVector *vector, ETC) 
{
    return bifunc_eval_map_vector_(function, 
				   variable->ident, vector, FALSE,
				   named_args,
				   context, 
				   error);
}

NxpElement* bifunc_simp_map_single_vector(NxpElement *function,
					  NxpElementVector *vector, ETC)
{
    NxpIdentifier* id;
    NxpElement* result;
    id = nxp_identifier_new_simple("_");
    result = bifunc_eval_map_vector_(function, id, 
				     vector, FALSE, named_args,
				     context, error);
    nxp_identifier_unref(id);
    return result;
}

static NxpElement* bifunc_eval_map_matrix_(NxpElement       *function,
					   NxpIdentifier    *id,
					   NxpElementMatrix *matrix,
					   gboolean	    eval, ETC)
{
    NxpElement *mapped;
    guint width, height, i, j;

    width = NXP_ELEMENT_MATRIX_WIDTH(matrix);
    height = NXP_ELEMENT_MATRIX_HEIGHT(matrix);

    mapped = nxp_element_matrix_new(width, height, NULL);
    for (i=0; i<width; i++) 
	for (j=0; j<height; j++) {
	    NXP_ELEMENT_MATRIX_ELEM_XY(NXP_ELEMENT_MATRIX(mapped), i, j) =
		bifunc_map_elem(function, id, NXP_ELEMENT_MATRIX_ELEM_XY(matrix, i, j),
				eval, named_args, context, error);
	}
    return mapped;
}

NxpElement* bifunc_eval_map_matrix(NxpElement *function,
				   NxpElementVariable *variable,
				   NxpElementMatrix *matrix, ETC) 
{
    return bifunc_eval_map_matrix_(function, 
				   variable->ident, matrix, TRUE,
				   named_args, 
				   context, 
				   error);
}


NxpElement* bifunc_eval_map_single_matrix(NxpElement *function,
					  NxpElementMatrix *matrix, ETC)
{
    NxpIdentifier* id;
    NxpElement* result;
    id = nxp_identifier_new_simple("_");
    result = bifunc_eval_map_matrix_(function, id, 
				     matrix, TRUE, named_args, 
				     context, error);
    nxp_identifier_unref(id);
    return result;
}

NxpElement* bifunc_simp_map_matrix(NxpElement *function,
				   NxpElementVariable *variable,
				   NxpElementMatrix *matrix, ETC) 
{
    return bifunc_eval_map_matrix_(function, 
				   variable->ident, matrix, FALSE,
				   named_args, 
				   context, 
				   error);
}


NxpElement* bifunc_simp_map_single_matrix(NxpElement *function,
					  NxpElementMatrix *matrix, ETC)
{
    NxpIdentifier* id;
    NxpElement* result;
    id = nxp_identifier_new_simple("_");
    result = bifunc_eval_map_matrix_(function, id, 
				     matrix, FALSE, named_args, 
 				     context, error);
     nxp_identifier_unref(id);
     return result;
}


/* for[variable, vector, what-to-repeat] */
NxpElement* bifunc_eval_for(NxpElementVariable *var, NxpElementVector *vector, NxpElement *el, ETC)
{
    gint i, size = vector->size;
    GError *err = NULL;
    NxpElement *tmp;

    tmp = nxp_element_void_new();
    for (i = 0; i < size; i++) {
	NXP_ELEMENT_UNREF(tmp);
	tmp = nxp_element_tree_eval_op(NXP_OP_COLON_EQ, NXP_ELEMENT(var), _element_vector_index(vector, i), context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
	NXP_ELEMENT_UNREF(tmp);
	tmp = nxp_element_eval(el, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
    }
    return tmp;
}

NxpElement* bifunc_simp_for(NxpElementVariable *var, NxpElementVector *vector, NxpElement *el, ETC)
{
    gint i, size = vector->size;
    GError *err = NULL;
    NxpElement *tmp;

    tmp = nxp_element_void_new();
    for (i = 0; i < size; i++) {
	NXP_ELEMENT_UNREF(tmp);
	tmp = nxp_element_tree_simplify_op(NXP_OP_COLON_EQ, NXP_ELEMENT(var), _element_vector_index(vector, i), context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
	NXP_ELEMENT_UNREF(tmp);
	tmp = nxp_element_simplify(el, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
    }
    return tmp;
}

/* while[condition, body-forms...] */
NxpElement* bifunc_eval_while(GSList* args, ETC)
{
    NxpElement *condition,  *body_form, *result;
    NxpElement *cond_evaled;
    gboolean    cond_val;
    GSList *arg;
    GError *err = NULL;

    if (!args) {
	nxp_eval_warning(context, _("'while' expression without arguments!"));
	return nxp_element_void_new();
    }
    if (!args->next) {
	nxp_eval_warning(context, _("'while' expression without body!"));
    }
    condition = args->data;
    args      = args->next;
    result    = nxp_element_void_new();

    while (1)
    {
	if(nxp_element_check_thread_abortion_request()) {
	    NXP_ELEMENT_XUNREF(result);
	    return NULL;
	}

	/* Test condition expression */
	cond_evaled = nxp_element_eval(condition, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_XUNREF(cond_evaled);
	    NXP_ELEMENT_XUNREF(result);
	    return NULL;
	}
	cond_val = nxp_element_is_true(cond_evaled, context, &err);
	NXP_ELEMENT_UNREF(cond_evaled);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_XUNREF(result);
	    return NULL;
	}
	if (!cond_val) break;

	/* Evaluate all body forms */
	for (arg = args; arg; arg = arg->next)
	{
	    NXP_ELEMENT_UNREF(result);
	    body_form = NXP_ELEMENT(arg->data);
	    result = nxp_element_eval(body_form, context, &err);
	    if (err) {
		g_propagate_error(error, err);
		NXP_ELEMENT_XUNREF(result);
		return NULL;
	    }
	    if (!result) return NULL; /* this shouldn't happen, but... */
	}
    }
    return result;
}

/* while[condition, body-forms...] */
NxpElement* bifunc_simp_while(GSList* args, ETC)
{
    NxpElement *condition, *simp_cond, *body_form, *result;
    gboolean    cond_val;
    GSList *arg;
    GError *err = NULL;

    if (!args) {
	nxp_eval_warning(context, _("'while' expression without arguments!"));
	return nxp_element_void_new();
    }
    if (!args->next) {
	nxp_eval_warning(context, _("'while' expression without body!"));
    }
    condition = args->data;
    args      = args->next;
    result    = nxp_element_void_new();

    while (1)
    {
	if(nxp_element_check_thread_abortion_request()) {
	    NXP_ELEMENT_XUNREF(result);
	    return NULL;
	}

	/* Test condition expression */
	simp_cond = nxp_element_simplify(condition, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_XUNREF(result);
	    return NULL;
	}
	cond_val = nxp_element_is_true(simp_cond, context, &err);
	NXP_ELEMENT_UNREF(simp_cond);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_XUNREF(result);
	    return NULL;
	}
	if (!cond_val) break;

	/* Evaluate all body forms */
	for (arg = args; arg; arg = arg->next)
	{
	    NXP_ELEMENT_UNREF(result);
	    body_form = NXP_ELEMENT(arg->data);
	    result = nxp_element_simplify(body_form, context, &err);
	    if (err) {
		g_propagate_error(error, err);
		NXP_ELEMENT_XUNREF(result);
		return NULL;
	    }
	    if (!result) return NULL; /* this shouldn't happen, but... */
	}
    }
    return result;
}


NxpElement* bifunc_eval_if(NxpElement *cond, NxpElement *true, NxpElement *false, ETC)
{
    GError *err = NULL;
    gboolean bool;
    NxpElement *value, *result, *branch;

    value = nxp_element_eval(cond, context, &err);
    if (err) {
	g_propagate_error(error, err);
	NXP_ELEMENT_XUNREF(value);
	return NULL;
    }
    
    bool = nxp_element_is_true(value, context, &err);
    NXP_ELEMENT_UNREF(value);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }

    branch = bool? true : false;
    if (branch) {
	result = nxp_element_eval(branch, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_XUNREF(result);
	    return NULL;
	} else
	    return result;
    } else
	return nxp_element_void_new();
}

NxpElement* bifunc_simp_if(NxpElement *cond, NxpElement *true, NxpElement *false, ETC)
{
    GError *err = NULL;
    gboolean bool;
    NxpElement *value, *result, *branch;

    value = nxp_element_simplify(cond, context, &err);
    if (err) {
	g_propagate_error(error, err);
	NXP_ELEMENT_XUNREF(value);
	return NULL;
    }

    bool = nxp_element_is_true(value, context, &err);
    NXP_ELEMENT_UNREF(value);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }

    branch = bool? true : false;
    if (branch) {
	result = nxp_element_simplify(branch, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_XUNREF(result);
	    return NULL;
	} else
	    return result;
    } else
	return nxp_element_void_new();
}


NxpElement* bifunc_eval_if2(NxpElement *cond, NxpElement *true, ETC)
{
    return bifunc_eval_if(cond, true, NULL, NULL, context, error);
}

NxpElement* bifunc_simp_if2(NxpElement *cond, NxpElement *true, ETC)
{
    return bifunc_simp_if(cond, true, NULL, NULL, context, error);
}


NxpElement* bifunc_eval_define_function(NxpElementFunction *function,
					NxpElement *definition, ETC)
{
    char *name;
    GSList* a;
    int  size;
    NxpIdentifier *fullname;

    a = NXP_ELEMENT_LIST(function->simple_args)->list;
    size = nxp_element_list_length(function->simple_args);
    /* Check user function prototype arguments */
    while (a)
    {
	NxpElement *el;
	el = (NxpElement_t*)a->data;
	/* Check if argument is a variable name */
	if (NXP_ELEMENT_GET_TYPE(el) != NXP_ELEMENT_TYPE_VARIABLE) {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK, _("See the manual for the correct syntax of define"));
	    return NULL;
	}
	/* Check if variable name is valid */
	if (nxp_identifier_to_function_argument(NXP_ELEMENT_VARIABLE(el)->ident) == NULL) {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK, _("See the manual for the correct syntax of define"));
	    return NULL;
	}
	a = g_slist_next(a);
    }
    fullname = nxp_identifier_create_function_identifier(NXP_ELEMENT_VARIABLE(function)->ident, size);
    nxp_user_function_define(fullname,
			     function->simple_args,
			     NXP_ELEMENT_REF(definition), context, error);
    name = nxp_identifier_to_string(fullname);
    nxp_output_message(context, _("Function %s defined"), name);
    g_free(name);
    nxp_identifier_unref(fullname);
    return nxp_element_void_new();
}



/* :[expression, expression, ...]
   also known as: progn[expression, expression, ...]
*/
NxpElement* bifunc_eval_two_dot(GSList * args, ETC)
{
    GSList *l;
    NxpElement *el;
    GError *our_error = NULL;

    l = args;
    el = nxp_element_void_new();

    while (l)
    {
        NXP_ELEMENT_XUNREF(el);
        el = nxp_element_eval(l->data, context, &our_error);
	if (our_error) {
	    g_propagate_error(error, our_error);
	    return NULL;
	}
        l = g_slist_next(l);
    }

    return el;
}

NxpElement* bifunc_eval_vector_fill(NxpElement *n, NxpElement *val, ETC)
{
    GError *err = NULL;
    nxp_int_t size, i;
    NxpElementVector *vector;

    size = nxp_element_int_get_si(n, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    vector = NXP_ELEMENT_VECTOR(nxp_element_vector_new(size, NULL));
    for (i = 0; i < size; i++) {
	_element_vector_index(vector, i) = nxp_element_eval(val, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    return NULL;
	}
    }
    return (NxpElement *)vector;
}

NxpElement* bifunc_simp_vector_fill(NxpElement *n, NxpElement *val, ETC)
{
    GError *err = NULL;
    nxp_int_t size, i;
    NxpElementVector *vector;

        size = nxp_element_int_get_si(n, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    vector = NXP_ELEMENT_VECTOR(nxp_element_vector_new(size, NULL));
    for (i = 0; i < size; i++) {
	_element_vector_index(vector, i) = nxp_element_simplify(val, context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_UNREF(vector);
	    return NULL;
	}
    }
    return (NxpElement *)vector;
}

NxpElement* bifunc_join_vector(NxpElementVector *v1, NxpElementVector *v2, gboolean simp, ETC)
{
    guint size1 = v1->size, size2 = v2->size, size = size1 + size2;
    NxpElementVector *retval;
    guint i, j;
    GError *err = NULL;

    retval = (NxpElementVector *)nxp_element_vector_new(size, NULL);
    for (i = 0; i < size1; i++) {
	if (simp)
	    _element_vector_index(retval, i) = nxp_element_simplify(_element_vector_index(v1, i), context, &err);
	else
	    _element_vector_index(retval, i) = nxp_element_eval(_element_vector_index(v1, i), context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_UNREF(retval);
	    return NULL;
	}
    }
    for (j = 0; j < size2; j++, i++) {
	if (simp)
	    _element_vector_index(retval, i) = nxp_element_simplify(_element_vector_index(v2, j), context, &err);
	else
	    _element_vector_index(retval, i) = nxp_element_eval(_element_vector_index(v2, j), context, &err);
	if (err) {
	    g_propagate_error(error, err);
	    NXP_ELEMENT_UNREF(retval);
	    return NULL;
	}
    }

    return (NxpElement *)retval;
}

NxpElement* bifunc_eval_join_vector(NxpElementVector *v1, NxpElementVector *v2, ETC) {
    return bifunc_join_vector(v1, v2, FALSE, named_args, context, error);
}

NxpElement* bifunc_simp_join_vector(NxpElementVector *v1, NxpElementVector *v2, ETC) {
    return bifunc_join_vector(v1, v2, TRUE, named_args, context, error);
}

static NxpElement* bifunc_join_matrix(NxpElementMatrix *m1, NxpElementMatrix *m2, gboolean simp, ETC)
{
    guint h1 = m1->height, h2 = m2->height;
    guint w1 = m1->width, w2 = m2->width, w = w1 + w2;
    NxpElementMatrix *retval;
    guint i, j;
    GError *err = NULL;

    if (h1 != h2) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("join: Matrices don't have the same height"));
	return NULL;
    }
    retval = (NxpElementMatrix *)nxp_element_matrix_new(w, h1, NULL);
    for (j = 0; j < h1; j++) {
	for (i = 0; i < w1; i++) {
	    if (simp)
		_element_matrix_index(retval, j, i) = nxp_element_simplify(_element_matrix_index(m1, j, i), context, &err);
	    else
		_element_matrix_index(retval, j, i) = nxp_element_eval(_element_matrix_index(m1, j, i), context, &err);
	    if (err) {
		g_propagate_error(error, err);
		NXP_ELEMENT_UNREF(retval);
		return NULL;
	    }
	}
	for (i = 0; i < w2; i++) {
	    if (simp)
		_element_matrix_index(retval, j, i + w1) = nxp_element_simplify(_element_matrix_index(m2, j, i), context, &err);
	    else
		_element_matrix_index(retval, j, i + w1) = nxp_element_eval(_element_matrix_index(m2, j, i), context, &err);
	    if (err) {
		g_propagate_error(error, err);
		NXP_ELEMENT_UNREF(retval);
		return NULL;
	    }
	}
    }
    return (NxpElement *)retval;
}

NxpElement* bifunc_eval_join_matrix(NxpElementMatrix *m1, NxpElementMatrix *m2, ETC) {
    return bifunc_join_matrix(m1, m2, FALSE, named_args, context, error);
}

NxpElement* bifunc_simp_join_matrix(NxpElementMatrix *m1, NxpElementMatrix *m2, ETC) {
    return bifunc_join_matrix(m1, m2, TRUE, named_args, context, error);
}

NxpElement* bifunc_vector_to_poly(NxpElementVector *v, NxpElementVariable *var, ETC)
{
    int i;
    NxpElement *tmp, *tmp1, *var1;
    NxpElement *retval;

    var1 = nxp_element_int_new(1);
    retval = nxp_element_int_new(0);
    for (i = 0; i < v->size; i++) {
	tmp = nxp_element_tree_simplify_op(NXP_OP_STAR, _element_vector_index(v, i), var1, context, error);
	tmp1 = nxp_element_tree_simplify_op(NXP_OP_PLUS, tmp, retval, context, error);
	NXP_ELEMENT_UNREF(tmp);
	NXP_ELEMENT_UNREF(retval);
	retval = tmp1;
	tmp = nxp_element_tree_simplify_op(NXP_OP_STAR, var1, NXP_ELEMENT(var), context, error);
	NXP_ELEMENT_UNREF(var1);
	var1 = tmp;
    }
    NXP_ELEMENT_UNREF(var1);
    return retval;
}
   
NxpElement* bifunc_size_vector(NxpElementVector *v, ETC)
{
    return nxp_element_int_new(v->size);
}

NxpElement* bifunc_size_matrix(NxpElementMatrix *m, ETC)
{
    NxpElement *retval;

    retval = nxp_element_vector_new(2, NULL);
    _element_vector_index(NXP_ELEMENT_VECTOR(retval), 0) = nxp_element_int_new(m->width);
    _element_vector_index(NXP_ELEMENT_VECTOR(retval), 1) = nxp_element_int_new(m->height);
    return retval;
}

/* :[expression, expression, ...]
   also known as: progn[expression, expression, ...]
*/
NxpElement* bifunc_simplify_two_dot(GSList * args, ETC)
{
    GSList *l;
    NxpElement *el;
    GError *our_error = NULL;

    l = args;
    el = nxp_element_void_new();

    while (l)
    {
        NXP_ELEMENT_XUNREF(el);
        el = nxp_element_simplify(l->data, context, &our_error);
	if (our_error) {
	    g_propagate_error(error, our_error);
	    return NULL;
	}
        l = g_slist_next(l);
    }

    return el;
}



static __inline__ NxpElement*
bifunc_eval_progx(GSList* args, gint x, ETC)
{
    GSList *l;
    NxpElement *el, *result = NULL;
    gint arity, nth;
    arity = g_slist_length(args);

    if (arity < x) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("prog%i: insufficient arguments"), x);
	return NULL;
    }
    for (l = args, nth = 1; l; l = l->next, nth++)
    {
        el = nxp_element_eval(l->data, context, NULL);
        if (el) {
	    if (nth == x)
		result = el;
	    else
		NXP_ELEMENT_UNREF(el);
	}
    }
    return result;
}



/* prog1[expression, expression, ...] -- as in Lisp */
NxpElement* bifunc_eval_prog1(GSList* args, ETC)
{
    return bifunc_eval_progx(args, 1, NULL, context, error);
}



/* prog2[expression, expression, ...] -- as in Lisp */
NxpElement* bifunc_eval_prog2(GSList* args, ETC)
{
    return bifunc_eval_progx(args, 2, NULL, context, error);
}


NxpElement* bifunc_eval_comment(GSList *args, ETC)
{
    return nxp_element_void_new();
}

