/* -*- Mode: C; c-file-style: "bsd"; -*- */
#include <stdio.h>
#include "element_output.h"
#include "element_type.h"
#include "config.h"
#include "i18n.h"

/* Takes an element and prints it to the stdout. */
void nxp_elem_printer(NxpElement* a)
{
    GString *gstr;
#ifdef NXP_ELEMENT_DEBUG
    GSList *trace_list;
#endif
    NxpElementTypeInfo *info;
    int i;

    a = NXP_ELEMENT(a);
    if (!a) return;
    gstr = nxp_element_repr(a);
    fprintf(stderr, "NxpElement at %p (ref_count=%i, type='%s'): %s\n", a, 
	    a->ref_count, NXP_ELEMENT_GET_TYPE_INFO(a)->name, gstr->str);

    for (i = 0, info = NXP_ELEMENT_GET_TYPE_INFO(a); info; info = info->parent)
	fprintf(stderr, "\ttype %d : %s\n", i++, info->name);

#ifdef NXP_ELEMENT_DEBUG
    for (trace_list = g_slist_reverse(a->ref_trace);
	 trace_list; trace_list = trace_list->next)
    {
	NxpElementRefTraceItem *trace;
	trace = trace_list->data;
	printf("  %-5s in %s,\tline %i,\tof file %s\n", trace->ref? "ref" : "unref",
	       trace->function, trace->line, trace->file);
    }
    g_slist_free(trace_list);
#endif
    g_string_free(gstr, TRUE);
}




/**
 * nxp_element_repr: - get a string representation of an element
 * @elem: the element
 *
 * Create a textual representation of @elem in format presentable to
 * the user. The resulting string might not be able to be parsed back
 * into an element -- it's just a user representation!
 * 
 * Return value: a GString with the textual representation
 **/
GString* nxp_element_repr(NxpElement *elem)
{
    NxpElementTypeInfo *typeinfo;
    GString *s;
    g_return_val_if_fail(elem, NULL);
    typeinfo = nxp_element_type_get_info(NXP_ELEMENT_GET_TYPE(elem));
    g_return_val_if_fail(typeinfo, NULL);
    if (typeinfo->repr_func) 
	return typeinfo->repr_func(elem, NULL);
    g_warning("This is a bug: no registered repr() function for type '%s'.",
	      NXP_ELEMENT_GET_TYPE_NAME(elem));
    return s = g_string_new(NULL), 
	g_string_printf(s, _("#<element of type '%s'>"), NXP_ELEMENT_GET_TYPE_NAME(elem)), 
	s;
}

/**
 * nxp_element_str: - get a parse-able string representation of an element
 * @elem: the element
 * 
 * Create a textual representation of @elem in format understandable
 * by the parser. Parsing the resulting string should give an element
 * equivalent to @elem.
 * 
 * Return value: a GString with the string representation
 **/
GString* nxp_element_str(NxpElement *elem)
{
    NxpElementTypeInfo *typeinfo;
    GString *s;
    g_return_val_if_fail(elem, NULL);
    typeinfo = nxp_element_type_get_info(NXP_ELEMENT_GET_TYPE(elem));
    g_return_val_if_fail(typeinfo, NULL);
    if (typeinfo->str_func) 
	return typeinfo->str_func(elem, NULL);
    g_warning("This is a bug: no registered str() function for type '%s'.",
	      NXP_ELEMENT_GET_TYPE_NAME(elem));
    return s = g_string_new(NULL), 
	g_string_printf(s, _("#<element of type '%s'>"), NXP_ELEMENT_GET_TYPE_NAME(elem)), 
	s;
}




/**
 * elem2string: - legacy function -- do not use
 * @string: 
 * @a: 
 * 
 * 
 * 
 * Return value: priority
 **/
int elem2string(GString **string, NxpElement* a)
{
    *string = nxp_element_str(a);
    return -1;
}
