/* -*- c -*- */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include "identifier.h"
#include "message.h"
#include "eval_context.h"
#include "nxp_config.h"
#include "parser/parser.h"
#include "types/variable.h"
#include "i18n.h"

//#define DEBUG


/* The (internal) format of the identifiers is a list of strings, one
   string for each name component. A NULL as first or last component
   represents a path separator.
   
   Examples:
   x		is stored as	("x")
   \x		is stored as	(NULL, "x")
   x\		is stored as	("x", NULL)
   x\y\z	is stored as	("x", "y" ,"z")
   ..\x		is stored as	("..", "x")

*/


/* this structure should not be accessed directly -- use the existing
   functions or make your own! */
struct _NxpIdentifier {
    volatile guint ref_count;
    GStringChunk *string_chunk;
    GSList *name_components;
};

/* number of identifiers in memory at any time -- for debugging */
volatile int nxp_identifier_global_count = 0;


inline static gboolean nxp_component_is_path_up(const char *component)
{
    return (strcmp(component, NXP_IDENTIFIER_PATH_UP) == 0);
}




/**
 * nxp_identifier_new:
 * 
 * Create a new NxpIdentifier
 * 
 * Return value: the new #NxpIdenfifier
 **/
NxpIdentifier *nxp_identifier_new(void)
{
    NxpIdentifier *self;

    self = g_new(NxpIdentifier, 1);
    self->string_chunk = g_string_chunk_new(16);
    self->name_components = NULL;
    self->ref_count = 1;
    ++nxp_identifier_global_count;
    return self;
}




/**
 * nxp_identifier_new_simple:
 * @name: const string with single component
 * 
 * Get new identifier from const string
 * 
 * Return value: new identifier
 **/
NxpIdentifier *nxp_identifier_new_simple(const char *name)
{
    NxpIdentifier *self;
    self = nxp_identifier_new();
    nxp_identifier_append_component(self, name);
    return self;
}




/**
 * nxp_identifier_ref:
 * @self: the NxpIdentifier
 * 
 * Reference NxpIdentifier object
 *
 * Return value: self
 **/
NxpIdentifier *nxp_identifier_ref(NxpIdentifier *self)
{
    g_return_val_if_fail(self, NULL);
    ++self->ref_count;
    return self;
}




/**
 * nxp_identifier_unref:
 * @self: the Idenfier
 * 
 * Unreference NxpIdentifier object
 *
 * Return value: %NULL
 **/
NxpIdentifier *nxp_identifier_unref(NxpIdentifier *self)
{
    g_return_val_if_fail(self, NULL);
    if (--self->ref_count == 0)
    {
	g_slist_free(self->name_components);
	g_string_chunk_free(self->string_chunk);
	g_free(self);
	--nxp_identifier_global_count;
    }
    return NULL;
}




/**
 * nxp_identifier_append_component:
 * @self: NxpIdentifier
 * @component: component to append;
 * 
 * Append one name component to the identifier. For use in the parser.
 **/
void nxp_identifier_append_component(NxpIdentifier *self, const char *component)
{
    char *s;
    GSList *last_component;
    g_return_if_fail(self);
    g_return_if_fail(component);
    last_component = g_slist_last(self->name_components);
    s = g_string_chunk_insert_const(self->string_chunk, component);
    if (last_component) {
	if (last_component == self->name_components) {
	    /* if last is also first */
	    if(last_component->data == NULL)
		self->name_components = g_slist_append(self->name_components, s);
	    else {
		g_warning("Two consecutive components!");
		return;
	    }
	} else {
	    if (last_component->data == NULL)
		last_component->data = s;
	    else {
		g_warning("Two consecutive components!");
		return;
	    }
	}
    } else {
	self->name_components = g_slist_append(self->name_components, s);
    }
}



/**
 * nxp_identifier_prepend_separator:
 * @self: NxpIdentifier
 *
 * Prepend a separator to the given identifier.
 **/
void nxp_identifier_prepend_separator(NxpIdentifier *self)
{
    if (!self->name_components)
	self->name_components = g_slist_prepend(self->name_components, NULL);
    else {
	if (self->name_components->data == NULL)
	    g_warning("Two consecutive separators!");
	else
	    self->name_components = g_slist_prepend(self->name_components, NULL);
    }
}
    

/**
 * nxp_identifier_append_separator:
 * @self: identifier
 * 
 * Append a path separator to an identifier. For use in the parser.
 **/
void nxp_identifier_append_separator(NxpIdentifier *self)
{
    GSList *last_component;
    g_return_if_fail(self);
    last_component = g_slist_last(self->name_components);

    if (!last_component)
	self->name_components = g_slist_append(self->name_components, NULL);
    else {
	if (last_component->data == NULL)
	    g_warning("Two consecutive separators!");
	else
	    self->name_components = g_slist_append(self->name_components, NULL);
    }
}



/**
 * nxp_identifier_is_relative:
 * @self: identifier
 * 
 * Find out if the identifier has a relative path, like x\y\z or ..\r, 
 * or an absolute path, like \x\y\z.
 * 
 * Return value: %TRUE if path is relative, %FALSE if it is absolute.
 **/
gboolean nxp_identifier_is_relative(NxpIdentifier *self)
{
    g_return_val_if_fail(self, TRUE);
    g_return_val_if_fail(self->name_components, TRUE);
    return (self->name_components->data != NULL);
}



/**
 * nxp_identifier_is_dir:
 * @self: identifier
 * 
 * Find out if the identifier points to dir (ends with %PATH_SEP) or not.
 * 
 * Return value: %TRUE if identifier is dir, %FALSE if not.
 **/
gboolean nxp_identifier_is_dir(NxpIdentifier *self)
{
    g_return_val_if_fail(self, FALSE);
    g_return_val_if_fail(self->name_components, TRUE);
    return (g_slist_last(self->name_components)->data == NULL);
}



/**
 * nxp_identifier_resolve_from_dir:
 * @self: NxpIdentifier
 * @cwd: the current working directory (NxpNameSpace)
 * @ns_ptr: address of variable that will contain the NxpNameSpace
 * refered to by the identifier, or NULL if the NxpNameSpace could not be
 * found (in that case, it's an error situation).
 * @name_ptr: address of variable that will contain the name component
 * of the identifier, or NULL if it's a "directory" or there was a
 * parse error.
 * 
 * Resolve the identifier, i.e., find out which NxpNameSpace and name it refers to.
 **/
void nxp_identifier_resolve_from_dir(NxpIdentifier *self, NxpNameSpace *cwd,
				     NxpNameSpace **ns_ptr, const char **name_ptr,
				     NxpEvalContext* context)
{
    GSList *l, *first;
    const char *component;
    g_return_if_fail(self);
    g_return_if_fail(ns_ptr);
    g_return_if_fail(name_ptr);
    *ns_ptr = NULL;
    *name_ptr = NULL;
    g_return_if_fail(self->name_components);
#ifdef DEBUG
    nxp_identifier_debug_print(self, __FUNCTION__);
#endif
    first = self->name_components;
    /* to speed up, check the most common situation */
    if (first->next == NULL && first->data) {
	*name_ptr = first->data;
	*ns_ptr = nxp_name_space_find_child(cwd, first->data);
	if (!*ns_ptr) *ns_ptr = cwd;
	return;
    }
    for (l = self->name_components; l; l = l->next)
    {
#ifdef DEBUG
	GString *ns_fullname;
#endif
	component = l->data;
#ifdef DEBUG
	ns_fullname = nxp_name_space_get_full_path_name(cwd);
	g_message("%s:  cwd is '%s' and looking at component '%s'",
		  __FUNCTION__, ns_fullname->str, component);
	g_string_free(ns_fullname, TRUE);
#endif
	if (component) {
	    if (nxp_component_is_path_up(component)) {
		if (!(cwd = cwd->parent)) {
		    g_warning(_("Could not resolve identifier: ascending above root name space.")); /* FIXME: GError ** */
		    return;
		}
	    } else {		/* if (component_is_path_up) */
		if (l->next == NULL) {
		    /* These three lines check for a child namespace with that name
		       This way, ::namespace can be used instead of ::namespace:: */
		    /* it's never a namespace because the parser forbids it, so
		     * I comment them.
		    *ns_ptr = nxp_name_space_find_child(cwd, component);
		    g_assert(*ns_ptr == NULL);
		    if (!*ns_ptr) *ns_ptr = cwd;
		    cwd = *ns_ptr;*/

		    *name_ptr = component;
#ifdef DEBUG
		    g_message("%s: last component = '%s'", __FUNCTION__,
			      component);
#endif
		} else {
		    cwd = nxp_name_space_find_child(cwd, component);
		    if (!cwd) {
			g_warning(_("Could not resolve identifier: no name space named '%s' found"), component); /* FIXME: GError ** */
			*ns_ptr = NULL;
			*name_ptr = NULL;
			return;
		    }
		}
	    }
	} else {		/* if (component) */
	    if (l == first) {
		cwd = context->root_name_space;
	    } else if (l->next == NULL) {
		/* is dir */
	    } else {
		g_warning("Malformed identifier. "
			  "NULL was found as a middle component.");
	    }
	}
    }
    *ns_ptr = cwd;
}




/**
 * nxp_identifier_resolve:
 * @self: see nxp_identifier_resolve_from_dir()
 * @context: an execution context (#NxpEvalContext)
 * @ns_ptr: see nxp_identifier_resolve_from_dir()
 * @name_ptr: see nxp_identifier_resolve_from_dir()
 * 
 * Same as nxp_identifier_resolve_from_dir(), but gets cwd from @context.
 **/
void nxp_identifier_resolve(NxpIdentifier *self, NxpEvalContext *context,
			    NxpNameSpace **ns_ptr, const char **name_ptr)
{
    nxp_identifier_resolve_from_dir(self, context->cwd, ns_ptr, name_ptr, context);
}



/**
 * nxp_identifier_equal:
 * @a: an identifier
 * @b: another identifier
 * @context: eval context
 * 
 * Compare two identifiers to see if they point to the same
 * variable/function or , if they're both dirs, if they point to the
 * same dir.
 * 
 * Return value: %TRUE if they are equal, %FALSE otherwise.
 **/
gboolean nxp_identifier_equal(NxpIdentifier *a, NxpIdentifier *b, NxpEvalContext *context)
{
    NxpNameSpace *ns_a, *ns_b;
    const char *name_a, *name_b;

    g_return_val_if_fail(a, FALSE);
    g_return_val_if_fail(b, FALSE);
    nxp_identifier_resolve(a, context, &ns_a, &name_a);
    g_return_val_if_fail(ns_a, FALSE);
    nxp_identifier_resolve(b, context, &ns_b, &name_b);
    g_return_val_if_fail(ns_b, FALSE);

    if (!name_a && !name_b)
	return (ns_a == ns_b);
    else if (name_a && name_b)
	return ns_a == ns_b && strcmp(name_a, name_b) == 0;
    else
	return FALSE;
}


/**
 * nxp_identifier_to_function_argument:
 * @self: identifier
 * 
 * Returns the name component of the identifier if it is admissible as
 * function argument name, or %NULL if it is not.
 * 
 * Return value: const string with argument name or %NULL. DON'T touch this string!
 **/
const char *nxp_identifier_to_function_argument(NxpIdentifier *self)
{
    g_return_val_if_fail(self, NULL);
    if (!self->name_components)
	return NULL;
    else if (self->name_components->next) 
	return NULL;		/* there are at least two name components */
    else {
	const char *component;
	component = self->name_components->data;
	if (nxp_component_is_path_up(component))
	    return NULL;
	else
	    return component;
    }
}




/**
 * nxp_identifier_to_function_definition: - find function definition from identifier and arity.
 * @self: identifier
 * @arity: the expected arity of the function; if a definition isn't
 * found with the given arity, %-1 is also tried.
 * @fullname: %NULL or address of variable that will receive the
 * string containing the full name ("name/arity") of the function; the
 * caller will be responsible for freeing this string, unless there
 * was an error.
 * 
 * Tries to find a function definition that matches the given identifier and arity.
 * 
 * Return value: a reference to the definition if found, %NULL otherwise.
 **/
NxpDefinition *nxp_identifier_to_function_definition(NxpIdentifier *self, 
						     NxpEvalContext *context,
						     int arity, char **fullname)
{
    NxpNameSpace *ns;
    const char *name;
    char *name_with_arity;
    NxpDefinition *def;

    g_return_val_if_fail(self, NULL);
    if (fullname) *fullname = NULL;
    
    nxp_identifier_resolve(self, context, &ns, &name);
    g_return_val_if_fail(ns, NULL);
    g_return_val_if_fail(name, NULL);

    name_with_arity = g_strdup_printf("%s/%d", name, arity);
    def = nxp_name_space_lookup_definition_r(ns, name_with_arity, FALSE);
    if (!def)
    {
	char *another_chance;
	/* See if there is a function that takes a variable number of
           arguments with this name. */
	another_chance = g_strdup_printf("%s/-1", name);
	def = nxp_name_space_lookup_definition_r(ns, another_chance, FALSE);
	if (!def)
	{
	    g_warning(_("Undefined Function: %s, or function has a different arity"), name_with_arity); /* FIXME: GError ** */
	} else {
	    free(name_with_arity);
	    name_with_arity = another_chance;
	}
    }
    if (fullname)
	*fullname = name_with_arity;
    else
	g_free(name_with_arity);
    return def;
}



/**
 * nxp_identifier_create_function_identifier:
 * @base: plain identifier
 * @arity: arity or -1
 * 
 * Create a new identifier from @base with "/"@arity appended.
 * 
 * Return value: the newly created identifier.
 **/
NxpIdentifier* nxp_identifier_create_function_identifier(NxpIdentifier *base, int arity)
{
    GSList *l, *first;
    NxpIdentifier *ident;
    const char *component;
    g_return_val_if_fail(base, NULL);
    g_return_val_if_fail(base->name_components, NULL);
    g_return_val_if_fail(!nxp_identifier_is_dir(base), NULL);
    
    ident = nxp_identifier_new();
    first = base->name_components;

    for (l = first; l; l = l->next)
    {
	component = l->data;
	if (!component) {
	    if (l == first)
		nxp_identifier_append_separator(ident);
	    else 
		g_warning("Bad identifier: NULL as middle component!");
	} else {
	    if (l->next == NULL) {
		char *name_with_arity;
		name_with_arity = g_strdup_printf("%s/%i", component, arity);
		nxp_identifier_append_component(ident, name_with_arity);
		g_free(name_with_arity);
	    } else {
		nxp_identifier_append_component(ident, component);
		nxp_identifier_append_separator(ident);
	    }
	}
    }
    return ident;
}


/**
 * nxp_identifier_to_gstring:
 * @self: the identifier
 * 
 * Create a string (GString) representation of an identifier.
 * 
 * Return value: a GString containing a representation of the identifier. 
 * It must be g_string_free'd by the caller.
 **/
GString *nxp_identifier_to_gstring(NxpIdentifier *self)
{
    GSList *l;
    const char *component;
    GString *s;
    g_return_val_if_fail(self, NULL);
    if (!self->name_components)
	return g_string_new(_("#<empty identifier>"));
    s = g_string_new(NULL);

    for (l = self->name_components; l; l = l->next)
    {
	if ((component = l->data))
	    s = g_string_append(s, component);
	if (l->next)
	    s = g_string_append(s, NXP_IDENTIFIER_PATH_SEP);
    }
    return s;
}


/**
 * nxp_identifier_to_string:
 * @self: the identifier
 * 
 * Create a string (char *) representation of an identifier.
 * 
 * Return value: a string containing a representation of the identifier. 
 * It must be g_free'd by the caller.
 **/
char *nxp_identifier_to_string(NxpIdentifier *self)
{
    GString *gstr;
    char *str;
    g_return_val_if_fail(self, NULL);
    gstr = nxp_identifier_to_gstring(self);
    g_return_val_if_fail(gstr, NULL);
    str = gstr->str;
    g_string_free(gstr, FALSE);
    return str;
}


/**
 * nxp_identifier_debug_print: -show contents of identifier for debugging
 * @self: the identifier to show
 * @message_prefix: message to prepend before the final message
 * 
 * Print a debugging message containing the string representation of
 * the identifier.
 **/
void nxp_identifier_debug_print(NxpIdentifier *self, const char *message_prefix)
{
    GString *gstr;

    g_return_if_fail(self);
    gstr = nxp_identifier_to_gstring(self);
    g_message("%s: identifier at %p = \"%s\"", message_prefix, self, gstr->str);
    g_string_free(gstr, TRUE);
}



/**
 * nxp_identifier_is_plain:
 * @self: the identifier
 * 
 * Find out if it is a plain identifier, i.e., it is not a dir and has
 * a simple, non-null, name component, like 'x'.
 * 
 * Return value: %TRUE if plain, %FALSE if not.
 **/
gboolean nxp_identifier_is_plain(NxpIdentifier *self)
{
    g_return_val_if_fail(self, FALSE);
    return self->name_components && self->name_components->data;
}

NxpIdentifier* nxp_identifier_new_with_full_path (int levels, ...) {
    NxpIdentifier* new;
    char* name;
    va_list argument_list;

    new = nxp_identifier_new();

    va_start(argument_list, levels);
    while(levels--) {
	nxp_identifier_append_separator(new);
	name = va_arg(argument_list, char*);
	nxp_identifier_append_component(new, name);
    }

    va_end(argument_list);
    return new;
}


/**
 * nxp_identifier_from_string:
 * @path: identifier string; path separators are handled correctly
 * @error: it is possible that (EXP_PARSE_ERROR_DOMAIN,
 * EXP_PARSE_ERROR) is set if the string cannot be parsed or
 * represents something other than an identifer.
 * 
 * Parse a string to obtain an identifier. Please not that this is
 * different than nxp_identifier_new_simple, which doesn't handle path
 * separators.
 * 
 * Return value: the idenfier
 **/
NxpIdentifier* nxp_identifier_from_string(const char      *path,
					  NxpEvalContext  *context,
					  GError         **error)
{
    NxpElement    *var_elem;
    GError        *error1 = NULL;
    NxpIdentifier *retval;

    if (error == NULL)
	error = &error1;
    var_elem = exp_parse_string(path, context, error);
    if (*error || !var_elem)
	return NULL;
    if (NXP_ELEMENT_GET_TYPE(var_elem) != NXP_ELEMENT_TYPE_VARIABLE) {
	g_set_error(error, EXP_PARSE_ERROR_DOMAIN, EXP_PARSE_ERROR,
		    "'%s' does not look like an identifier", path);
	retval = NULL;
    } else {
	retval = NXP_ELEMENT_VARIABLE(var_elem)->ident;
	nxp_identifier_ref(retval);
    }
    NXP_ELEMENT_UNREF(var_elem);
    g_clear_error(&error1);
    return retval;
}

