/* -*- Mode: C; c-file-style: "stroustrup"; encoding: latin-1-unix -*- */
#include <pthread.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <sys/signal.h>

#include "namespace.h"
#include "identifier.h"
#include "prototypes.h"
#include "nxp_debug.h"



#define nxp_name_space_set_modified(ns) 	g_atomic_int_set(&ns->modified, 1)



static void _definition_destroy(NxpDefinition *def);



/**
 * nxp_definition_ref:
 * @def: a #NxpDefinition
 * 
 * Increase reference count of @def
 * 
 * Return value: @def
 **/
inline NxpDefinition* nxp_definition_ref(NxpDefinition *def)
{
    g_return_val_if_fail(def, NULL);
#ifdef DEBUG_ELEMENTS
    if (def->ref_count == 0)
	G_BREAKPOINT();
#endif
    ++def->ref_count;
    return def;
}



/**
 * nxp_definition_unref:
 * @def: a #NxpDefinition
 * 
 * Decrease reference count of @def
 * 
 **/
inline void nxp_definition_unref(NxpDefinition *def)
{
    g_return_if_fail(def);
    if (--def->ref_count == 0)
	_definition_destroy(def);
}


/*
  ----------------------------------------------
  NxpNameSpace -- a collection of NxpDefinition's
  ----------------------------------------------
*/


/**
 * nxp_name_space_find_child:
 * @self: a #NxpNameSpace
 * @name: the name of the child to search for
 * 
 * Find the child of NxpNameSpace @self that has the name @name.
 * 
 * Return value: the NxpNameSpace if found, or NULL
 **/
NxpNameSpace* nxp_name_space_find_child(NxpNameSpace *self, const char *name)
{
    NxpNameSpace *child, *the_child = NULL;
    g_return_val_if_fail(self, NULL);
    nxp_name_space_lock(self);
    for (child = self->child; child; child = child->next)
	if (child->name && strcmp(child->name, name) == 0) {
	    the_child = child;
	    break;
	}
    nxp_name_space_unlock(self);
    return the_child;
}


static void nxp_name_space_construct(NxpNameSpace *self, NxpNameSpace *parent, const char *name)
{
    pthread_mutexattr_t attr;

    self->definitions = g_hash_table_new(g_str_hash, g_str_equal);
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);
    pthread_mutex_init(&self->mutex, &attr);
    pthread_mutexattr_destroy(&attr);
    if (!parent)
	self->parent = NULL;
    else {
	/* TODO: check if there is another NxpNameSpace with the same
           name and the same parent. Would that be worth it? Wouldn't
           it be slow? */
	self->next = parent->child;
	parent->child = self;
	self->parent = parent;
    }
    self->reentrant = TRUE;
    self->name = name? g_strdup(name): NULL;
    self->listeners = NULL;
    self->modified = -1;	/* negative or zero value means false */
    if (self->parent)
	nxp_name_space_set_modified(self->parent);
}


/**
 * nxp_name_space_new:
 * @parent: parent namespace or %NULL
 * @name: name of the namespace or %NULL
 * 
 * Creates a namespace object.
 * 
 * Return value: the newly created namespace
 **/
NxpNameSpace *nxp_name_space_new(NxpNameSpace *parent, const char *name)
{
    NxpNameSpace *self;

    /* Check if the namespace already exists, unless the new namespace is anonymous. */
    if (parent && name && (self = nxp_name_space_find_child(parent, name))) return self;

    /* it does not exists... we can create it */
    self = g_new0(NxpNameSpace, 1);
    nxp_name_space_construct(self, parent, name);
    self->module = nxp_current_module;
    if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) {
	GString *s;
	s = nxp_name_space_get_full_path_name(self);
	g_message("Creating name space '%s'.", s->str);
	g_string_free(s, TRUE);
    }
    return self;
}




/**
 * nxp_name_space_add_definition:
 * @self: namespace
 * @def: the definition to add; will take ownership of the reference!
 * 
 * Inserts definition into the hash table. Any previous definition
 * with the same name is removed and unreferenced.
 * Non-reentrant! Use nxp_name_space_lock() first! 
 * 
 **/
void nxp_name_space_add_definition(NxpNameSpace *self, NxpDefinition *def)
{
    NxpDefinition *old_def;
    g_return_if_fail(self);
    g_return_if_fail(def);
    g_return_if_fail(def->name);

    def->module = nxp_current_module;

#ifdef NXP_NAMESPACE_DEBUG
    if (def->parent_namespace) {
	fprintf(stderr, "WARNING: Adding definition to namespace,"
		" but definition already has a parent!\n");
	if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) G_BREAKPOINT();
    }
    def->parent_namespace = self;

#ifdef DEBUG_ELEMENTS
  	GString *s; 
  	s = nxp_name_space_get_full_path_name(self); 
  	g_message("Adding definition of '%s' to name space '%s'.", 
  		  def->name, s->str); 
  	g_string_free(s, TRUE); 
#endif

#endif
    old_def = g_hash_table_lookup(self->definitions, def->name);
    /* If the old definition isn't removed first, the hash table will
       keep using the same key, which is owned by the old definition,
       which is about to be destroyed. */
    if (old_def) {
	//	g_hash_table_freeze(self->definitions);
	g_hash_table_remove(self->definitions, old_def->name);
#ifdef NXP_NAMESPACE_DEBUG
	old_def->parent_namespace = NULL;
#endif
    }
    g_hash_table_insert(self->definitions, def->name, def);

    if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) {
	NxpDefinition *new_def = NULL;
	gchar *key = NULL;
	/*  */
	if (!g_hash_table_lookup_extended(self->definitions, def->name, 
					  (gpointer)&key, (gpointer)&new_def))
	    g_warning("NxpDefinition couldn't be found after insertion!");
	else {
	    if (key != new_def->name) {
		if (key == old_def->name)
		    g_warning("Hash table uses old definition's key!");
		else
		    g_warning("Hash table uses a different key!");
	    }
	}
    }
    if (old_def) {
#ifdef NXP_NAMESPACE_DEBUG
	g_message("Definition %s (ref_count=%i) has been replaced",
		  old_def->name, old_def->ref_count);
#endif
	nxp_definition_unref(old_def);
    }
    nxp_name_space_set_modified(self);
}



/**
 * nxp_name_space_add_definition_r:
 * @self: the #NxpNameSpace
 * @def: the #NxpDefinition to add
 * 
 * thread safe version of nxp_name_space_add_definition
 **/
void nxp_name_space_add_definition_r(NxpNameSpace *self, NxpDefinition *def)
{
    nxp_name_space_lock(self);
    nxp_name_space_add_definition(self, def);
    nxp_name_space_unlock(self);
}




/**
 * nxp_name_space_lookup_definition_r:
 * @self: the #NxpNameSpace
 * @name: name of the #NxpDefinition to search
 * 
 * thread safe version of nxp_name_space_lookup_definition
 * 
 * Return value: 
 **/
NxpDefinition* nxp_name_space_lookup_definition_r(NxpNameSpace *self, const char *name, gboolean rec)
{
    NxpDefinition *def;
    nxp_name_space_lock(self);
    def = nxp_name_space_lookup_definition_extended(self, name, NULL, rec);
    nxp_name_space_unlock(self);
    return def;
}

/**
 * nxp_name_space_lookup_definition:
 * @self: namespace to (start) search
 * @name: definition name
 * 
 * Search namespace for definition. If not found there search parents recursively.
 * 
 * Return value: a reference to the definition if found, NULL otherwise.
 **/
NxpDefinition* nxp_name_space_lookup_definition(NxpNameSpace *self, const char *name)
{
    return nxp_name_space_lookup_definition_extended(self, name, NULL, TRUE);
}


/**
 * nxp_name_space_lookup_definition_extended:
 * @self: namespace
 * @name: name string
 * @actual_name_space: If not NULL, store here a pointer to the NxpNameSpace where (if) it actually was found.
 * @recursive: if TRUE, parent NxpNameSpaces are also searched if the definition isn't found in this one.
 * 
 * Lookup definition in the NxpNameSpace. Non-reentrant!
 *  
 * 
 * Return value: a reference to the definition if found, or NULL if it wasn't found.
 **/
NxpDefinition* nxp_name_space_lookup_definition_extended(NxpNameSpace *self, const char *name,
							 NxpNameSpace **actual_name_space, 
							 gboolean recursive)
{
    NxpNameSpace *parent;
    NxpDefinition *def = NULL;

    g_return_val_if_fail(self, NULL);
    g_return_val_if_fail(name, NULL);

    /* first lookup inside self */
    parent = recursive? self->parent: NULL;
    def = g_hash_table_lookup(self->definitions, name);
    if (def) nxp_definition_ref(def);
    if (def && actual_name_space) *actual_name_space = self;
    if (!def && parent) {
	/* then, lookup inside parent */
	def = nxp_name_space_lookup_definition_extended_r(
	    parent, name, actual_name_space, recursive);
	if (actual_name_space) {
	    if (def)
		*actual_name_space = parent;
	    else
		*actual_name_space = NULL;
	}
    }
    return def;
}


/**
 * nxp_name_space_lookup_definition_extended_r:
 * @self: see nxp_name_space_lookup_definition_extended()
 * @name: see nxp_name_space_lookup_definition_extended()
 * @actual_name_space: see nxp_name_space_lookup_definition_extended()
 * @recursive: see nxp_name_space_lookup_definition_extended()
 * 
 * thread safe version of nxp_name_space_lookup_definition_extended()
 * 
 * Return value: see nxp_name_space_lookup_definition_extended()
 **/
NxpDefinition* nxp_name_space_lookup_definition_extended_r(NxpNameSpace *self, const char *name,
						    NxpNameSpace **actual_name_space, 
						    gboolean recursive)
{
    NxpDefinition *def;
    nxp_name_space_lock(self);
    def = nxp_name_space_lookup_definition_extended(self, name, actual_name_space, recursive);
    nxp_name_space_unlock(self);
    return def;
}

/* What's this? */
gboolean nxp_name_space_resolve_path(NxpNameSpace *cwd, NxpIdentifier *ident)
{
    return 0;
}




/**
 * nxp_name_space_remove_definition:
 * @self: the namespace
 * @def: the definition to remove, which must belong to the given namespace
 * 
 * Remove the definition from the NxpNameSpace. Non-reentrant!
 **/
void nxp_name_space_remove_definition(NxpNameSpace *self, NxpDefinition *def)
{
    g_return_if_fail(self);
    g_return_if_fail(def);
    g_return_if_fail(def->name);

    nxp_name_space_lock(self);
    g_hash_table_remove(self->definitions, def->name);
#ifdef NXP_NAMESPACE_DEBUG
    def->parent_namespace = NULL;
#endif
    nxp_name_space_unlock(self);
    nxp_definition_unref(def);
    nxp_name_space_set_modified(self);
}



/**
 * nxp_name_space_remove_definition_r:
 * @self: see nxp_name_space_remove_definition()
 * @def: see nxp_name_space_remove_definition()
 * 
 * thread safe version of nxp_name_space_remove_definition
 **/
void nxp_name_space_remove_definition_r(NxpNameSpace *self, NxpDefinition *def)
{
    nxp_name_space_lock(self);
    nxp_name_space_remove_definition(self, def);
    nxp_name_space_unlock(self);
}




static gboolean nxp_name_space_remove_func(gpointer key, 
				       gpointer value, 
				       gpointer user_data)
{
    NxpDefinition *def = value;
#ifdef NXP_NAMESPACE_DEBUG
    def->parent_namespace = NULL;
    if (def->ref_count > 1) {
	g_message("Destroying namespace but definition '%s' has"
		  " ref_count = %i -- will remain in memory!", 
		  def->name, def->ref_count);
	if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) G_BREAKPOINT();
    }
#endif
    nxp_definition_unref(def);
    return TRUE;
}


/**
 * nxp_name_space_destroy:
 * @self: the #NxpNameSpace to destroy
 * 
 * Destroys the given NxpNameSpace. Recursively destroys any children of
 * this NxpNameSpace. Removes itself from the child list of its
 * parent. Every definition in it is also unreferenced.
 **/
void nxp_name_space_destroy(NxpNameSpace *self)
{
    NxpNameSpace *ns;
    g_return_if_fail(self);
    

    if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) {
	GString *s;
	s = nxp_name_space_get_full_path_name(self);
	g_message("Destroying name space '%s'.", s->str);
	g_string_free(s, TRUE);
    }
    /* Destroy any remaining children. */
    /* (removes the first child until there are no more children. */
    while (self->child)
	nxp_name_space_destroy(self->child);
    g_hash_table_foreach_remove(self->definitions, nxp_name_space_remove_func, NULL);
    g_hash_table_destroy(self->definitions);
    g_free(self->name);
    /* Remove oneself from the parent's child list. */
    if (self->parent) {
	/* TODO: locking of parent, maybe? */
	ns = self->parent->child;
	if (ns == self) {
	    /* We are the first child. Need to update parent's child
               pointer. */
	    self->parent->child = self->next;
	} else {
	    /* We are not the first child. Find our position in the
               list and remove the link to us. */
	    while (ns) {
		if (ns->next == self) {
		    ns->next = self->next;
		    break;
		}
		ns = ns->next;
		if (!ns)
		    g_warning("Could not find self in the NxpNameSpace linked list.");
	    }
	}
    }
/*      element_container_destroy(self->container); */
    pthread_mutex_destroy(&self->mutex);
    /* TODO: free listeners */
    g_free(self);
}


/**
 * nxp_name_space_get_full_path_name:
 * @self: the name space
 * 
 * Return a string with path of to this name space
 * 
 * Return value: a GString -- caller is responsible for freeing it.
 **/
GString *nxp_name_space_get_full_path_name(NxpNameSpace *self)
{
    GString *path;
    g_return_val_if_fail(self, NULL);
    if (!self->parent) {
	path = g_string_new(NXP_IDENTIFIER_PATH_SEP);
	return path;
    } else
	path = nxp_name_space_get_full_path_name(self->parent);
    path = g_string_append(path, self->name? self->name : "#<anonymous>");
    path = g_string_append(path, NXP_IDENTIFIER_PATH_SEP);
    return path;
}



/**
 * nxp_name_space_lock:
 * @self: a #NxpNameSpace
 * 
 * Lock the name space; until unlocked, no other thread can lock it.
 **/
void nxp_name_space_lock(NxpNameSpace *self)
{
    g_return_if_fail(self);
    if (!self->reentrant)
	return;
    if (pthread_mutex_lock(&self->mutex) == EDEADLK) {
	if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) {
	    printf("NxpNameSpace deadlock!\n");
	    G_BREAKPOINT();
	}
    }
}


/**
 * nxp_name_space_unlock:
 * @self: a #NxpNameSpace
 * 
 * Unlock the name space.
 **/
void nxp_name_space_unlock(NxpNameSpace *self)
{
    g_return_if_fail(self);
    if (self->reentrant)
	pthread_mutex_unlock(&self->mutex);
}




/*
   -------------------------------------------------------------
   Definitions -- these are the objects held by NxpNameSpace's
   -------------------------------------------------------------
*/



/**
 * nxp_def_variable_new:
 * @name: name of variable. This function will take ownership of this string. Do not free or change it!
 * @value: an element containing the value for the variable
 * 
 * Create variable definition.
 * 
 * Return value: the new definition.
 **/
NxpDefVariable* nxp_def_variable_new(char *name, NxpElement *value)
{
    NxpDefVariable *def;

    def = g_new0(NxpDefVariable, 1);
    def->definition.type = NXP_DEF_VARIABLE;
    def->definition.name = name;
    def->value = NXP_ELEMENT_REF(value);
    def->definition.ref_count = 1;
    return def;
}

/* private -- do not use. */
static void inline _variable_definition_clean(NxpDefVariable *def)
{
    def->value = NXP_ELEMENT_UNREF(def->value);
}


/**
 * nxp_def_bi_func_new:
 * @name: name of definition. The function will take ownership of this string. Do not free or change it!
 * 
 * Create an empty (without prototypes) definition for a built-in function
 * 
 * Return value: the new definition.
 **/
NxpDefBIFunc* nxp_def_bi_func_new(char *name)
{
    NxpDefBIFunc *def;

    def = g_new0(NxpDefBIFunc, 1);
    def->definition.type = NXP_DEF_BI_FUNC;
    def->definition.name = name;
    def->definition.ref_count = 1;
    return def;
}


static void inline _bi_function_definition_clean(NxpDefBIFunc *def)
{
    g_slist_foreach(def->prototypes, (GFunc)nxp_proto_instance_free, NULL);
    g_slist_free(def->prototypes);
}


/**
 * nxp_def_user_func_new:
 * @name: name of definition. The function will take ownership of this string. Do not free or change it!
 * @expression: NxpElement tree containing body of the function.
 * 
 * Create a user defined function definition
 * 
 * Return value: the definition
 **/
NxpDefUserFunc* nxp_def_user_func_new(char *name, NxpElement *expression)
{
    NxpDefUserFunc *def;

    def = g_new0(NxpDefUserFunc, 1);
    def->definition.type = NXP_DEF_USER_FUNC;
    def->definition.name = name;
    def->argument_strings = g_string_chunk_new(16);
    def->argument_list = NULL;
    def->expression = NXP_ELEMENT_REF(expression);
    def->definition.ref_count = 1;
    return def;
}





/**
 * nxp_def_user_func_append_argument:
 * @def: A #NxpDefUserFunc
 * @arg: name of argument to append
 * 
 * Append an argument name to @def
 **/
void nxp_def_user_func_append_argument(NxpDefUserFunc *def, const char *arg)
{
    g_return_if_fail(def);
    def->argument_list = g_slist_append
	(def->argument_list, g_string_chunk_insert(def->argument_strings, arg));
}





/* private -- do not use. */
static void inline _user_function_definition_clean(NxpDefUserFunc *def)
{
    if (def->argument_strings)
	g_string_chunk_free(def->argument_strings);
    else g_warning("Destroying user function definition with NULL argument strings chunk");

    if (def->argument_list)
	g_slist_free(def->argument_list);
    else g_warning("Destroying user function definition with NULL argument list");

    if (def->expression)
	def->expression = NXP_ELEMENT_UNREF(def->expression);
    else g_warning("Destroying user function definition with NULL element");
}


/* private -- do not use. */
static void _definition_destroy(NxpDefinition *def)
{
#ifdef NXP_NAMESPACE_DEBUG
    if (def->parent_namespace) {
	fprintf(stderr, "WARNING: premature destruction of definition --"
		" it still belongs to a namespace!\n");
	if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS) G_BREAKPOINT();
    }
    def->parent_namespace = NULL;
#endif
    if (_numexp_debug_flags&NUMEXP_DEBUG_ELEMENTS)
	g_message("Destroying definition '%s'", def->name);
    switch (def->type)
    {
    case NXP_DEF_VARIABLE:
	_variable_definition_clean(NXP_DEF_VARIABLE(def));
	break;
    case NXP_DEF_BI_FUNC:
	_bi_function_definition_clean(NXP_DEF_BI_FUNC(def));
	break;
    case NXP_DEF_USER_FUNC:
	_user_function_definition_clean(NXP_DEF_USER_FUNC(def));
	break;
    default:
	g_assert_not_reached();
    }
    g_free(def->name);
    g_free(def);
}


/*
   -------------------------------------------------------------
       Listeners -- namespace change notifications
   -------------------------------------------------------------
*/


static GStaticMutex listeners_mutex =  G_STATIC_MUTEX_INIT;


/**
 * nxp_name_space_add_listener:
 * @self: namespace
 * @func: listener callback function
 * @data: user data to be supplied to function
 * 
 * Adds a listener to a namespace. The listener is a hook function and
 * data that gets called when changes in namespace @self or its
 * descendents have occurred.
 **/
void nxp_name_space_add_listener(NxpNameSpace *self, NxpNSListenerFunc  func, 
				 gpointer data, NxpEvalContext *context)
{
    NxpNSListener *listener;

    g_return_if_fail(self);
    g_return_if_fail(func);
    g_return_if_fail(context);

    listener = g_new0(NxpNSListener, 1);
    listener->func = func;
    listener->user_data = data;
    listener->context = context;

    nxp_name_space_lock(self);
    self->listeners = g_slist_prepend(self->listeners, listener);
    nxp_name_space_unlock(self);
    context->monitored_namespaces = 
	g_slist_prepend(context->monitored_namespaces, self);
}




static void ns_run_listeners(NxpNameSpace *self, NxpEvalContext *context)
{
    NxpNameSpace *ns;
    GSList *l, *m;
    NxpNSListener *listener;
    for (l = self->listeners; l; l = l->next) {
	listener = l->data;
	if (listener->context != context) continue;
	for (m = listener->nslist; m; m = m->next) {
	    ns = m->data;
	    listener->func(ns, context, listener->user_data);
	}
	g_slist_free(listener->nslist);
	listener->nslist = NULL;
    }
}


static GSList* ns_check_modified(NxpNameSpace *self)
{
    NxpNameSpace *child;
    GSList *change_list = NULL, *child_list;

    if (g_atomic_int_compare_and_exchange(&self->modified, 1, 0)) {
	/* namespace has been modified */
	change_list = g_slist_prepend(change_list, self);
    }
    /* recurse for every child, skipping anonymous namespaces */
    for (child = self->child; child && child->name; child = child->next) {
        nxp_name_space_lock(child);
	child_list = ns_check_modified(child);
        nxp_name_space_unlock(child);
	change_list = g_slist_concat(change_list, child_list);
    }
    /* If self_changed, place a copy of the change list in each
     * listener in this namespace */
    if (change_list) {
	GSList *l;
	NxpNSListener *lis;
	for (l = self->listeners; l; l = l->next) {
	    lis = l->data;
	    /* append copy of new list into listener */
	    lis->nslist = g_slist_concat(lis->nslist, g_slist_copy(change_list));
	    /* TODO: remove duplicates */
	}
    }
    return change_list;
}


/**
 * nxp_name_space_wakeup_listeners: -run listener notifications
 * @context: a #NxpEvalContext
 * 
 * Iterate through all namespaces being monitored (in context
 * @context) and invoke the apropriate listener functions for the ones
 * that have changed. This function should not be called too often,
 * otherwise it might cause a severe performance penalty.
 * 
 * Return value: always returns %FALSE, so that it can be registered as an idle handler.
 **/
gboolean nxp_name_space_wakeup_listeners(NxpEvalContext *context)
{
    GSList *l;
    NxpNameSpace *ns;
    g_return_val_if_fail(context, FALSE);
    g_return_val_if_fail(context->cwd, FALSE);

    g_static_mutex_lock(&listeners_mutex);
    for (l = context->monitored_namespaces; l; l = l->next) {
	ns = l->data;
	g_slist_free(ns_check_modified(ns));
	ns_run_listeners(ns, context);
    }

    g_static_mutex_unlock(&listeners_mutex);

    return FALSE;
}


