/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include "errors.h"
#include "i18n.h"
#include "nxpintpoly.h"
#include "nxpintpoly_mod.h"
#include "types/mpoly.h"

void
nxp_int_poly_view(NxpIntPoly *a)
{
    int i;
    char *str;
    g_print("NxpIntPoly : { ");
    if (a->deg < 0) {
	g_print ("NULL }\n");
	return;
    }
    for (i = 0; i <= a->deg; i++) {
	str = mpz_get_str(NULL, 10, a->coeff[i]);
	g_print("%s ", str);
	g_free(str);
    }
    g_print("}\ndeg = %d, lmax = %d\n", a->deg, a->lmax);
}

void
nxp_int_poly_init(NxpIntPoly *a, gint size)
{
    int i;
    g_assert(size >= 0);
    a->lmax = size;
    if (size > 0)
	a->coeff = g_new(mpz_t, size);
    else
	a->coeff = NULL;
    for (i = 0; i < a->lmax; i++)
	mpz_init(a->coeff[i]);
    a->deg = -1;
}

void
nxp_int_poly_realloc(NxpIntPoly *a, gint newsize)
{
    int i;
    gint oldsize = a->lmax;
    g_assert(newsize > oldsize);
    a->lmax = newsize;
    a->coeff = g_renew(mpz_t, a->coeff, newsize);
    for (i = oldsize; i < newsize; i++)
	mpz_init(a->coeff[i]);
}

void
nxp_int_poly_init_set(NxpIntPoly *a, NxpIntPoly *b)
{
    int i;
    int size = b->deg + 1;

    a->lmax = size;
    a->coeff = g_new(mpz_t, size);
    a->deg = b->deg;
    for (i = 0; i <= b->deg; i++)
	mpz_init_set(a->coeff[i], b->coeff[i]);
}

void
nxp_int_poly_init_set_mpoly(NxpIntPoly *aa, NxpElement *a, GError **error)
{
    GSList *mons;
    GError *err = NULL;
    int deg = -1, n, i;
    NxpElement *coeff, *mon;

    if (nxp_element_list_length(NXP_ELEMENT(NXP_ELEMENT_MPOLY(a)->args)) > 1) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_INVALID_ARGUMENT, _("I cannot convert a multivariate polynomial to a single variate polynomial."));
	return;
    }
    for (mons = NXP_ELEMENT_MPOLY(a)->mon->list; mons; mons = mons->next) {
	mon = mons->data;
	n = nxp_element_int_get_si(_element_vector_index(NXP_ELEMENT_VECTOR(NXP_MONOMIAL_VECTOR(mon)), 0), &err);
	if (err) {
	    g_propagate_error(error, err);
	    return;
	}
	if (n > deg)
	    deg = n;
    }

    nxp_int_poly_init(aa, deg + 1);
    for (i = 0; i <= deg; i++)
	mpz_set_ui(aa->coeff[i], 0);
    aa->deg = deg;

    for (mons = NXP_ELEMENT_MPOLY(a)->mon->list; mons; mons = mons->next) {
	coeff = NXP_MONOMIAL_COEFF(mons->data);
	n = nxp_element_int_get_si(_element_vector_index(NXP_ELEMENT_VECTOR(NXP_MONOMIAL_VECTOR(mons->data)), 0), &err);
	mpz_set(aa->coeff[n], NXP_ELEMENT_INT(coeff)->value);
    }
}

NxpElement*
nxp_element_mpoly_new_from_int_poly(NxpIntPoly *a, int size, NxpElement *vars)
{
    g_return_val_if_fail(a->deg >= size, NULL);
    int i, s;
    NxpElement *retval, *coeff, *mon, *vector, *list;

    s = (size < 0) ? a->deg : size - 1;
    while (s >= 0 && mpz_sgn(a->coeff[s]) == 0)
	s--;
    if (s < 0) {
	coeff = nxp_element_int_new(0);
	vector = nxp_element_vector_new(1, NULL);
	_element_vector_index(NXP_ELEMENT_VECTOR(vector), 0) = NXP_ELEMENT_REF(coeff);
	mon = nxp_monomial_new(coeff, vector);
	NXP_ELEMENT_UNREF(vector);
	NXP_ELEMENT_UNREF(coeff);
	retval = nxp_element_mpoly_new_simple(mon, vars);
	NXP_ELEMENT_UNREF(mon);
    }
    else {
	list = nxp_element_list_new();
	for (i = 0; i <= s; i++) {
	    mpz_t z;
	    mpz_init_set(z, a->coeff[i]);
	    coeff = nxp_element_int_new_from_mpz(z);
	    vector = nxp_element_vector_new(1, NULL);
	    _element_vector_index(NXP_ELEMENT_VECTOR(vector), 0) = nxp_element_int_new(i);
	    mon = nxp_monomial_new(coeff, vector);
	    NXP_ELEMENT_UNREF(vector);
	    NXP_ELEMENT_UNREF(coeff);
	    nxp_element_list_prepend(list, mon);
	    NXP_ELEMENT_UNREF(mon);
	}
	retval = nxp_element_mpoly_new(list, vars);
	NXP_ELEMENT_UNREF(list);
    }
    return retval;
}

void
nxp_int_poly_set(NxpIntPoly *r, NxpIntPoly *a)
{
    int i;
    if (r == a)
	return;
    if (r->lmax <= a->deg) {
	nxp_int_poly_realloc(r, a->deg + 1);
	r->lmax = a->deg + 1;
    }

    for (i = 0; i <= a->deg; i++)
	mpz_set(r->coeff[i], a->coeff[i]);
    r->deg = a->deg;
}

gboolean
nxp_int_poly_equal(NxpIntPoly *a, NxpIntPoly *b)
{
    int i;
    if (a->deg == b->deg) {
	i = 0;
	while (i <= a->deg && mpz_cmp(a->coeff[i], b->coeff[i]) == 0)
	    i++;
	if (i > a->deg)
	    return TRUE;
    }
    return FALSE;
}

int inline
nxp_int_poly_deg(NxpIntPoly *a)
{
    return a->deg;
}

void
nxp_int_poly_clear(NxpIntPoly *a)
{
    int i;
    for (i = 0; i < a->lmax; i++)
	mpz_clear(a->coeff[i]);
    g_free(a->coeff);
}

void
nxp_int_poly_derive(NxpIntPoly *r, NxpIntPoly *a)
{
    int i;
    if (r->lmax < a->deg)
	nxp_int_poly_realloc(r, a->deg);

    for (i = 1; i <= a->deg; i++)
	mpz_mul_ui(r->coeff[i-1], a->coeff[i], i);
    r->deg = a->deg - 1;
}

void
nxp_int_poly_add(NxpIntPoly *r, NxpIntPoly *a, NxpIntPoly *b)
{
    int deg, max, i;
    int dega, degb;
    mpz_t coeff;

    dega = a->deg;
    degb = b->deg;
    if (dega == degb) {
	if (dega == -1) {
	    r->deg = -1;
	    return;
	}
	mpz_init(coeff);
	deg = dega;
	mpz_add(coeff, a->coeff[deg], b->coeff[deg]);
	while (deg >= 1 && mpz_sgn(coeff) == 0) {
	    deg --;
	    mpz_add(coeff, a->coeff[deg], b->coeff[deg]);
	}
	if (mpz_sgn(coeff) == 0) {
	    mpz_clear(coeff);
	    r->deg = -1;
	    return;
	}
	if (r->lmax <= deg)
	    nxp_int_poly_realloc(r, deg + 1);
	r->deg = deg;
	mpz_set(r->coeff[deg], coeff);
	deg --;
	mpz_clear(coeff);
    }
    else {
	if (dega > degb) {
	    max = r->deg = dega;
	    deg = degb;
	}
	else {
	    max = r->deg = degb;
	    deg = dega;
	}
	if (r->lmax <= max)
	    nxp_int_poly_realloc(r, max + 1);
	if (r != a)
	    for (i = deg + 1; i <= dega; i++)
		mpz_set(r->coeff[i], a->coeff[i]);
	if (r != b)
	    for (i = dega + 1; i <= degb; i++)
		mpz_set(r->coeff[i], b->coeff[i]);
    }
    for (i = 0; i <= deg; i++)
	mpz_add(r->coeff[i], a->coeff[i], b->coeff[i]);
}

void
nxp_int_poly_sub(NxpIntPoly *r, NxpIntPoly *a, NxpIntPoly *b)
{
    int deg, max, dega, degb;
    int i;

    dega = a->deg;
    degb = b->deg;
    if (dega == degb) {
	deg = dega;
	while (deg >= 0 && mpz_cmp(a->coeff[deg], b->coeff[deg]) == 0)
	    deg --;
	if (deg < 0) {
	    r->deg = -1;
	    return;
	}
	if (r->lmax <= deg)
	    nxp_int_poly_realloc(r, deg + 1);
	r->deg = deg;
    }
    else {
	if (dega > degb) {
	    max = dega;
	    deg = degb;
	}
	else {
	    max = degb;
	    deg = dega;
	}
	if (r->lmax <= max)
	    nxp_int_poly_realloc(r, max + 1);
	r->deg = max;
	for (i = deg + 1; i <= dega; i++)
	    if (r->coeff != a->coeff)
		mpz_set(r->coeff[i], a->coeff[i]);
	for ( ; i <= degb; i++)
	    mpz_neg(r->coeff[i], b->coeff[i]);
    }
    for (i = 0; i <= deg; i++)
	mpz_sub(r->coeff[i], a->coeff[i], b->coeff[i]);
}

void
nxp_int_poly_mul(NxpIntPoly *ret, NxpIntPoly *p1, NxpIntPoly *p2)
{
    int size;
    int sa0, sb0, sa1, sb1, i;
    NxpIntPoly u[1], z[1];
    NxpIntPoly res_tmp[1];
    NxpIntPoly *retval;
    mpz_t *tmp, *tmp1;
    NxpIntPoly a0[1], b0[1], a1[1], b1[1];
    mpz_t c;

    if (p1->deg == -1 || p2->deg == -1) {
	ret->deg = -1;
	return;
    }
    else if (p1->deg == 0) {
	mpz_init_set(c, p1->coeff[0]);
	if (ret->lmax <= p2->deg)
	    nxp_int_poly_realloc(ret, p2->deg + 1);
	for (i = 0; i <= p2->deg; i++)
	    mpz_mul(ret->coeff[i], p2->coeff[i], c);
	ret->deg = p2->deg;
	mpz_clear(c);
	return;
    }
    else if (p2->deg == 0) {
	mpz_init_set(c, p2->coeff[0]);
	if (ret->lmax <= p1->deg)
	    nxp_int_poly_realloc(ret, p1->deg + 1);
	for (i = 0; i <= p1->deg; i++)
	    mpz_mul(ret->coeff[i], p1->coeff[i], c);
	ret->deg = p1->deg;
	mpz_clear(c);
	return;
    }
    else {
	if (p1->deg < p2->deg)
	    size = (p2->deg + 1) >> 1;
	else
	    size = (p1->deg + 1) >> 1;

	/* For a0 and a1 we don't create new polynomials, they are just links to
	 * some parts of p1 */
	sa0 = MIN(size - 1, p1->deg);
	while (sa0 >= 0 && mpz_sgn(p1->coeff[sa0]) == 0)
		sa0--;
	a0->deg = sa0;
	a0->coeff = p1->coeff;
	a0->lmax = sa0 + 1;

	if (p1->deg < size)
	    a1->deg = -1;
	else {
	    sa1 = p1->deg;
	    while (sa1 >= size && mpz_sgn(p1->coeff[sa1]) == 0)
		sa1--;
	    a1->deg = sa1 - size;
	    a1->coeff = p1->coeff + size;
	    a1->lmax = a1->deg + 1;
	}

	/* For b0 and b1 we don't create new polynomials, they are just links to
	 * some parts of p2 */
	sb0 = MIN(size - 1, p2->deg);
	while (sb0 >= 0 && mpz_sgn(p2->coeff[sb0]) == 0)
		sb0--;
	b0->deg = sb0;
	b0->coeff = p2->coeff;
	b0->lmax = sb0 + 1;

	if (p2->deg < size)
	    b1->deg = -1;
	else {
	    sb1 = p2->deg;
	    while (sb1 >= size && mpz_sgn(p2->coeff[sb1]) == 0)
		sb1--;
	    b1->deg = sb1 - size;
	    b1->coeff = p2->coeff + size;
	    b1->lmax = b1->deg + 1;
	}
	if (ret != p1 && ret != p2) {
	    if (ret->lmax <= p1->deg + p2->deg)
		nxp_int_poly_realloc(ret, p1->deg + p2->deg + 1);
	    retval = ret;
	}
	else {
	    nxp_int_poly_init(res_tmp, p1->deg + p2->deg + 1);
	    retval = res_tmp;
	}

	for (i = 0; i < retval->lmax; i++)
	    mpz_set_ui(retval->coeff[i], 0);

	nxp_int_poly_init(u, MAX(a0->deg + b0->deg, 0) + 1);
	nxp_int_poly_mul(u, a0, b0);

	tmp = retval->coeff + size;
	for (i = 0; i <= u->deg; i++) {
	    mpz_add(retval->coeff[i], retval->coeff[i], u->coeff[i]);
	    mpz_sub(tmp[i], tmp[i], u->coeff[i]);
	}
	nxp_int_poly_mul(u, a1, b1);
	tmp1 = retval->coeff + (size << 1);
	for (i = 0; i <= u->deg; i++) {
	    mpz_add(tmp1[i], tmp1[i], u->coeff[i]);
	    mpz_sub(tmp[i], tmp[i], u->coeff[i]);
	}

	nxp_int_poly_add(u, a0, a1);
	nxp_int_poly_init(z, MAX(b0->deg, b1->deg) + 1);
	nxp_int_poly_add(z, b0, b1);
	nxp_int_poly_mul(u, u, z);

	for (i = 0; i <= u->deg; i++)
	    mpz_add(tmp[i], tmp[i], u->coeff[i]);
	nxp_int_poly_clear(u);
	nxp_int_poly_clear(z);
	i = retval->lmax -1;
	while (i >= 0 && mpz_sgn(retval->coeff[i]) == 0)
	    i--;
	retval->deg = i;
	if (res_tmp == retval) {
	    nxp_int_poly_clear(ret);
	    *ret = *res_tmp;
	}
    }
}

void
nxp_int_poly_power(NxpIntPoly *res, NxpIntPoly *a, int pow)
{
    int k;
    NxpIntPoly tmp[1];

    k = a->deg * pow + 1;
    nxp_int_poly_init(tmp, k);
    nxp_int_poly_set(tmp, a);
    if (res->lmax < k)
	nxp_int_poly_realloc(res, k);
    if (pow & 1)
	nxp_int_poly_set(res, tmp);
    else {
	mpz_set_ui(res->coeff[0], 1);
	res->deg = 0;
    }
    pow >>= 1;

    while (pow) {
	nxp_int_poly_mul(tmp, tmp, tmp);
	if (pow & 1)
	    nxp_int_poly_mul(res, res, tmp);
	pow >>= 1;
    }
    nxp_int_poly_clear(tmp);
}

void
nxp_int_poly_pseudo_div_qr(NxpIntPoly *q, NxpIntPoly *r,
			   NxpIntPoly *u, NxpIntPoly *v)
{
    int k, j;
    int m = u->deg, n = v->deg;
    NxpIntPoly Q[1], R[1];
    mpz_t d, dd;

    g_return_if_fail(n != -1);

    nxp_int_poly_init_set(R, u);
    if (m < n) {
	q->deg = -1;
	nxp_int_poly_clear(r);
	r->coeff = R->coeff;
	r->deg = n;
	r->lmax = R->lmax;
	return;
    }
    else {
	mpz_init_set(d, v->coeff[n]);
	mpz_init(dd);
	nxp_int_poly_init(Q, m - n + 1);
	for (k = m - n; k >= 0; k--) {
	    mpz_pow_ui(dd, d, k);
	    mpz_mul(Q->coeff[k], R->coeff[n + k], dd);
	    for (j = n + k -1; j >= k; j--) {
		mpz_mul(R->coeff[j], v->coeff[n], R->coeff[j]);
		mpz_submul(R->coeff[j], R->coeff[n + k], v->coeff[j - k]);
	    }
	    for ( ; j >= 0; j--)
		mpz_mul(R->coeff[j], v->coeff[n], R->coeff[j]);
	}
	if (r) {
	    do
		n--;
	    while(n >= 0 && mpz_sgn(R->coeff[n]) == 0);

	    nxp_int_poly_clear(r);
	    r->coeff = R->coeff;
	    r->deg = n;
	    r->lmax = R->lmax;
	}
	else
	    nxp_int_poly_clear(R);

	nxp_int_poly_clear(q);
	q->coeff = Q->coeff;
	q->deg = m - n;
	q->lmax = Q->lmax;
	mpz_clear(dd);
	mpz_clear(d);
    }
}

void
nxp_int_poly_pseudo_div_r(NxpIntPoly *r, NxpIntPoly *u, NxpIntPoly *v)
{
    int k, j;
    int m = u->deg, n = v->deg;

    g_return_if_fail(n != -1);

    nxp_int_poly_set(r, u);
    if (m < n)
	return;
    else {
	for (k = m - n; k >= 0; k--) {
	    for (j = n + k -1; j >= k; j--) {
		mpz_mul(r->coeff[j], v->coeff[n], r->coeff[j]);
		mpz_submul(r->coeff[j], r->coeff[n + k], v->coeff[j - k]);
	    }
	    for ( ; j >= 0; j--)
		mpz_mul(r->coeff[j], v->coeff[n], r->coeff[j]);
	}
	do
	    n--;
	while(n >= 0 && mpz_sgn(r->coeff[n]) == 0);

	r->deg = n;
    }
}

gboolean
nxp_int_poly_divisible(NxpIntPoly *a, NxpIntPoly *b)
{
    NxpIntPoly r[1];
    gboolean test;

    nxp_int_poly_init(r, MAX(a->deg - b->deg, 0) + 1);
    test = nxp_int_poly_divexact(r, a, b);
    nxp_int_poly_clear(r);
    return test;
}

void
nxp_int_poly_cont(mpz_t cont, NxpIntPoly *p)
{
    int i;
    int sgn;

    if (p->deg < 0) {
	mpz_set_ui(cont, 0);
	return;
    }
    else {
	mpz_set(cont, p->coeff[p->deg]);
	if (mpz_sgn(cont) < 0)
	    sgn = -1;
	else
	    sgn = 1;
	for (i = p->deg - 1; i >= 0; i--)
	    if (mpz_sgn(p->coeff[i]))
		mpz_gcd(cont, cont, p->coeff[i]);
    }
    if (sgn < 0)
	mpz_neg(cont, cont);
}

/* Compute the gcd of the two given polynomials, those last ones must have
 * integer coefficients-> */
void
nxp_int_poly_gcd(NxpIntPoly *d, NxpIntPoly *a, NxpIntPoly *b)
{
    NxpIntPoly n[1], r[1];
    mpz_t aa, g, h, c, dd, bb;
    int i, delta;

    if (a->deg < 0) {
	nxp_int_poly_set(d, b);
	return;
    }
    if (b->deg < 0) {
	nxp_int_poly_set(d, a);
	return;
    }
    if (a->deg > b->deg) {
	nxp_int_poly_init(n, a->deg + 1);
	nxp_int_poly_set(n, a);
	nxp_int_poly_set(d, b);
    }
    else {
	nxp_int_poly_init(n, b->deg + 1);
	nxp_int_poly_set(n, b);
	nxp_int_poly_set(d, a);
    }

    mpz_init(aa);
    nxp_int_poly_cont(aa, n);
    mpz_init(bb);
    nxp_int_poly_cont(bb, d);
    mpz_init(dd);
    mpz_gcd(dd, aa, bb);
    for (i = 0; i <= n->deg; i++)
	mpz_divexact(n->coeff[i], n->coeff[i], aa);

    for (i = 0; i <= d->deg; i++)
	mpz_divexact(d->coeff[i], d->coeff[i], bb);

    mpz_init_set_ui(g, 1);
    mpz_init_set_ui(h, 1);
    mpz_init(c);
    delta = n->deg - d->deg;
    nxp_int_poly_init(r, d->deg);
    do {
	nxp_int_poly_pseudo_div_r(r, n, d);
	if (r->deg > 0) {
	    nxp_int_poly_set(n, d);

	    mpz_pow_ui(c, h, delta);
	    mpz_mul(c, c, g);

	    for (i = 0; i <= r->deg; i++)
		mpz_divexact(d->coeff[i], r->coeff[i], c);
	    d->deg = r->deg;
	    mpz_set(g, n->coeff[n->deg]);
	    mpz_pow_ui(c, g, delta);
	    if (delta > 1) {
		mpz_pow_ui(h, h, delta - 1);
		mpz_divexact(h, c, h);
	    }
	    else {
		mpz_pow_ui(h, h, 1 - delta);
		mpz_mul(h, h, c);
	    }
	}
	else {
	    if (r->deg == 0) {
		mpz_set_ui(d->coeff[0], 1);
		d->deg = 0;
	    }
	    break;
	}
	delta = n->deg - d->deg;
    }
    while(TRUE);
    nxp_int_poly_cont(bb, d);

    for (i = 0; i <= d->deg; i++) {
	mpz_mul(d->coeff[i], d->coeff[i], dd);
	mpz_divexact(d->coeff[i], d->coeff[i], bb);
    }
    nxp_int_poly_clear(n);
    nxp_int_poly_clear(r);
    mpz_clear(aa);
    mpz_clear(bb);
    mpz_clear(dd);
    mpz_clear(g);
    mpz_clear(h);
    mpz_clear(c);
}

GSList *
nxp_int_poly_sff(NxpIntPoly *p)
{
    GSList *list = NULL;
    NxpIntPoly w[1], y[1], g[1];
    NxpIntPoly *tmp;

    nxp_int_poly_init_set(w, p);
    nxp_int_poly_init(y, w->deg);
    nxp_int_poly_init(g, w->deg);
    nxp_int_poly_derive(y, w);
    nxp_int_poly_ezgcd(g, w, y);
    do {
	nxp_int_poly_divexact(w, w, g);
	nxp_int_poly_divexact(y, y, g);
	nxp_int_poly_derive(g, w);
	nxp_int_poly_sub(y, y, g);
	nxp_int_poly_ezgcd(g, w, y);

	tmp = g_new(NxpIntPoly, 1);
	nxp_int_poly_init_set(tmp, g);
	list = g_slist_append(list, tmp);
    }
    while(y->deg >= 0);
    nxp_int_poly_clear(w);
    nxp_int_poly_clear(y);
    nxp_int_poly_clear(g);
    return list;
}

void inline
nxp_int_poly_divexact_mpz(NxpIntPoly *res, NxpIntPoly *a, mpz_t b)
{
    int i;

    if (res->lmax <= a->deg)
	nxp_int_poly_realloc(res, a->deg + 1);
    for (i = 0; i <= a->deg; i++)
	mpz_divexact(res->coeff[i], a->coeff[i], b);
    res->deg = a->deg;
}

void inline
nxp_int_poly_mul_si(NxpIntPoly *res, NxpIntPoly *a, nxp_int_t b)
{
    int i;

    if (res->lmax <= a->deg)
	nxp_int_poly_realloc(res, a->deg + 1);
    for (i = 0; i <= a->deg; i++)
	mpz_mul_si(res->coeff[i], a->coeff[i], b);
    res->deg = a->deg;
}

void
nxp_int_poly_sub_si(NxpIntPoly *res, NxpIntPoly *a, nxp_int_t b)
{
    nxp_int_poly_set(res, a);
    if (b != 0) {
	if (res->deg >= 0) {
	    if (b > 0)
		mpz_sub_ui(res->coeff[0], res->coeff[0], b);
	    else
		mpz_add_ui(res->coeff[0], res->coeff[0], -b);

	    if (res->deg == 0 && mpz_sgn(res->coeff[0]) == 0)
		res->deg = -1;
	}
	else {
	    if (res->lmax < 1)
		nxp_int_poly_realloc(res, 1);
	    mpz_set_si(res->coeff[0], -b);
	    res->deg = 0;
	}
    }
}

void inline
nxp_int_poly_mul_mpz(NxpIntPoly *res, NxpIntPoly *a, mpz_t b)
{
    int i;

    if (res->lmax <= a->deg)
	nxp_int_poly_realloc(res, a->deg + 1);
    for (i = 0; i <= a->deg; i++)
	mpz_mul(res->coeff[i], a->coeff[i], b);
    res->deg = a->deg;
}

void
nxp_int_poly_mignotte(mpz_t bound, NxpIntPoly *a, int n)
{
    int d, i;
    mpz_t na, am, aa, bb, v;

    d = a->deg;

    /* First, we compute the norm of a */
    mpz_init_set_ui(na, 0);
    for (i = 0; i <= d; i++)
	mpz_addmul(na, a->coeff[i], a->coeff[i]);
    mpz_sqrt(na, na);

    mpz_init_set(am, a->coeff[d]);
    if (mpz_sgn(am) < 0)
	mpz_neg(am, am);

    mpz_init_set_ui(aa, 1);
    mpz_init_set_ui(bb, 0);
    mpz_init(v);
    mpz_set_ui(bound, 0);
    for (i = 0; i < n; i++) {
	mpz_mul(v, aa, na);
	mpz_addmul(v, bb, am);
	if (mpz_cmp(v, bound) > 0)
	    mpz_set(bound, v);
	mpz_set(bb, aa);
	mpz_mul_ui(aa, bb, n - i - 1);
	mpz_tdiv_q_ui(aa, aa, i + 1);
    }
    mpz_mul(v, aa, na);
    mpz_addmul(v, bb, am);
    if (mpz_cmp(v, bound) > 0)
	mpz_set(bound, v);
    mpz_clear(na);
    mpz_clear(am);
    mpz_clear(aa);
    mpz_clear(bb);
    mpz_clear(v);
}

/* Divides a by b and puts the quotient in q. If the division fails returns FALSE else returns TRUE. It's a good algorithm (I hope so :-) ) to test if b divides a because in each step we have a test to know if b divides a. */
gboolean
nxp_int_poly_divexact(NxpIntPoly *q, NxpIntPoly *a, NxpIntPoly *b)
{
    NxpIntPoly quo[1], rem[1], prod[1];
    int pitcha, pitchb, i, pitch;

    /* rem is not the remainder of an euclidean division, it's the remainder
     * of our division. */

    pitchb = 0;
    while(mpz_sgn(b->coeff[pitchb]) == 0)
	pitchb ++;

    nxp_int_poly_init_set(rem, a);

    pitcha = 0;
    while(mpz_sgn(rem->coeff[pitcha]) == 0)
	pitcha ++;

    pitch = pitcha - pitchb;
    if (pitch >= 0 && (rem->deg - pitcha) >= (b->deg - pitchb)) {
	nxp_int_poly_init(quo, rem->deg - b->deg + 1);
	nxp_int_poly_init(prod, rem->deg + 1);
	q->deg = -1;
	if (rem->deg >= q->lmax)
	    nxp_int_poly_realloc(q, rem->deg + 1);
	do {
	    if (!mpz_divisible_p(rem->coeff[pitcha], b->coeff[pitchb]) || !mpz_divisible_p(rem->coeff[rem->deg], b->coeff[b->deg]))
		break;
	    quo->deg = rem->deg - b->deg;
	    for (i = 0; i < quo->deg; i++)
		mpz_set_ui(quo->coeff[i], 0);
	    if (quo->deg != 0)
		mpz_divexact(quo->coeff[quo->deg], rem->coeff[rem->deg], b->coeff[b->deg]);
	    mpz_divexact(quo->coeff[pitch], rem->coeff[pitcha], b->coeff[pitchb]);
	    nxp_int_poly_mul(prod, quo, b);
	    nxp_int_poly_sub(rem, rem, prod);

	    nxp_int_poly_add(q, q, quo);
	    if (rem->deg < 0) {
		nxp_int_poly_clear(quo);
		nxp_int_poly_clear(prod);
		nxp_int_poly_clear(rem);
		return TRUE;
	    }
	    /* We know that rem->coeff[pitcha] is null */
	    pitcha ++;
	    while(pitcha <= rem->deg && mpz_sgn(rem->coeff[pitcha]) == 0)
		pitcha ++;
	    pitch = pitcha - pitchb;
	}
	while(pitch >= 0 && (rem->deg - pitcha) >= (b->deg - pitchb));
	nxp_int_poly_clear(quo);
	nxp_int_poly_clear(prod);
	nxp_int_poly_clear(rem);
    }
    return FALSE;
}

/* pol is the product of the lift's elements modulo p that is suffisantly big to have factors near of those we could have in Z. But those factors are too numerous. We must group some of them... */
GSList*
nxp_int_poly_build_factors(NxpIntPoly *pol, GArray *lift, mpz_t p)
{
    mpz_t lcpol, lc, p1;
    NxpIntPoly D[1], *index, *index1, tmp[1], q[1];
    gboolean test;
    GSList *retval = NULL;
    int *v, size, i, j, l, d, k, n;

    mpz_init_set(lcpol, pol->coeff[pol->deg]);
    mpz_init(lc);
    size = lift->len;
    nxp_int_poly_init(D, pol->deg + 1);
    nxp_int_poly_init(q, D->deg + 1);
    nxp_int_poly_init(tmp, pol->deg + 1);
    nxp_int_poly_mul_mpz(D, pol, lcpol);

    n = size;
    for (i = 0; i < size; i++) {
	index = &g_array_index(lift, NxpIntPoly, i);
	if (mpz_cmp_ui(index->coeff[index->deg], 1) != 0) {
	    mpz_invert(lc,  index->coeff[index->deg], p);
	    nxp_int_poly_mul_mpz(index, index, lc);
	    nxp_int_poly_normalize(index, index, p);
	}
	if (mpz_cmp(lcpol, index->coeff[index->deg]) != 0) {
	    nxp_int_poly_mul_mpz(tmp, index, lcpol);
	    nxp_int_poly_normalize(tmp, tmp, p);
	}
	else
	    nxp_int_poly_set(tmp, index);

	test = nxp_int_poly_divexact(q, D, tmp);
	if (test) {
	    nxp_int_poly_cont(lc, tmp);
	    nxp_int_poly_divexact_mpz(tmp, tmp, lc);
	    nxp_int_poly_mul_mpz(q, q, lc);
	    nxp_int_poly_divexact_mpz(pol, q, lcpol);
	    mpz_set(lcpol, pol->coeff[pol->deg]);
	    nxp_int_poly_mul_mpz(D, pol, lcpol);
	    index1 = g_new(NxpIntPoly, 1);
	    nxp_int_poly_init_set(index1, tmp);
	    index->deg = -1;
	    n--;
	    retval = g_slist_prepend(retval, index1);
	}
    }

    if (n) {
	mpz_init_set(p1, p);
	mpz_tdiv_q_2exp(p1, p1, 1);
	k = 2;
	v = g_new(int, k);

	j = 0;
	for (i = 0; i < k; i++) {
	    while (g_array_index(lift, NxpIntPoly, j).deg < 0)
		j++;
	    v[i] = j++;
	}
	while (n) {
	    /* First we test degrees. */
	    d = g_array_index(lift, NxpIntPoly, v[0]).deg;
	    for (i = 1; d < (D->deg >> 1) && i < k; i++)
		d += g_array_index(lift, NxpIntPoly, v[i]).deg;

	    if (d > (D->deg >> 1))
		test = FALSE;
	    else {
		/* Second, we test constant coefficient */
		mpz_set(lc, lcpol);
		for (i = 0; i < k; i++) {
		    mpz_mul(lc, lc, g_array_index(lift, NxpIntPoly, v[i]).coeff[0]);
		    mpz_mod(lc, lc, p);
		}
		if (mpz_cmp(lc, p1) > 0)
		    mpz_sub(lc, lc, p);
		if (!mpz_divisible_p(D->coeff[0], lc))
		    test = FALSE;
		else {
		    nxp_int_poly_set(tmp, &g_array_index(lift, NxpIntPoly, v[0]));
		    for (i = 1; i < k; i++) {
			nxp_int_poly_mul(tmp, tmp, &g_array_index(lift, NxpIntPoly, v[i]));
			nxp_int_poly_normalize(tmp, tmp, p);
		    }
		    if (mpz_cmp(lcpol, tmp->coeff[tmp->deg]) != 0) {
			nxp_int_poly_mul_mpz(tmp, tmp, lcpol);
			nxp_int_poly_normalize(tmp, tmp, p);
		    }
		    test = nxp_int_poly_divexact(q, D, tmp);
		    if (test) {
			nxp_int_poly_cont(lc, tmp);
			nxp_int_poly_divexact_mpz(tmp, tmp, lc);
			nxp_int_poly_mul_mpz(q, q, lc);
			nxp_int_poly_divexact_mpz(pol, q, lcpol);
			mpz_set(lcpol, pol->coeff[pol->deg]);
			nxp_int_poly_mul_mpz(D, pol, lcpol);
			index = g_new(NxpIntPoly, 1);
			nxp_int_poly_init_set(index, tmp);
			retval = g_slist_prepend(retval, index);
			for (i = k - 1; i >= 0; i--) {
			    nxp_int_poly_clear(&g_array_index(lift, NxpIntPoly, v[i]));
			    g_array_index(lift, NxpIntPoly, v[i]).deg = -1;
			    n --;
			}

			j = 0;
			if (k > n)
			    break;
			for (i = 0; i < k; i++) {
			    while (g_array_index(lift, NxpIntPoly, j).deg < 0)
				j++;
			    v[i] = j++;
			}
		    }
		}
	    }

	    if (!test) {
		j = k - 1;
		while(j >= 0) {
		    do
			v[j] ++;
		    while (v[j] < size && g_array_index(lift, NxpIntPoly, v[j]).deg < 0);
		    if (v[j] == size) {
			j --;
			continue;
		    }
		    else if (j < k - 1) {
			l = j + 1;
			for (i = v[j] + 1; i < size && l < k; i++) {
			    if (g_array_index(lift, NxpIntPoly, i).deg >= 0)
				v[l++] = i;
			}
			if (l < k) {
			    j --;
			    continue;
			}
		    }
		    break;
		}
		if (j < 0) {
		    k ++;
		    if (k > n)
			break;
		    g_free(v);
		    v = g_new(int, k);

		    j = 0;
		    for (i = 0; i < k; i++) {
			while (g_array_index(lift, NxpIntPoly, j).deg < 0)
			    j++;
			v[i] = j;
		    }
		}
	    }
	}
	g_free(v);
	mpz_clear(p1);
    }
    if (n) {
	index = g_new(NxpIntPoly, 1);
	nxp_int_poly_init_set(index, pol);
	retval = g_slist_prepend(retval, index);
    }
    mpz_clear(lcpol);
    mpz_clear(lc);
    nxp_int_poly_clear(D);
    nxp_int_poly_clear(q);
    nxp_int_poly_clear(tmp);
    return retval;
}

void
nxp_int_poly_ezgcd(NxpIntPoly *gcd, NxpIntPoly *f, NxpIntPoly *g)
{
    mpz_t cf, cg, cd, p, p1, q;
    NxpIntPoly ppf[1], ppg[1], ppf_mod[1], ppg_mod[1], gcd_mod[1], fp[1], gp[1], dp[1];
    NxpIntPoly *c;
    int d, st, d1, e;
    char *str;

    g_assert(f->deg >= g->deg);
    if (f->deg < 0) {
	nxp_int_poly_set(gcd, g);
	return;
    }
    if (g->deg < 0) {
	nxp_int_poly_set(gcd, f);
	return;
    }

    mpz_init(cf);
    nxp_int_poly_cont(cf, f);
    nxp_int_poly_init(ppf, f->deg + 1);
    nxp_int_poly_divexact_mpz(ppf, f, cf);
    mpz_init(cg);
    nxp_int_poly_cont(cg, g);
    nxp_int_poly_init(ppg, g->deg + 1);
    nxp_int_poly_divexact_mpz(ppg, g, cg);
    mpz_init(cd);
    mpz_gcd(cd, cf, cg);
    nxp_int_poly_init(ppf_mod, ppf->deg + 1);
    nxp_int_poly_init(ppg_mod, ppg->deg + 1);
    nxp_int_poly_init(dp, MIN(ppf->deg, ppg->deg) + 1);
    nxp_int_poly_init(gcd_mod, MIN(ppf->deg, ppg->deg) + 1);
    nxp_int_poly_init(fp, ppf->deg + 1);
    nxp_int_poly_init(gp, ppg->deg + 1);
    mpz_init_set_ui(p, 100);
    mpz_init2(p1, mpz_sizeinbase(p, 2));
    mpz_init(q);
    d=-2;
    do {
	mpz_set(p1, p);
	for (st = 0; st <= 1; st++) {
	    do
		mpz_nextprime(p1, p1);
	    while (mpz_divisible_p(ppf->coeff[ppf->deg], p1) || mpz_divisible_p(ppg->coeff[ppg->deg], p1));

	    nxp_int_poly_normalize(ppf_mod, ppf, p1);
	    nxp_int_poly_normalize(ppg_mod, ppg, p1);
	    nxp_int_poly_gcd_mod(gcd_mod, ppf_mod, ppg_mod, p1);

	    d1 = gcd_mod->deg;
	    if (d1 == 0) {
		/* f and g are coprime !! */
		gcd->deg = 0;
		mpz_set(gcd->coeff[0], cd);
		goto end;
	    }
	    if (d1 > d && d != -2) {
		g_print("We already have seen a better prime\n");
		st--;
	    }
	    else if (d1 < d || d == -2) {
		/* The smaller deg(gcd_mod) is and the nearer we are of the gcd
		 * in Z[X] */
		d = d1;
		mpz_set(p, p1);
		nxp_int_poly_set(fp, ppf_mod);
		nxp_int_poly_set(gp, ppg_mod);
		nxp_int_poly_set(dp, gcd_mod);
	    }
	}

	/* Here, dp = gcd(fp, gp) mod p and d=deg(dp) */

	if (d == gp->deg) {
	    /* Here, if p is a good prime, we know that ppg is a multiple of the
	     * gcd dp. So we test the divisibility... */
	    /* We assume at the beginning of the function that deg(g)<=deg(f),
	     * that's why we don't test if d = fp->deg */
	    if (nxp_int_poly_divexact(dp, ppf, ppg)) {
		nxp_int_poly_mul_mpz(gcd, ppg, cd);
		goto end;
	    }
	    str = mpz_get_str(NULL, 10, p);
	    g_warning("%s is an unlucky prime...\n", str);
	    g_free(str);
	}
	else {
	    nxp_int_poly_div_q_mod(ppf_mod, gp, dp, p);
	    nxp_int_poly_gcd_mod(ppg_mod, dp, ppf_mod, p);
	    if (ppg_mod->deg == 0 && mpz_cmp_ui(ppg_mod->coeff[0], 1) == 0)
		c = ppg;
	    else {
		nxp_int_poly_div_q_mod(ppf_mod, fp, dp, p);
		nxp_int_poly_gcd_mod(ppg_mod, ppf_mod, dp, p);
		if (ppg_mod->deg == 0 && mpz_cmp_ui(ppg_mod->coeff[0], 1) == 0)
		    c = ppf;
		else {
		    /* Here, gp/dp nor fp/dp are coprime with dp
		     * (where dp=gcd(fp, gp) mod p ). But we use this gcd
		     * computation only in the sff algorithm where this case
		     * shouldn't occur. */
		    g_warning ("I'm not ready to compute such a gcd...\n");
		    g_assert_not_reached();
		    goto end;
		}
	    }

	    nxp_int_poly_mignotte(q, c, c->deg >> 1);
	    mpz_mul(q, q, c->coeff[c->deg]);
	    mpz_mul_2exp(q, q, 1);

	    mpz_set(p1, p);
	    e = 1;
	    while(mpz_cmp(p1, q) <= 0) {
		mpz_mul(p1, p1, p);
		e ++;
	    } /* p1 is p^e */
	    nxp_int_poly_hensel_quad(dp, ppf_mod, c, dp, ppf_mod, p, e);

	    mpz_invert(q, dp->coeff[dp->deg], p1);
	    mpz_mul(q, q, c->coeff[c->deg]);
	    mpz_mod(q, q, p1);
	    nxp_int_poly_mul_mpz(dp, dp, q);
	    nxp_int_poly_normalize(dp, dp, p1);
	    nxp_int_poly_cont(q, dp);
	    nxp_int_poly_divexact_mpz(dp, dp, q);

	    mpz_invert(q, ppf_mod->coeff[ppf_mod->deg], p1);
	    mpz_mul(q, q, c->coeff[c->deg]);
	    mpz_mod(q, q, p1);
	    nxp_int_poly_mul_mpz(ppf_mod, ppf_mod, q);
	    nxp_int_poly_normalize(ppf_mod, ppf_mod, p1);
	    nxp_int_poly_cont(q, ppf_mod);
	    nxp_int_poly_divexact_mpz(ppf_mod, ppf_mod, q);

	    nxp_int_poly_mul(gcd, dp, ppf_mod);
	    if (mpz_cmp(gcd->coeff[gcd->deg], c->coeff[c->deg]) == 0)
		nxp_int_poly_sub(gcd, gcd, c);
	    else
		nxp_int_poly_add(gcd, gcd, c);

	    if (gcd->deg < 0) {
		if ((c == ppg && nxp_int_poly_divisible(ppf, dp)) || (c == ppf && nxp_int_poly_divisible(ppg, dp))) {
		    nxp_int_poly_mul_mpz(gcd, dp, cd);
		    goto end;
		}
	    }
	}
    }
    while(TRUE);
   end:
    nxp_int_poly_clear(ppf);
    nxp_int_poly_clear(ppg);
    nxp_int_poly_clear(ppf_mod);
    nxp_int_poly_clear(ppg_mod);
    nxp_int_poly_clear(dp);
    nxp_int_poly_clear(gcd_mod);
    nxp_int_poly_clear(fp);
    nxp_int_poly_clear(gp);
    mpz_clear(p);
    mpz_clear(p1);
    mpz_clear(q);
    mpz_clear(cd);
    mpz_clear(cg);
    mpz_clear(cf);
}

