/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include "nxpintpoly.h"

/* This function put all the a's coefficients into -n/2 and n/2, here we work
 * modulo n-> This function works everytimes but should be only used after a
 * multiplication or a division-> See the next one for other operations->*/
void
nxp_int_poly_normalize(NxpIntPoly *r, NxpIntPoly *a, mpz_t n)
{
    int i;
    mpz_t p2, mp2;

    mpz_init(p2);
    mpz_init(mp2);
    mpz_tdiv_q_2exp(p2, n, 1);
    mpz_neg(mp2, p2);

    /* The reallocation of r is not optimal since a mod n can be null and then
     * we reserve an array of (a->deg + 1) mpz_t for nothing->->-> */
    if (r->lmax <= a->deg)
	nxp_int_poly_realloc(r, a->deg + 1);
    for (i = 0; i <= a->deg; i++) {
	mpz_mod(r->coeff[i], a->coeff[i], n);
	if (mpz_cmp(r->coeff[i], p2) > 0)
	    mpz_sub(r->coeff[i], r->coeff[i], n);
	else if (mpz_cmp(r->coeff[i], mp2) < 0)
	    mpz_add(r->coeff[i], r->coeff[i], n);
    }
    i = a->deg;
    while(i >= 0 && mpz_sgn(r->coeff[i]) == 0)
	i --;
    r->deg = i;
    mpz_clear(p2);
    mpz_clear(mp2);
}

void
nxp_int_poly_normalize_ui(NxpIntPoly *r, NxpIntPoly *a, guint n)
{
    int i;
    nxp_int_t p2, mp2;

    p2 = n >> 1;
    mp2 = -p2;

    /* The reallocation of r is not optimal since a mod n can be null and then
     * we reserve an array of (a->deg + 1) mpz_t or nothing->->-> */
    if (r->lmax <= a->deg)
	nxp_int_poly_realloc(r, a->deg + 1);
    for (i = 0; i <= a->deg; i++) {
	mpz_mod_ui(r->coeff[i], a->coeff[i], n);
	if (mpz_cmp_si(r->coeff[i], p2) > 0)
	    mpz_sub_ui(r->coeff[i], r->coeff[i], n);
	else if (mpz_cmp_si(r->coeff[i], mp2) <= 0)
	    mpz_add_ui(r->coeff[i], r->coeff[i], n);
    }
    i = a->deg;
    while(i >= 0 && mpz_sgn(r->coeff[i]) == 0)
	i --;
    r->deg = i;
}

/* This function put all the a's coefficients into -n/2 and n/2, here we work
 * modulo n-> This function only works after an addition or a substraction of
 * polynomials and only if those ones were already normalized-> But it is much
 * faster than nxp_int_poly_normalize->*/
void
nxp_int_poly_normalize1(NxpIntPoly *r, NxpIntPoly *a, mpz_t n)
{
    int i;
    mpz_t p2, mp2;

    mpz_init(p2);
    mpz_init(mp2);
    mpz_tdiv_q_2exp(p2, n, 1);
    mpz_neg(mp2, p2);

    /* The reallocation of r is not optimal since a mod n can be null and then
     * we reserve an array of (a->deg + 1) mpz_t or nothing->->-> */
    if (r->lmax <= a->deg)
	nxp_int_poly_realloc(r, a->deg + 1);
    for (i = 0; i <= a->deg; i++) {
	if (mpz_cmp(r->coeff[i], p2) > 0)
	    mpz_sub(r->coeff[i], r->coeff[i], n);
	else if (mpz_cmp(r->coeff[i], mp2) < 0)
	    mpz_add(r->coeff[i], r->coeff[i], n);
    }
    i = a->deg;
    while(i >= 0 && mpz_sgn(r->coeff[i]) == 0)
	i --;
    r->deg = i;
    mpz_clear(p2);
    mpz_clear(mp2);
}

/* Returns a^n mod p */
void
nxp_int_poly_power_mod(NxpIntPoly *ret, NxpIntPoly *a, nxp_int_t n, nxp_int_t p)
{
    int i, j, k, s, r, q;
    NxpIntPoly b[1], tmp[1];

    if (a->deg < 0) {
	if (n != 0)
	    ret->deg = -1;
	else {
	    if (ret->lmax < 1)
		nxp_int_poly_realloc(ret, 1);
	    mpz_set_ui(ret->coeff[0], 1);
	    ret->deg = 0;
	}
	return;
    }

    q = n / p;
    r = n % p;

    k = n * a->deg + 1;
    nxp_int_poly_init(tmp, k);
    nxp_int_poly_set(tmp, a);
    nxp_int_poly_init(b, a->deg * r + 1);
    if (r & 1)
	nxp_int_poly_set(b, tmp);
    else {
	mpz_set_ui(b->coeff[0], 1);
	b->deg = 0;
    }

    while(r) {
	nxp_int_poly_mul(tmp, tmp, tmp);
	nxp_int_poly_normalize_ui(tmp, tmp, p);
	r >>= 1;
	if (r & 1) {
	    nxp_int_poly_mul(b, b, tmp);
	    nxp_int_poly_normalize_ui(b, b, p);
	}
    }

    if (q != 0) {
	if (q != 1)
	    nxp_int_poly_power_mod(tmp, a, q, p);
	else
	    nxp_int_poly_set(tmp, a);
	s = tmp->deg * p;
	for (i = s, j = s, k = tmp->deg; i >= 0; i--) {
	    if (i == j) {
		mpz_set(tmp->coeff[i], tmp->coeff[k]);
		j -= p;
		k --;
	    }
	    else
		mpz_set_ui(tmp->coeff[i], 0);
	}
	tmp->deg = s;
	if (b->deg == 0)
	    g_assert(mpz_sgn(b->coeff[b->deg]) != 0);

	nxp_int_poly_mul(ret, tmp, b);
	nxp_int_poly_normalize_ui(ret, ret, p);
    }
    else
	nxp_int_poly_set(ret, b);
    nxp_int_poly_clear(b);
    nxp_int_poly_clear(tmp);
}

/* compute @q and @r in Z/pZ such that u=q*v + r and deg r < deg v.
 * p is not necessarily a prime number, we just need the leading
 * coefficient of v to be inversible in Z/pZ.
 */
void
nxp_int_poly_div_qr_mod(NxpIntPoly *q, NxpIntPoly *r,
			NxpIntPoly *u, NxpIntPoly *v, mpz_t p)
{
    int m = u->deg, n = v->deg, k, j;
    NxpIntPoly Q[1], R[1];
    NxpIntPoly *q1, *r1;
    mpz_t inv;

    if (m < n) {
	q->deg = -1;
	nxp_int_poly_set(r, u);
	return;
    }
    else {
	if (q == u || q == v) {
	    nxp_int_poly_init(Q, m - n + 1);
	    q1 = Q;
	}
	else {
	    if (m - n >= q->lmax)
		nxp_int_poly_realloc(q, m - n + 1);
	    q1 = q;
	}

	if (r != v) {
	    nxp_int_poly_set(r, u);
	    r1 = r;
	}
	else {
	    nxp_int_poly_init(R, m + 1);
	    nxp_int_poly_set(R, u);
	    r1 = R;
	}

	mpz_init(inv);
	q1->deg = m - n;
	r1->deg = n - 1;
	if (mpz_invert(inv, v->coeff[n], p)) {
	    for (k = m - n; k >= 0; k--) {
		mpz_mul(q1->coeff[k], r1->coeff[n + k], inv);
		for (j = n + k -1; j >= k; j--)
		    mpz_submul(r1->coeff[j], q1->coeff[k], v->coeff[j - k]);
	    }
	}
	else {
	    g_warning("The leading coefficient of the polynomial v has no inverse.\n");
	    g_assert_not_reached();
	}
	mpz_clear(inv);
	if (q1 != q) {
	    nxp_int_poly_clear(q);
	    q->coeff = q1->coeff;
	    q->deg = q1->deg;
	    q->lmax = q1->lmax;
	}
	if (r != r1) {
	    nxp_int_poly_clear(r);
	    r->coeff = r1->coeff;
	    r->deg = r1->deg;
	    r->lmax = r1->lmax;
	}

	nxp_int_poly_normalize(q, q, p);
	nxp_int_poly_normalize(r, r, p);

	j = q->deg;
	while (j >= 0 && mpz_sgn(q->coeff[j]) == 0)
	    j--;
	q->deg = j;

	j = r->deg;
	while (j >= 0 && mpz_sgn(r->coeff[j]) == 0)
	    j--;
	r->deg = j;
    }
}

void
nxp_int_poly_div_q_mod(NxpIntPoly *q,
			NxpIntPoly *u, NxpIntPoly *v, mpz_t p)
{
    int m = u->deg, n = v->deg, k, j;
    NxpIntPoly Q[1], R[1];
    NxpIntPoly *q1;
    mpz_t inv;

    if (m < n) {
	q->deg = -1;
	return;
    }
    else {
	if (q == u || q == v) {
	    nxp_int_poly_init(Q, m - n + 1);
	    q1 = Q;
	}
	else {
	    if (m - n >= q->lmax)
		nxp_int_poly_realloc(q, m - n + 1);
	    q1 = q;
	}

	nxp_int_poly_init(R, m + 1);
	nxp_int_poly_set(R, u);

	mpz_init(inv);
	q1->deg = m - n;
	R->deg = n - 1;
	if (mpz_invert(inv, v->coeff[n], p)) {
	    for (k = m - n; k >= 0; k--) {
		mpz_mul(q1->coeff[k], R->coeff[n + k], inv);
		for (j = n + k -1; j >= k; j--)
		    mpz_submul(R->coeff[j], q1->coeff[k], v->coeff[j - k]);
	    }
	}
	else {
	    g_warning("The leading coefficient of the polynomial v has no inverse.\n");
	    g_assert_not_reached();
	}
	mpz_clear(inv);
	if (q1 != q) {
	    nxp_int_poly_clear(q);
	    q->coeff = q1->coeff;
	    q->deg = q1->deg;
	    q->lmax = q1->lmax;
	}

	nxp_int_poly_normalize(q, q, p);
	j = q->deg;
	while (j >= 0 && mpz_sgn(q->coeff[j]) == 0)
	    j--;
	q->deg = j;
	nxp_int_poly_clear(R);
    }
}

/* compute @r in Z/pZ such that u=q*v + r and deg r < deg v.
 * p is not necessarily a prime number, we just need the leading
 * coefficient of v to be inversible in Z/pZ.
 */
void
nxp_int_poly_div_r_mod(NxpIntPoly *r, NxpIntPoly *u, NxpIntPoly *v, mpz_t p)
{
    int m = u->deg, n = v->deg, k, j;
    NxpIntPoly Q[1], R[1];
    NxpIntPoly *r1;
    mpz_t inv;

    if (m < n) {
	nxp_int_poly_set(r, u);
	return;
    }
    else {
	nxp_int_poly_init(Q, m - n + 1);

	if (r != v) {
	    nxp_int_poly_set(r, u);
	    r1 = r;
	}
	else {
	    nxp_int_poly_init(R, m + 1);
	    nxp_int_poly_set(R, u);
	    r1 = R;
	}

	mpz_init(inv);
	Q->deg = m - n;
	r1->deg = n - 1;

	if (mpz_invert(inv, v->coeff[n], p)) {
	    for (k = m - n; k >= 0; k--) {
		mpz_mul(Q->coeff[k], r1->coeff[n + k], inv);
		for (j = n + k -1; j >= k; j--)
		    mpz_submul(r1->coeff[j], Q->coeff[k], v->coeff[j - k]);
	    }
	}
	else {
	    g_warning("The leading coefficient of the polynomial v has no inverse.\n");
	    nxp_int_poly_view(v);
	    g_assert_not_reached();
	}
	mpz_clear(inv);
	nxp_int_poly_clear(Q);
	if (r != r1) {
	    nxp_int_poly_clear(r);
	    r->coeff = r1->coeff;
	    r->deg = r1->deg;
	    r->lmax = r1->lmax;
	}

	nxp_int_poly_normalize(r, r, p);

	j = r->deg;
	while (j >= 0 && mpz_sgn(r->coeff[j]) == 0)
	    j--;
	r->deg = j;
    }
}

void
nxp_int_poly_gcd_mod(NxpIntPoly *d, NxpIntPoly *a, NxpIntPoly *b, mpz_t p)
{
    NxpIntPoly r[1], t[1];
    mpz_t inv;

    nxp_int_poly_init_set(t, b);
    nxp_int_poly_set(d, a);
    nxp_int_poly_normalize(t, t, p);
    nxp_int_poly_init(r, d->deg + 1);
    while (t->deg >= 0) {
	nxp_int_poly_div_r_mod(r, d, t, p);
	nxp_int_poly_set(d, t);
	nxp_int_poly_set(t, r);
    }
    nxp_int_poly_clear(r);
    nxp_int_poly_clear(t);
    mpz_init(inv);
    mpz_invert(inv, d->coeff[d->deg], p);
    nxp_int_poly_mul_mpz(d, d, inv);
    nxp_int_poly_normalize(d, d, p);
    mpz_clear(inv);
}

void
nxp_int_poly_gcdext_mod(NxpIntPoly *d, NxpIntPoly *u, NxpIntPoly *v,
			NxpIntPoly *a, NxpIntPoly *b, mpz_t p)
{
    NxpIntPoly v1[1], v3[1], t1[1], t3[1], q[1];

    nxp_int_poly_set(d, a);
    u->deg = 0;
    mpz_set_ui(u->coeff[0], 1);
    if (b->deg < 0) {
	v->deg = -1;
	return;
    }
    nxp_int_poly_init(v1, 1);
    nxp_int_poly_init(t3, 1);
    nxp_int_poly_init(t1, 1);
    nxp_int_poly_init(q, 1);
    v1->deg = -1;
    nxp_int_poly_init_set(v3, b);
    while (v3->deg >= 0) {
	nxp_int_poly_div_qr_mod(q, t3, d, v3, p);
	nxp_int_poly_mul(t1, q, v1);
	nxp_int_poly_normalize(t1, t1, p);
	nxp_int_poly_sub(t1, u, t1);
	nxp_int_poly_normalize1(t1, t1, p);
	nxp_int_poly_set(u, v1);
	nxp_int_poly_set(d, v3);
	nxp_int_poly_set(v1, t1);
	nxp_int_poly_set(v3, t3);
    }
    nxp_int_poly_mul(v, a, u);
    nxp_int_poly_normalize(v, v, p);
    nxp_int_poly_sub(v, d, v);
    nxp_int_poly_normalize1(v, v, p);
    nxp_int_poly_div_q_mod(v, v, b, p);
    if (d->deg <= 0) {
	if (mpz_cmp_ui(d->coeff[0], 1)) {
	    mpz_t inv;

	    mpz_init(inv);
	    mpz_invert(inv, d->coeff[0], p);
	    nxp_int_poly_mul_mpz(d, d, inv);
	    nxp_int_poly_mul_mpz(u, u, inv);
	    nxp_int_poly_mul_mpz(v, v, inv);
	    mpz_clear(inv);
	}
    }
    nxp_int_poly_normalize(d, d, p);
    nxp_int_poly_normalize(u, u, p);
    nxp_int_poly_normalize(v, v, p);
    nxp_int_poly_clear(v1);
    nxp_int_poly_clear(v3);
    nxp_int_poly_clear(t1);
    nxp_int_poly_clear(t3);
    nxp_int_poly_clear(q);
}

/**
 * nxp_int_poly_hensel_linear:
 *
 * @a1, @b1: Polynomials returned by the procedure.
 * @pol, @a, @b: Polynomials given to the procedure.
 * @p: An GMP integer not necessarily prime.
 * @e: A positive integer.
 *
 * We give to the procedure @pol, @a and @b such that pol=a*b mod p.
 *
 * It returns @a1 and @b1 such that a1=a mod p, b1=b mod p and pol=a1*b1 mod p^e
 *
 **/
void
nxp_int_poly_hensel_linear(NxpIntPoly *a1, NxpIntPoly *b1, NxpIntPoly *pol,
			   NxpIntPoly *a, NxpIntPoly *b, mpz_t p, int e)
{
    NxpIntPoly u[1], v[1], d[1], w[1], s[1], r[1];
    mpz_t q, q1;
    int k;

    nxp_int_poly_init(u, b->deg);
    nxp_int_poly_init(v, a->deg);
    nxp_int_poly_init(d, pol->deg + 1);
    nxp_int_poly_gcdext_mod(d, u, v, a, b, p);
    nxp_int_poly_normalize(a1, a, p);
    nxp_int_poly_normalize(b1, b, p);
    mpz_init_set(q, p);
    mpz_init(q1);
    nxp_int_poly_init(s, pol->deg + 1);
    nxp_int_poly_init(r, pol->deg + 1);
    nxp_int_poly_init(w, pol->deg + 1);
    for (k = 2; k <= e; k++) {
	mpz_mul(q1, p, q);
	nxp_int_poly_mul(d, a1, b1);
	nxp_int_poly_sub(d, pol, d);
	nxp_int_poly_divexact_mpz(d, d, q);
	nxp_int_poly_normalize(d, d, p);
	nxp_int_poly_mul(s, v, d);
	nxp_int_poly_div_qr_mod(w, s, s, a1, p);
	nxp_int_poly_mul(w, w, b1);
	nxp_int_poly_mul(r, u, d);
	nxp_int_poly_add(r, r, w);
	nxp_int_poly_normalize(r, r, p);
	nxp_int_poly_mul_mpz(s, s, q);
	nxp_int_poly_add(a1, a1, s);
	nxp_int_poly_normalize(a1, a1, q1);
	nxp_int_poly_mul_mpz(r, r, q);
	nxp_int_poly_add(b1, b1, r);
	nxp_int_poly_normalize(b1, b1, q1);
	mpz_set(q, q1);
    }
    nxp_int_poly_clear(u);
    nxp_int_poly_clear(v);
    nxp_int_poly_clear(d);
    nxp_int_poly_clear(s);
    nxp_int_poly_clear(r);
    nxp_int_poly_clear(w);
    mpz_clear(q);
    mpz_clear(q1);
}

/**
 * nxp_int_poly_hensel_quad:
 *
 * @a1, @b1: Polynomials returned by the procedure.
 * @pol, @a, @b: Polynomials given to the procedure.
 * @p: An GMP integer not necessarily prime.
 * @e: A positive integer.
 *
 * We give to the procedure @pol, @a and @b such that pol=a*b mod p.
 *
 * It returns @a1 and @b1 such that a1=a mod p, b1=b mod p and pol=a1*b1 mod p^e
 * Compared with hensel_linear the powers of p are doubled each time, so we
 * should obtain the result faster but, when the powers become big, the
 * computations become longer. So hensel_quad is not so much better than
 * hensel_linear.
 **/
void
nxp_int_poly_hensel_quad(NxpIntPoly *a1, NxpIntPoly *b1, NxpIntPoly *pol,
			   NxpIntPoly *a, NxpIntPoly *b, mpz_t p, int e)
{
    NxpIntPoly u[1], v[1], d[1], t[1], s[1], r[1];
    NxpIntPoly polbis[1];
    NxpIntPoly *pol1;

    mpz_t q, q1;
    int k;

    if (pol == a1 || pol == b1) {
	nxp_int_poly_init_set(polbis, pol);
	pol1 = polbis;
    }
    else
	pol1 = pol;

    nxp_int_poly_normalize(a1, a, p);/* Maybe a copy of a should be sufficient */
    nxp_int_poly_normalize(b1, b, p);
    k = 1;
    if (e < k)
	return;
    nxp_int_poly_init(u, b1->deg + 1);
    nxp_int_poly_init(v, a1->deg + 1);
    nxp_int_poly_init(d, pol1->deg + 1);
    nxp_int_poly_gcdext_mod(d, u, v, a1, b1, p);
    mpz_init_set(q, p);
    mpz_init(q1);
    nxp_int_poly_init(s, pol1->deg + 1);
    nxp_int_poly_init(r, pol1->deg + 1);
    nxp_int_poly_init(t, pol1->deg + 1);
    while(k <= e) {
	mpz_mul(q1, q, q);
	nxp_int_poly_mul(d, a1, b1);
	nxp_int_poly_sub(d, pol1, d);

	nxp_int_poly_divexact_mpz(d, d, q);
	nxp_int_poly_normalize(d, d, q);

	nxp_int_poly_mul(s, v, d);
	nxp_int_poly_normalize(s, s, q);
	nxp_int_poly_div_qr_mod(t, s, s, a1, q);
	nxp_int_poly_mul(t, t, b1);
	nxp_int_poly_mul(r, u, d);
	nxp_int_poly_add(r, r, t);
	nxp_int_poly_normalize(r, r, q);
	nxp_int_poly_mul_mpz(s, s, q);
	nxp_int_poly_add(a1, a1, s);
	nxp_int_poly_normalize(a1, a1, q1);
	nxp_int_poly_mul_mpz(r, r, q);
	nxp_int_poly_add(b1, b1, r);
	nxp_int_poly_normalize(b1, b1, q1);

	nxp_int_poly_mul(r, u, a1);
	nxp_int_poly_mul(s, v, b1);
	d->deg = 0;
	mpz_set_ui(d->coeff[0], 1);
	nxp_int_poly_sub(d, d, r);
	nxp_int_poly_sub(d, d, s);
	nxp_int_poly_divexact_mpz(d, d, q);
	nxp_int_poly_mul(s, u, d);
	nxp_int_poly_div_qr_mod(t, r, s, b1, q);
	nxp_int_poly_mul(s, v, d);
	nxp_int_poly_mul(t, t, a1);
	nxp_int_poly_add(s, s, t);
	nxp_int_poly_normalize(s, s, q);
	nxp_int_poly_mul_mpz(r, r, q);
	nxp_int_poly_add(u, u, r);
	nxp_int_poly_normalize(u, u, q1);
	nxp_int_poly_mul_mpz(s, s, q);
	nxp_int_poly_add(v, v, s);
	nxp_int_poly_normalize(v, v, q1);
	mpz_set(q, q1);
	k <<= 1;
    }
    nxp_int_poly_clear(u);
    nxp_int_poly_clear(v);
    nxp_int_poly_clear(d);
    nxp_int_poly_clear(s);
    nxp_int_poly_clear(r);
    nxp_int_poly_clear(t);
    mpz_clear(q);
    mpz_clear(q1);
    if (pol1 != pol)
	nxp_int_poly_clear(pol1);
}

/* @array is the decomposition of pol modulo p. When the function terminates,
 * @array is the decomposition of pol modulo p^e.
 * */
void
nxp_int_poly_hensel_array(NxpIntPoly *pol, GArray *array, mpz_t p, int e)
{
    NxpIntPoly *s;
    int i, size, rsize;
    NxpIntPoly tmp[1], tmp1[1];
    mpz_t p1, q, q1;

    mpz_init(p1);
    mpz_pow_ui(p1, p, e);
    rsize = array->len;

    size = rsize - 1;
    s = g_new(NxpIntPoly, size);
    nxp_int_poly_init_set(tmp, &g_array_index(array, NxpIntPoly, size));
    nxp_int_poly_init(tmp1, tmp->deg + 1);
    nxp_int_poly_init_set(&s[size - 1], tmp);
    for (i = size - 2; i >= 0; i--) {
	nxp_int_poly_mul(tmp, tmp ,&g_array_index(array, NxpIntPoly, i + 1));
	nxp_int_poly_normalize(tmp, tmp, p);
	nxp_int_poly_init_set(&s[i], tmp);
    }
    nxp_int_poly_set(tmp, pol);
    mpz_init(q);
    mpz_init(q1);

    for (i = 0; i < size; i++) {
	nxp_int_poly_hensel_quad(&g_array_index(array, NxpIntPoly, i), tmp,
				 tmp,
				 &g_array_index(array, NxpIntPoly, i),
				 &s[i],
				 p, e);

	nxp_int_poly_normalize(&g_array_index(array, NxpIntPoly, i),
			       &g_array_index(array, NxpIntPoly, i), p1);
    }
    nxp_int_poly_set(&g_array_index(array, NxpIntPoly, i), tmp);
    nxp_int_poly_normalize(&g_array_index(array, NxpIntPoly, i),
			   &g_array_index(array, NxpIntPoly, i), p1);

    nxp_int_poly_clear(tmp);
    nxp_int_poly_clear(tmp1);
    for (i = 0; i < size; i++)
	nxp_int_poly_clear(&s[i]);
    g_free(s);
    mpz_clear(p1);
    mpz_clear(q);
    mpz_clear(q1);
}

static nxp_int_t
_inverse(nxp_int_t b, nxp_int_t nb)
{
    nxp_int_t u, v, d, v1, v3, q, t1, t3;


    u = 1;
    d = b;
    v1 = 0;
    v3 = nb;
    while (v3 != 0) {
	q = d / v3;
	t3 = d % v3;
	t1 = u - q*v1;
	u = v1;
	d = v3;
	v1 = t1;
	v3 = t3;
    }
    v = (d - b * u) / nb;
    switch (d) {
	case 1:
	    return u;
	    break;
	case -1:
	    return -u;
	    break;
	default:
	    g_warning ("%lld has no inverse modulo %lld ; d = %lld\n", b, nb, d);
	    return 0;
    }
}

/* Compute a basis of the kernel of the matrix a. We always work in Z/pZ and
 * p should be a prime number to too big so we only work with nxp_int_t numbers.
 * Instead of returning directly the basis, we prefer to return polynomials
 * obtained with this basis.
 */
static GArray *
kernel_mod(nxp_int_t *m, int n, mpz_t p)
{
    int i, j, k, s;
    nxp_int_t *c;
    nxp_int_t *d;
    nxp_int_t r, dp;
    nxp_int_t pp = mpz_get_si(p);
    GArray *x;
    NxpIntPoly *v;

    c = g_new0(nxp_int_t, n);
    d = g_new(nxp_int_t, n);

    r = 0;
    for (k = 0; k < n; k++) {
	j = 0;
	while (j < n && (m[j * n + k] % pp == 0 || c[j] != 0))
	    j++;
	if (j < n) {
	    dp = -_inverse(m[j * n + k], pp);
	    m[j * n + k] = -1;
	    for (s = k + 1; s < n; s ++)
		m[j * n + s] = m[j * n + s] * dp % pp;
	    for (i = 0; i < n; i++) {
		if (i != j) {
		    dp = m[i * n + k] % pp;
		    m[i * n + k] = 0;
		    for (s = k + 1; s < n; s++)
			m[i * n + s] = (m[i * n + s] + dp * m[j * n + s]) % pp;
		}
	    }
	    c[j] = k + 1;
	    d[k] = j;	
	}
	else {
	    r++;
	    d[k] = -1;
	}
    }
    x = g_array_sized_new(FALSE, FALSE, sizeof(NxpIntPoly), r);
    j = 0;
    for (k = 0; j < r; k++) {	/* we know we need exactly r vectors */
	if (d[k] < 0) {
	    v = g_new(NxpIntPoly, 1);
	    nxp_int_poly_init(v, n);
	    for (i = 0; i < n; i++) {
		if (d[i] >= 0)
		    mpz_set_si(v->coeff[i], m[d[i] * n + k]);
		else if (i == k)
		    mpz_set_ui(v->coeff[i], 1);
		else
		    mpz_set_ui(v->coeff[i], 0);
	    }
	    i = n - 1;
	    while (i >= 0 && mpz_sgn(v->coeff[i]) == 0)
		i--;
	    v->deg = i;
	    g_array_append_val(x, *v);
	    j++;
	}
    }
    g_free(c);
    g_free(d);
    return x;
}

static GSList *
nxp_int_poly_berlekamp(NxpIntPoly *a, mpz_t p, int d)
{
    int i, k, n, r;
    nxp_int_t pp, p1;
    GSList *e = NULL, *l;
    nxp_int_t *q;
    NxpIntPoly ttmp[1], aa[1], *item, tt[1], quo[1];
    mpz_t rnd;
    GArray *t;

    /* We must have pp > 2, Berlekamp needs a prime number greater than 2 */
    pp = mpz_get_si(p);

    /* Compute Q */
    n = a->deg;

    g_assert(n != d);	    /* If n == d then a is irreducible. */

    nxp_int_poly_init(aa, a->deg + 1);
    q = g_new0(nxp_int_t, n * n);
    nxp_int_poly_init(ttmp, pp * (n - 1) + 1);
    for (k = 0; k < n; k++) {
	/* tmp = X^(pk) */
	for (i = 0; i < pp * k; i++)
	    mpz_set_ui(ttmp->coeff[i], 0);
	mpz_set_ui(ttmp->coeff[i], 1);
	ttmp->deg = i;
	nxp_int_poly_div_r_mod(ttmp, ttmp, a, p);
	for (i = 0; i <= ttmp->deg; i++)
	    q[i * n + k] = mpz_get_si(ttmp->coeff[i]);
    }
    /* We make the difference of q and In */
    for (i = 0; i < n; i++)
	q[i * (n + 1)] -= 1;
    /* Q is done, we compute now the kernel of Q-I, the basis is a vector of polynomials of the variable a->arg. */
    t = kernel_mod(q, n, p);
    g_free(q);

    r = t->len;

    item = g_new(NxpIntPoly, 1);
    nxp_int_poly_init_set(item, a);
    e = g_slist_prepend(e, item);
    k = 1;

    p1 = (pp >> 1) + 1;
    mpz_init(rnd);
    nxp_int_poly_init(tt, n * p1 + 1);
    nxp_int_poly_init(quo, n + 1);
    while (k != r) {
	do {
	    aa->deg = -1;
	    for (i = 0; i < r; i++) {
		mpz_set_ui(rnd, g_random_int_range(0, pp));
		nxp_int_poly_mul_mpz(ttmp, &g_array_index(t, NxpIntPoly, i), rnd);
		nxp_int_poly_normalize(ttmp, ttmp, p);
		nxp_int_poly_add(aa, aa, ttmp);
		nxp_int_poly_normalize1(aa, aa, p);
	    }
	}
	while (aa->deg < 0);

	l = e;
	nxp_int_poly_power_mod(tt, aa, p1, pp);
	nxp_int_poly_sub_si(tt, tt, 1);
	nxp_int_poly_normalize1(tt, tt, p);
	while (l) {
	    item = l->data;
	    if (item->deg > d) {
		nxp_int_poly_gcd_mod(ttmp, tt, item, p);
		if (ttmp->deg > 0 && ttmp->deg < item->deg) {
		    nxp_int_poly_div_q_mod(quo, item, ttmp, p);
		    nxp_int_poly_set(item, ttmp);
		    item = g_new(NxpIntPoly, 1);
		    nxp_int_poly_init_set(item, quo);
		    e = g_slist_append(e, item);
		    k ++;
		    if (k == r)
			break;
		}
		else
		    l = l->next;
	    }
	    else
		l = l->next;
	}
    }
    mpz_clear(rnd);
    nxp_int_poly_clear(aa);
    nxp_int_poly_clear(tt);
    nxp_int_poly_clear(quo);
    nxp_int_poly_clear(ttmp);
    return e;
}

/* Assuming a is a squarefree polynomial, we decompose @a into a DISTINCT
 * DEGREE FACTORIZATION in Z/pZ */
GSList *
nxp_int_poly_ddf(NxpIntPoly *a, mpz_t p)
{
    GSList *list, *list1;
    NxpIntPoly q[1], x[1], pol[1], g[1];
    int k, d, pp;
    NxpIntPoly *tmp;

    nxp_int_poly_init(q, 2);
    q->deg = 1;
    mpz_set_ui(q->coeff[0], 0);
    mpz_set_ui(q->coeff[1], 1);
    nxp_int_poly_init_set(x, q);
    list = NULL;
    pp = mpz_get_si(p);

    nxp_int_poly_init_set(pol, a);
    nxp_int_poly_init(g, pol->deg + 1);

    k = 1;
    d = pol->deg;
    while (d > 0) {
	if (k > (d >> 1)) {
	    tmp = g_new(NxpIntPoly, 1);
	    nxp_int_poly_init_set(tmp, pol);
	    list = g_slist_append(list, tmp);
	    break;
	}
	nxp_int_poly_power_mod(q, q, pp, pp);
	nxp_int_poly_div_r_mod(q, q, pol, p);
	nxp_int_poly_sub(g, q, x);
	nxp_int_poly_gcd_mod(g, g, pol, p);
	if (g->deg == pol->deg)
	    nxp_int_poly_set(g, pol);
	if (g->deg > 0) {
	    /* g shouldn't be equal to 1 */
	    if (g->deg == k) {
		tmp = g_new(NxpIntPoly, 1);
		nxp_int_poly_init_set(tmp, g);
		list = g_slist_append(list, tmp);
	    }
	    else {
		list1 = nxp_int_poly_berlekamp(g, p, k);
		list = g_slist_concat(list, list1);
	    }
	    nxp_int_poly_div_q_mod(pol, pol, g, p);
	}
	k++;
	d = pol->deg;
    }
    nxp_int_poly_clear(q);
    nxp_int_poly_clear(x);
    nxp_int_poly_clear(pol);
    nxp_int_poly_clear(g);
    return list;
}

/* pol is a polynomial with integral coefficients without double roots.
 * list is the list of pol divisors modulo p.
 *
 * the return value is the exact divisors of pol in Z[X].
 * */
GSList *
nxp_int_poly_get_factors(NxpIntPoly *pol, GSList *list, mpz_t p)
{
    GSList *lst;
    int i, e;
    int d = pol->deg;
    int size = g_slist_length(list);
    GArray *array;
    mpz_t pp, mignotte;
    NxpIntPoly pptmp[1];
    GSList *factors;
    /* Here we compute the Mignotte-Landau constant. Any polynomials with
     * integral coefficients that divides pol has his coefficients bounded by
     * this constant. :-) */

    mpz_init(mignotte);
    nxp_int_poly_mignotte(mignotte, pol, d >> 1);

    /* We have a list of potential dividers of pol (modulo p). We lift those
     * dividers modulo p^e where e is choosed such that p^e>2*l(pol)*mignotte.
     * It's not useful to lift those dividers indefinitely... */
    /* Let's compute e... */

    mpz_add(mignotte, mignotte, mignotte);
    mpz_mul(mignotte, mignotte, pol->coeff[d]);

    mpz_init_set(pp, p);
    e = 1;

    while (mpz_cmp(pp, mignotte) <= 0) {
	e ++;
	mpz_mul(pp, pp, p);
    }

    /* We should have pp = p^e */

    /* We apply Hensel lift to lift the factors from modulo p to modulo p^e */

    /* FIXME: Conversion of the list to an array, we should change this... */
    array = g_array_sized_new(FALSE, FALSE, sizeof(NxpIntPoly), size);
    for (lst = list, i = 0; i < size; i++) {
	nxp_int_poly_init_set(pptmp, lst->data);
	g_array_append_val(array, pptmp);
	lst = lst->next;
    }

    nxp_int_poly_hensel_array(pol, array, p, e);

    factors = nxp_int_poly_build_factors(pol, array, pp);
    g_array_free(array, TRUE);

    mpz_clear(pp);
    mpz_clear(mignotte);
    return factors;
}
