/* -*- c -*- */

#include "typecasts.h"
#include "prototypes.h"
#include "types/types.h"
#include "errors.h"
#include "namespace.h"
#include "i18n.h"

#define abs(x) (x>0?x:-(x))

NxpProtoInstance* nxp_proto_instance_new(void)
{
    return g_new0(NxpProtoInstance, 1);
}


void nxp_proto_instance_free(NxpProtoInstance *proto)
{
    g_return_if_fail(proto);

    if (proto->user_data) {
	if (!proto->free_user_data) {
	    g_warning("Created prototype userdata without free function");
	} else {
	    (proto->free_user_data)(proto->user_data);
	}
    }

    g_free(proto);
} 


static gint protoinstcomp(gconstpointer a, gconstpointer b)
{
    const NxpProtoInstance *p1 = a, *p2 = b;
    if (!p1->protoclass || !p2->protoclass) return 0;
    return p1->protoclass->arity < p2->protoclass->arity;
}


/*
 * nxp_bifunc_define_global:
 * @name: name
 * @prototype: pointer to an existing #NxpProtoClass structure describing prototype
 * @eval_func: eval function
 * @derive_func: derive function or %NULL
 * @simplify_func: simplify function or %NULL
 * @context: eval context (to specify where is the global namespace)
 * @error: may be set to NXP_ERROR_ALREADY_DEFINED
 * 
 * Define a built-in function with a prototype as global (at the root namespace).
 **/

void nxp_bifunc_define_global(const char *name, const NxpProtoClass *prototype,
			      gpointer eval_func, gpointer derive_func,
			      gpointer simplify_func, NxpEvalContext *context, 
			      GError **error)
{
    nxp_bifunc_define_global_full(name, prototype, eval_func, derive_func,
				  simplify_func, NULL, NULL, context, error);
}

void nxp_bifunc_define_global_volatile(const char *name, const NxpProtoClass *prototype,
				       gpointer eval_func, gpointer derive_func,
				       gpointer simplify_func, NxpEvalContext *context, 
				       GError **error)
{
    nxp_bifunc_define_global_full_volatile(name, prototype, eval_func, derive_func,
					   simplify_func, NULL, NULL, context, error);
}

void nxp_bifunc_define_global_full(const char *name, const NxpProtoClass *prototype,
				   gpointer eval_func, gpointer derive_func,
				   gpointer simplify_func, gpointer user_data,
				   NxpFreeData free_data,
				   NxpEvalContext *context, GError **error)
{
    NxpIdentifier *ident;
    ident = nxp_identifier_new();
    nxp_identifier_append_separator(ident);
    nxp_identifier_append_component(ident, name);
    nxp_bifunc_define_full(ident, prototype, eval_func, derive_func, simplify_func,
			   user_data, free_data, context, error);
    nxp_identifier_unref(ident);
}

void nxp_bifunc_define_global_full_volatile(const char *name, const NxpProtoClass *prototype,
					    gpointer eval_func, gpointer derive_func,
					    gpointer simplify_func, gpointer user_data,
					    NxpFreeData free_data,
					    NxpEvalContext *context, GError **error)
{
    NxpIdentifier *ident;
    NxpDefinition *def;
    ident = nxp_identifier_new();
    nxp_identifier_append_separator(ident);
    nxp_identifier_append_component(ident, name);
    def = nxp_bifunc_define_detailed(ident, prototype, eval_func, derive_func,
				     simplify_func, user_data, free_data, context,
				     NXP_PROTO_INSTANCE_VOLATILE, error);
    nxp_identifier_unref(ident);
    nxp_definition_unref(def);
}

/**
 * nxp_bifunc_define:
 * @ident: name
 * @prototype: pointer to an existing #NxpProtoClass structure describing prototype
 * @eval_func: eval function (mandatory)
 * @derive_func: derive function or %NULL
 * @simplify_func: simplify function or %NULL
 * @context: eval context (to specify where is the global namespace)
 * @error: may be set to NXP_ERROR_ALREADY_DEFINED
 * 
 * Define a built-in function with a prototype.
 **/
void nxp_bifunc_define(NxpIdentifier *ident, const NxpProtoClass *prototype,
		       gpointer eval_func, gpointer derive_func,
		       gpointer simplify_func, NxpEvalContext *context, GError **error)
{
    nxp_bifunc_define_full(ident, prototype, eval_func, derive_func,
			    simplify_func, NULL, NULL, context, error);
}

void nxp_bifunc_define_full(NxpIdentifier *ident, const NxpProtoClass *prototype,
			    gpointer eval_func, gpointer derive_func,
			    gpointer simplify_func, gpointer user_data,
			    NxpFreeData free, NxpEvalContext *context, GError **error)
{
    NxpDefinition *def;
    def = nxp_bifunc_define_detailed(ident, prototype, eval_func, derive_func,
				     simplify_func, user_data, free, context, 0, error);
    nxp_definition_unref(def);
}


/**
 * nxp_bifunc_define_return_def:
 * @ident: name
 * @prototype: pointer to an existing #NxpProtoClass structure
 * describing prototype. Should be %NULL in case the flag
 * %NXP_PROTO_INSTANCE_DETAILED is set
 * @eval_func: eval function (mandatory)
 * @derive_func: derive function or %NULL
 * @simplify_func: simplify function or %NULL
 * @context: eval context (to specify where is the current namespace)
 * @flags: see #NxpProtoInstanceFlags
 * @error: may be set to NXP_ERROR_ALREADY_DEFINED
 * 
 * Define a built-in function with a prototype and returns the definition.
 *
 * Return value: a new reference to the new or existing
 * definition. Caller takes ownership of this.
 **/
NxpDefinition* nxp_bifunc_define_detailed(NxpIdentifier *ident,
					  const NxpProtoClass *prototype,
					  gpointer eval_func, gpointer derive_func,
					  gpointer simplify_func,
					  gpointer user_data,
					  NxpFreeData free,
					  NxpEvalContext *context, 
					  NxpProtoInstanceFlags flags,
					  GError **error)
{
    NxpProtoInstance *inst;
    NxpNameSpace *ns;
    const char *name;
    NxpDefinition *def;
    NxpDefBIFunc *bifuncdef;

    /* Try to get existing definition */
    nxp_identifier_resolve(ident, context, &ns, &name);
    g_assert(ns && name);

    def = nxp_name_space_lookup_definition_extended(ns, name, NULL, FALSE);
    if (def)
    {
	if (def->type == NXP_DEF_BI_FUNC)
	{
	    GSList *l;
	    bifuncdef = NXP_DEF_BI_FUNC(def);
	    /* Check if there's already a prototype instance with the same class */
	    for (l = bifuncdef->prototypes; l; l = l->next) {
		NxpProtoInstance *instance = (NxpProtoInstance*)l->data;
		if (instance->protoclass == prototype)
		{
		    if (instance->flags & NXP_PROTO_INSTANCE_VOLATILE) {
			nxp_proto_instance_free(instance);
			instance = nxp_proto_instance_new();
			instance->protoclass = prototype;
			instance->eval_func = eval_func;
			instance->derive_func = derive_func;
			instance->simplify_func = simplify_func;
			instance->user_data = user_data;
			instance->free_user_data = free;
			instance->flags = flags;
			l->data = instance;
		    } else {
			g_set_error(error, nxp_error_domain, NXP_ERROR_ALREADY_DEFINED,
				    _("Function already defined with same prototype"));
		    }
		    return def;
		}
	    }
	} else {
	    g_set_error(error, nxp_error_domain, NXP_ERROR_ALREADY_DEFINED, _("Function already defined with different definition type"));
	    return def;
	}
    } else {
	bifuncdef = nxp_def_bi_func_new(g_strdup(name));
	nxp_definition_ref(NXP_DEFINITION(bifuncdef));
	nxp_name_space_add_definition(ns, NXP_DEFINITION(bifuncdef));
    }
    inst = nxp_proto_instance_new();
    inst->protoclass = prototype;
    inst->eval_func = eval_func;
    inst->derive_func = derive_func;
    inst->simplify_func = simplify_func;
    inst->user_data = user_data;
    inst->free_user_data = free;
    inst->flags = flags;
    bifuncdef->prototypes = g_slist_insert_sorted(bifuncdef->prototypes,
						  inst, protoinstcomp);
    return NXP_DEFINITION(bifuncdef);
}



/* Check if first n arguments match prototype */

typedef struct {
    gint t;
    NxpElement *new;
} NxpMatch;

static NxpMatch args_match(int n, GSList *args,
			   NxpElementType *types,
			   NxpEvalContext *context, gboolean simp, GError **error)
{
    GError *err = NULL;
    NxpMatch ret;
    /* return:
       - 0   if don't match;
       - 50  if some TYPE_ANY has matched;
       - 75  if we need to typecast something;
       - 100 if all types matched */
    NxpElementType *type;

    ret.t = 0;
    ret.new = nxp_element_list_new();

    for (type = types; n; type++, n--, args = args->next) {
	if (*type == NXP_ELEMENT_TYPE_ANY) {
	    ret.t += 50;
	    nxp_element_list_append(ret.new, args->data);
	}
	else if (nxp_element_check_type(NXP_ELEMENT(args->data), *type)) {
	    ret.t += 100;
	    nxp_element_list_append(ret.new, args->data);
	} 
	else {
	    NxpElement *el;

	    if (simp)
		el = nxp_element_simplify(args->data, context, &err);
	    else
		el = nxp_element_eval(args->data, context, &err);

	    if (el && nxp_element_check_type(el, *type)) {
		ret.t += 95;
		nxp_element_list_append(ret.new, el);
	    }
	    else if (nxp_typecasts_exists(NXP_ELEMENT_GET_TYPE(NXP_ELEMENT(args->data)),*type)) {
		ret.t += 75;
		nxp_element_list_append(ret.new, args->data);
	    }
	    else if (el && nxp_typecasts_exists(NXP_ELEMENT_GET_TYPE(el),*type)) {
		ret.t += 70;
		nxp_element_list_append(ret.new, el);
	    } else {
		ret.t = 0;
		NXP_ELEMENT_UNREF(ret.new);
		NXP_ELEMENT_XUNREF(el);
		ret.new = NULL;

		/* Nothing was found and maybe it is due to an error, so we
		 * propagate it */
		if (err)
		    g_propagate_error(error, err);
		return ret;
	    }
	    NXP_ELEMENT_XUNREF(el);
	}
    }
    /* We found something, so we can clear the error if it exists */
    if (err)
	g_clear_error(&err);
    return ret;
}

static NxpProtoInstance* nxp_find_prototype(NxpElement *el_args,
				     NxpElement **new_args,
				     GSList *proto_list,
				     NxpEvalContext *context, gboolean simp, GError **error)
{
    GSList *args = NXP_ELEMENT_LIST(el_args)->list;
    guint nargs = g_slist_length(args);
    int arity;
    GSList *le;
    NxpMatch t;
    const NxpProtoClass *proto;
    NxpProtoInstance *proto_instance;
    NxpProtoInstance* better = NULL;
    NxpMatch max = {0, NULL};
    GError *err = NULL;

    g_return_val_if_fail(proto_list, NULL);

    for (le = proto_list; le; le = le->next)
    {
	proto_instance = (NxpProtoInstance*) le->data;
	proto = proto_instance->protoclass;
	if (!proto) {
	    *new_args = NXP_ELEMENT_REF(el_args);
	    return proto_instance;
	}
	arity = proto->arity;
	/* simplest case of fixed number of arguments (arity is
           positive or zero) */
	if (arity >= 0) {

	    /* Use the old max algorithm to find the best match :-) */
	    t = args_match(nargs, args, proto->types, context, simp, &err);
	    if (arity == nargs && t.t > max.t ) {
		better = proto_instance;
		if (max.new)
		    NXP_ELEMENT_UNREF(max.new);
		max = t;
		if (max.t == arity * 100) {
		    if (new_args)
			*new_args = max.new;
		    else
			NXP_ELEMENT_UNREF(max.new);
		    return better;
		}
	    }
	    else if (err) {
		g_propagate_error(error, err);
		return proto_instance;
	    }
	    else continue;
	} else /* arity < 0 */ {
	    /* "any number of arguments with zero mandatory arguments"
               prototype always matches anything */
	    if (arity == INT_MIN) {
		if (new_args)
		    *new_args = NXP_ELEMENT_REF(el_args);
		return proto_instance;
	    }
	    /* the only case left is "any number of arguments with
               abs(arity) mandatory arguments" */
	    t = args_match(abs(arity), args, proto->types, context, simp, &err);
	    if (nargs >= abs(arity) && t.t) {
		if (new_args)
		    *new_args = t.new;
		else
		    NXP_ELEMENT_UNREF(t.new);

		return proto_instance;
	    }
	    else if (err) {
		g_propagate_error(error, err);
		return proto_instance;
	    }
	}
    }
    /* If better is NULL at this point, no prototype was found */
    if (new_args)
	*new_args = max.new;
    return better;
}

/**
 * nxp_find_simp_prototype:
 * @args: argument list: list of elements
 * @new_args: list of new arguments
 * @proto_list: list of NxpProtoInstance
 * 
 * 
 * Find prototype that more closely matches argument list. The function can 
 * simplify arguments, the new arguments to give to the function are new_args.
 * 
 * Return value: prototype instance that matched or %NULL if none matched.
 **/
inline NxpProtoInstance* nxp_find_simp_prototype(NxpElement *el_args,
						NxpElement **new_args,
						GSList *proto_list,
						NxpEvalContext *context,
						GError **error)
{
    return nxp_find_prototype(el_args, new_args, proto_list, context, TRUE, error);
}

/**
 * nxp_find_eval_prototype:
 * @args: argument list: list of elements
 * @new_args: List of new arguments
 * @proto_list: list of NxpProtoInstance
 * 
 * 
 * Find prototype that more closely matches argument list. The function can 
 * evaluate arguments, the new arguments to give to the function are new_args.
 * 
 * Return value: prototype instance that matched or %NULL if none matched.
 **/
inline NxpProtoInstance* nxp_find_eval_prototype(NxpElement *el_args,
						NxpElement **new_args,
						GSList *proto_list,
						NxpEvalContext *context,
						GError **error)
{
    return nxp_find_prototype(el_args, new_args, proto_list, context, FALSE, error);
}
