/* -*- Mode: C; c-file-style: "stroustrup"; -*- */

#include <stdio.h>
#include <math.h>
#include "frac.h"
#include "int.h"
#include "tree.h"
#include "r.h"
#include "real.h"
#include "message.h"
#include "op_bindings.h"
#include "types_misc.h"
#include "errors.h"
#include "parser/plugins.h"
#include "i18n.h"

static NxpElementTypeInfo *typeinfo = NULL;

static NxpElement* nxp_frac_simplify(NxpElement* elem,
				     NxpEvalContext *context, GError **error)
{
    return NXP_ELEMENT_REF(elem);
}

NxpElement* nxp_element_frac_new(nxp_int_t numerator,
				 nxp_int_t denumerator) 
{
    NxpElementFrac *elem;

    elem = (NxpElementFrac *)nxp_element_alloc(typeinfo, sizeof(NxpElementFrac));
    mpq_init(elem->value);
    mpz_set_si(mpq_numref(elem->value), numerator);
    mpz_set_si(mpq_denref(elem->value), denumerator);
    mpq_canonicalize(elem->value);
    return (NxpElement *)elem;
}

NxpElement* nxp_element_frac_new_from_mpq(mpq_t q)
{
    NxpElementFrac *elem;
    elem = (NxpElementFrac *)nxp_element_alloc(typeinfo, sizeof(NxpElementFrac));
    mpq_init(elem->value);
    elem->value[0] = q[0];
    return (NxpElement *)elem;
}

NxpElement* nxp_element_frac_new_from_string(const char* n)
{
    NxpElementFrac *elem;
    elem = (NxpElementFrac *)nxp_element_alloc(typeinfo, sizeof(NxpElementFrac)); 
    mpq_init(elem->value);
    if (mpq_set_str(elem->value, n, 10) < 0) {
	NXP_ELEMENT_UNREF(elem);
	g_warning("divide by zero : file %s line %d", __FILE__, __LINE__);
	return NULL;
    }
    else {
	mpq_canonicalize(elem->value);
	return (NxpElement *)elem;
    }
}

static GString* str_func(NxpElement *el, int *precedence) {
    GString *gstr;
    char *str;

    gmp_asprintf(&str, "%Qi", NXP_ELEMENT_FRAC(el)->value);
    gstr = g_string_new(str);
    g_free(str);

    if (precedence)
	*precedence = 40;

    return gstr;
} 

#define FRAC_FRAC(func,mpq_cmd)   static NxpElement* func(NxpOp op, NxpElement* a, NxpElement* b, \
   			                             NxpEvalContext *context, GError **error) {\
     mpq_t i; NxpElement *el;\
     mpq_init(i); mpq_cmd(i, NXP_ELEMENT_FRAC(a)->value, NXP_ELEMENT_FRAC(b)->value);\
     el = nxp_element_frac_new_from_mpq(i);\
     return el; \
}

FRAC_FRAC(mul, mpq_mul);
FRAC_FRAC(add, mpq_add);
FRAC_FRAC(sub, mpq_sub);

#undef FRAC_FRAC

static NxpElement *divide(NxpOp op, NxpElement* a, NxpElement *b,
			  NxpEvalContext *context, GError **error) {
    mpq_t i;
    NxpElement *el;

    if (mpz_sgn(mpq_numref(NXP_ELEMENT_FRAC(b)->value)) != 0) {
	mpq_init(i);
	mpq_div(i, NXP_ELEMENT_FRAC(a)->value, NXP_ELEMENT_FRAC(b)->value);
	el = nxp_element_frac_new_from_mpq(i);
    }
    else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO,
			_("divide by zero"));
	return NULL;
    }
    return el;
}

static inline NxpElement* nxp_element_frac_opposed(NxpElementNumber *self)
{
    NxpElement *el;
    mpq_t q;

    mpq_init(q);
    mpq_set(q, (NXP_ELEMENT_FRAC(self)->value));
    mpq_neg(q, q);
    el = nxp_element_frac_new_from_mpq(q);
    return el;
}

static NxpElement *uminus(NxpOp op, NxpElement* a, NxpElement *b,
			  NxpEvalContext *context, GError **error) {
    return nxp_element_frac_opposed(NXP_ELEMENT_NUMBER(b));
}

#define BIND(op,func)		    nxp_bind_operator_function (op, 	  \
 			            NXP_ELEMENT_TYPE_FRAC, NXP_ELEMENT_TYPE_FRAC, \
   			            func, func, 0, NULL);

#define BIND_LEFT(op,func)	    nxp_bind_operator_function (op, 	  \
 			            0, NXP_ELEMENT_TYPE_FRAC, \
   			            func, func, 0, NULL);

#define BIND_FUNC_LEFT(func,sfunc,op)   nxp_bind_operator_function (op, \
    				        0, NXP_ELEMENT_TYPE_FRAC, 	 \
   				        func, sfunc, 0, NULL);


static NxpElement* nxp_frac_eval(NxpElement *a, NxpEvalContext *context, GError **error)
{ 
    NxpElement *el;
    mpfr_t result;

    if (mpz_cmp_si(mpq_denref(NXP_ELEMENT_FRAC(a)->value), 1)) {
	mpfr_init_set_q(result, NXP_ELEMENT_FRAC(a)->value, GMP_RNDN);
	el = nxp_element_real_new_from_mpfr(result);
    }
    else {
	mpz_t z;
	mpz_init_set(z, mpq_numref(NXP_ELEMENT_FRAC(a)->value));
	el = nxp_element_int_new_from_mpz(z);
    }
    return el;
} 


static gboolean truth_func(NxpElement *elem, NxpEvalContext *context, GError **error)
{ 
    return (mpq_cmp_si(NXP_ELEMENT_FRAC(elem)->value, 0, 1) != 0);
}

static int nxp_element_frac_compare_int(NxpElementNumber *self, nxp_int_t nb)
{
    return mpq_cmp_si(NXP_ELEMENT_FRAC(self)->value, nb, 1);
}

static gboolean nxp_element_frac_is_integer(NxpElementNumber *self)
{
    return FALSE;
}

static double nxp_element_frac_get_value_as_double(NxpElementR *self)
{
    return mpq_get_d(NXP_ELEMENT_FRAC(self)->value);
}

static nxp_int_t nxp_element_frac_get_value_as_int(NxpElementR *self)
{
    return (nxp_int_t)mpq_get_d(NXP_ELEMENT_FRAC(self)->value);
}

NxpElement *nxp_element_frac_get_num(NxpElement *self)
{
    mpz_t z;

    mpz_init_set(z, mpq_numref(NXP_ELEMENT_FRAC(self)->value));
    return nxp_element_int_new_from_mpz(z);
}

NxpElement *nxp_element_frac_get_den(NxpElement *self)
{
    mpz_t z;

    mpz_init_set(z, mpq_denref(NXP_ELEMENT_FRAC(self)->value));
    return nxp_element_int_new_from_mpz(z);
}

static void destroy(NxpElement *el) {
    mpq_clear(NXP_ELEMENT_FRAC(el)->value);
}

NxpElementType nxp_element_frac_get_type(void)
{
    static NxpElementType nxp_element_type_frac = 0;
    if (nxp_element_type_frac == 0) {
	nxp_element_type_frac = nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
				    nxp_element_type_frac, /* The new type */
				    sizeof(NxpElementRTypeInfo), /* the size of our TypeInfo */
				    NXP_ELEMENT_TYPE_R, /* The type of the parent */
				    sizeof(NxpElementRTypeInfo)); /* size of the typeinfo of the parent */
    }
    return nxp_element_type_frac;
}

void nxp_element_frac_type_init() 
{
    NxpElementNumberTypeInfo *nb_typeinfo;
    NxpElementRTypeInfo *r_typeinfo;
    typeinfo = nxp_element_type_get_info(nxp_element_frac_get_type());
    nb_typeinfo = (NxpElementNumberTypeInfo *)typeinfo;
    r_typeinfo = (NxpElementRTypeInfo *)typeinfo;
    typeinfo->name = g_strdup("frac"); 
    typeinfo->eval_func = nxp_frac_eval;
    typeinfo->simplify_func = nxp_frac_simplify;
    typeinfo->str_func = str_func;
    typeinfo->repr_func = str_func;
    typeinfo->truth_func = truth_func;
    typeinfo->destroy_func = destroy;

    nb_typeinfo->compare_int = nxp_element_frac_compare_int;
    nb_typeinfo->is_integer = nxp_element_frac_is_integer;
    nb_typeinfo->opposed = nxp_element_frac_opposed;
    r_typeinfo->get_value_as_double = nxp_element_frac_get_value_as_double;
    r_typeinfo->get_value_as_int = nxp_element_frac_get_value_as_int;

    nxp_parser_register_func(NXP_ATOM_CLASS_FRAC,
			     nxp_element_frac_new_from_string, 50);

    BIND(NXP_OP_PLUS, add);
    BIND(NXP_OP_MINUS, sub);
    BIND(NXP_OP_STAR, mul);
    BIND(NXP_OP_SLASH, divide);
    BIND_LEFT(NXP_OP_MINUS_LEFT, uminus);

/*     BIND(NXP_OP_PERCENT); */
/*     BIND(NXP_OP_HAT); */
/*     BIND(NXP_OP_DOT3); */
/*     BIND_LEFT(NXP_OP_PLUS2_LEFT); */
/*     BIND_LEFT(NXP_OP_MINUS2_LEFT); */
/*     BIND_RIGHT(NXP_OP_PLUS2_RIGHT); */
/*     BIND_RIGHT(NXP_OP_MINUS2_RIGHT); */

/*     BIND_FUNC(gt,NXP_OP_GT); */
/*     BIND_FUNC(ge,NXP_OP_GT_EQ); */
/*     BIND_FUNC(lt,NXP_OP_LT); */
/*     BIND_FUNC(le,NXP_OP_LT_EQ); */
/*     BIND_FUNC(eq,NXP_OP_EQ2); */
/*     BIND_FUNC(ne,NXP_OP_EXCL_EQ); */

/*     BIND_FUNC(nxp_int_and_or_int,NXP_OP_SLASH_BSLASH); */
/*     BIND_FUNC(nxp_int_and_or_int,NXP_OP_BSLASH_SLASH); */

/*     BIND_LEFT(NXP_OP_TILDE_LEFT); */
/*     BIND_FUNC_RIGHT(nxp_int_factorial,NXP_OP_EXCL_RIGHT); */
/*     BIND(NXP_OP_EQ); */
/*     BIND(NXP_OP_COLON_EQ); */
}
