/* -*- Mode: C; c-file-style: "stroustrup"; -*- */
#include <glib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "r.h"
#include "real.h"
#include "int.h"
#include "tree.h"
#include "op_bindings.h"
#include "element_type.h"
#include "types_misc.h"
#include "errors.h"
#include "parser/plugins.h"
#include "i18n.h"

static NxpElementTypeInfo *typeinfo = NULL;

NxpElement* nxp_element_real_new_from_string(const gchar *string)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementReal));
    mpfr_init_set_str(NXP_ELEMENT_REAL(elem)->value, string, 10, GMP_RNDN);
    return elem;
}


static inline NxpElement* _element_real_new(nxp_real_t value)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementReal));
    mpfr_init_set_d(NXP_ELEMENT_REAL(elem)->value, value, GMP_RNDN);
    return elem;
}

/**
 * nxp_element_real_new:
 * @value: value of element
 * 
 * Create a new NxpElementReal
 * 
 * Return value: the new element
 **/
NxpElement* nxp_element_real_new(nxp_real_t value)
{
    static NxpElement *zero = NULL, *one = NULL, *minus_one = NULL;

    if (!zero) zero = _element_real_new(0);
    if (!one) one = _element_real_new(1.0);
    if (!minus_one) minus_one = _element_real_new(-1.0);

    if (value == 0.0) return NXP_ELEMENT_REF(zero);
    if (value == 1.0) return NXP_ELEMENT_REF(one);
    if (value == -1.0) return NXP_ELEMENT_REF(minus_one);

    return _element_real_new(value);
}

NxpElement* nxp_element_real_new_from_mpfr(mpfr_t value)
{
    NxpElement *elem;
    elem = nxp_element_alloc(typeinfo, sizeof(NxpElementReal));
    NXP_ELEMENT_REAL(elem)->value[0] = value[0];
    return elem;
}

#define REAL_REAL(func, mpfr_cmd)   static NxpElement* func(NxpOp op, NxpElement* a, NxpElement* b, \
   			                             NxpEvalContext *context, GError **error) {\
    mpfr_t i;\
    mpz_t n;\
    NxpElement *el;\
    mpfr_init(i);\
    mpfr_cmd(i, NXP_ELEMENT_REAL(a)->value, NXP_ELEMENT_REAL(b)->value, GMP_RNDN);\
    if (mpfr_integer_p(i)) {\
	mpz_init(n);\
	mpfr_get_z(n, i, GMP_RNDN);\
	el = nxp_element_int_new_from_mpz(n);\
	mpfr_clear(i);\
    }\
    else\
	el = nxp_element_real_new_from_mpfr(i);\
    return el;\
}

REAL_REAL(times, mpfr_mul);
REAL_REAL(add, mpfr_add);
REAL_REAL(sub, mpfr_sub);
#undef REAL_REAL

static NxpElement*
divide(NxpOp op, NxpElement* a, NxpElement* b, NxpEvalContext *context, GError **error)
{
    mpfr_t i;
    mpz_t n;
    NxpElement *el;
    
    if (mpfr_sgn(NXP_ELEMENT_REAL(b)->value) == 0) {
        g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO,
		    _("divide by zero"));
        return NULL;
    }
    mpfr_init(i);
    mpfr_div(i, NXP_ELEMENT_REAL(a)->value, NXP_ELEMENT_REAL(b)->value, GMP_RNDN);
    if (mpfr_integer_p(i)) {
	mpz_init(n);
	mpfr_get_z(n, i, GMP_RNDN);
	mpfr_clear(i);
	el = nxp_element_int_new_from_mpz(n);
    }
    else
	el = nxp_element_real_new_from_mpfr(i);
    return el;
};

static NxpElement* eval_real_hat(NxpOp op, NxpElement *a, NxpElement *b, 
				 NxpEvalContext *context, GError **error)
{
    mpfr_t r;

    mpfr_init(r);
    mpfr_pow(r, NXP_ELEMENT_REAL(a)->value,
	     NXP_ELEMENT_REAL(b)->value, GMP_RNDN);
    return nxp_element_real_new_from_mpfr(r);
}

static GString* str_func(NxpElement *el, int *precedence)
{
    GString *gstr;
    char *str, *str1, *beg, *end, *virg, *s, *s1;
    int l, nl, neg;
    mp_exp_t e;

    if (precedence)
	*precedence = 0;
    str = mpfr_get_str(NULL, &e, 10, 0, NXP_ELEMENT_REAL(el)->value, GMP_RNDN);
    l = strlen(str) - 1;
    if (str[0] == '-') {
	beg = str + 1;
	l--;
	neg = 1;
    }
    else {
	beg = str;
	neg = 0;
    }
    end = str + l;
    while(end >= beg && *end == '0')
	end--;
    if (end < beg) {
	g_free(str);
	return g_string_new("0.0");
    }
    end[1] = 0;
    nl = end - beg + 1;

    if (e >= 0 && e <= nl) {
	/* the separator is between the first and the last character */

	if (e == 0) {
	    str1 = g_new(char, neg + nl + 3);
	    s = str;
	    s1 = str1;
	    if (neg)
		*s1++ = *s++;
	    *s1++ = '0';
	}
	else {
	    virg = beg + e;
	    str1 = g_new(char, neg + nl + 3);
	    s1 = str1;
	    s = str;
	    for (s = str; s != virg; s++, s1++)
		*s1 = *s;
	}
	*s1++ = '.';
	if (s > end)
	    *s1++ = '0';
	else {
	    for (; s <= end; s++, s1++)
		*s1 = *s;
	}
	*s1 = 0;
    }
    else if (e < 0) {
	/* the separator is before the first character */
	if (-e <= l - nl) {
	    /* if writing all the zeros is shorter than the original length */
	    str1 = g_new(char, neg - e + nl + 3);
	    s1 = str1;
	    if (neg)
		*s1++ = '-';
	    *s1++ = '0';
	    *s1++ = '.';
	    while (e) {
		*s1 = '0';
		s1++;
		e++;
	    }
	    for (s = beg; s <= end; s++, s1++)
		*s1 = *s;
	    *s1 = 0;
	}
	else {
	    /* if writing all the zeros is longer */
	    char *exp;

	    exp = g_strdup_printf("e%+ld", e - 1);
	    str1 = g_new(char, neg + nl + strlen(exp) + 2);
	    s1 = str1;
	    s = str;
	    if (neg) {
		*s1 = *s;
		s1++;
		s++;
	    }
	    *s1 = *s;
	    s1++;
	    s++;
	    *s1 = '.';
	    s1++;
	    for ( ; s <= end; s++, s1++)
		*s1 = *s;
	    for (s = exp; *s != 0; s++, s1++)
		*s1 = *s;
	    *s1 = 0;
	    g_free(exp);
	}
    }
    else {
	/* The separator should be after the last character */
	if (e <= l) {
	    /* There is a place sufficient to store all the necessary zeros */
	    str1 = g_new(char, neg + e + 3);
	    s1 = str1;
	    s = str;
	    for (s = str; s <= end; s++, s1++)
		*s1 = *s;
	    for (e = e - nl; e > 0; e--, s1++)
		*s1 = '0';
	    *s1++ = '.';
	    *s1++ = '0';
	    *s1 = 0;
	}
	else {
	    char *exp;

	    exp = g_strdup_printf("e%+ld", e - 1);
	    str1 = g_new(char, neg + nl + strlen(exp) + 3);
	    s1 = str1;
	    s = str;
	    if (neg) {
		*s1 = *s;
		s1++;
		s++;
	    }
	    *s1++ = *s++;
	    *s1++ = '.';
	    if (s > end)
		*s1++ = '0';
	    else {
		for ( ; s <= end; s++, s1++)
		    *s1 = *s;
	    }
	    for (s = exp; *s != 0; s++, s1++)
		*s1 = *s;
	    *s1 = 0;
	    g_free(exp);
	}
    }
    g_free(str);
    gstr = g_string_new(str1);
    g_free(str1);
    return gstr;
}

static NxpElement* nxp_real_and_or_real(NxpOp op, NxpElement *a, NxpElement *b,
					NxpEvalContext *context, GError **error)
{
    if (op == NXP_OP_SLASH_BSLASH) {
	if (mpfr_cmp_ui(NXP_ELEMENT_REAL(a)->value, 0))
	    return nxp_element_int_new(mpfr_cmp_ui(NXP_ELEMENT_REAL(b)->value, 0));
	else
	    return nxp_element_int_new(0);
    }
    else {
	if (mpfr_cmp_ui(NXP_ELEMENT_REAL(a)->value, 0))
	    return nxp_element_int_new(1);
	else
	    return nxp_element_int_new(mpfr_cmp_ui(NXP_ELEMENT_REAL(b)->value, 0));
    }
}

static NxpElement* inc_dec_error(NxpOp op, NxpElement *a, NxpElement *b, 
				 NxpEvalContext *context, GError **error)
{
    /* TODO: define a non lvalue GError and use it here */
    g_set_error(error, nxp_error_domain, NXP_ERROR_NO_LVALUE, _("Incrementation-Decrementation don't work with real numbers."));
    return NULL;
}

static NxpElement* eval_gt(NxpOp op, NxpElement *a, NxpElement *b,
			   NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpfr_cmp(
			       NXP_ELEMENT_REAL(a)->value,
			       NXP_ELEMENT_REAL(b)->value) > 0);
}

static NxpElement* eval_gte(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpfr_cmp(
			       NXP_ELEMENT_REAL(a)->value,
			       NXP_ELEMENT_REAL(b)->value) >= 0);
}

static NxpElement* eval_lt(NxpOp op, NxpElement *a, NxpElement *b, 
			   NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpfr_cmp(
			       NXP_ELEMENT_REAL(a)->value,
			       NXP_ELEMENT_REAL(b)->value) < 0);
}

static NxpElement* eval_lte(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpfr_cmp(
			       NXP_ELEMENT_REAL(a)->value,
			       NXP_ELEMENT_REAL(b)->value) <= 0);
}

static NxpElement* eval_eq(NxpOp op, NxpElement *a, NxpElement *b, 
			   NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpfr_cmp(
			       NXP_ELEMENT_REAL(a)->value,
			       NXP_ELEMENT_REAL(b)->value) == 0);
}

static NxpElement* eval_neq(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpfr_cmp(
			       NXP_ELEMENT_REAL(a)->value,
			       NXP_ELEMENT_REAL(b)->value) != 0);
}

static NxpElement* eval_not(NxpOp op, NxpElement *a, NxpElement *b, 
			    NxpEvalContext *context, GError **error)
{
    return nxp_element_int_new((signed long int)mpfr_cmp_si(
			       NXP_ELEMENT_REAL(b)->value, 0));
}

static inline NxpElement* nxp_element_real_opposed(NxpElementNumber *self)
{
    NxpElement *el;
    mpfr_t o;

    mpfr_init(o);
    mpfr_neg(o, NXP_ELEMENT_REAL(self)->value, GMP_RNDN);
    el = nxp_element_real_new_from_mpfr(o);
    return el;
}

static NxpElement* eval_uminus(NxpOp op, NxpElement *a, NxpElement *b,
			       NxpEvalContext *context, GError **error)
{
    return nxp_element_real_opposed(NXP_ELEMENT_NUMBER(b));
}

static gboolean truth_func(NxpElement *elem, NxpEvalContext *context, GError **error)
{
    return mpfr_cmp_si(NXP_ELEMENT_REAL(elem)->value, 0);
}

static int nxp_element_real_compare_int(NxpElementNumber *self, nxp_int_t nb)
{
    return mpfr_cmp_si(NXP_ELEMENT_REAL(self)->value, nb);
}

nxp_real_t nxp_element_real_get_double(NxpElement *self, GError **error)
{
    return mpfr_get_d(NXP_ELEMENT_REAL(self)->value, GMP_RNDN);
}

static double nxp_element_real_get_value_as_double(NxpElementR *self)
{
    return mpfr_get_d(NXP_ELEMENT_REAL(self)->value, GMP_RNDN);
}

/*static nxp_int_t nxp_element_real_get_value_as_int(NxpElementR *self)
{
    return (nxp_int_t)(NXP_ELEMENT_REAL(self)->value);
}
*/

static void destroy(NxpElement *el) {
    mpfr_clear(NXP_ELEMENT_REAL(el)->value);
}

NxpElementType nxp_element_real_get_type(void)
{
    static NxpElementType nxp_element_type_real = 0;
    if (nxp_element_type_real == 0) {
	nxp_element_type_real = nxp_element_next_type();
	typeinfo = nxp_element_type_register_info(
					    nxp_element_type_real,
					    sizeof(NxpElementRTypeInfo),
					    NXP_ELEMENT_TYPE_R,
					    sizeof(NxpElementRTypeInfo));
    }
    return nxp_element_type_real;
}

#define BIND(op,func)      nxp_bind_operator_function (op, 		\
     		  	   NXP_ELEMENT_TYPE_REAL, NXP_ELEMENT_TYPE_REAL, 	\
     				func, func, 0, NULL); 

void nxp_element_real_type_init(void)
{
    NxpElementNumberTypeInfo *nb_typeinfo;
    NxpElementRTypeInfo *r_typeinfo;

    typeinfo = nxp_element_type_get_info(nxp_element_real_get_type());
    nb_typeinfo = (NxpElementNumberTypeInfo *)typeinfo;
    r_typeinfo = (NxpElementRTypeInfo *)typeinfo;
    typeinfo->name = g_strdup("real");
    typeinfo->eval_func = nxp_eval_return_self;
    typeinfo->simplify_func = nxp_eval_return_self;
    typeinfo->str_func = str_func;
    typeinfo->repr_func = str_func;
    typeinfo->truth_func = truth_func;
    typeinfo->destroy_func = destroy;

    nb_typeinfo->compare_int = nxp_element_real_compare_int;
    nb_typeinfo->opposed = nxp_element_real_opposed;
    r_typeinfo->get_value_as_double = nxp_element_real_get_value_as_double;
//    r_typeinfo->get_value_as_int = nxp_element_real_get_value_as_int;

    BIND(NXP_OP_PLUS, add);
    BIND(NXP_OP_MINUS, sub);
    
    nxp_bind_operator_function(NXP_OP_MINUS_LEFT, 0, NXP_ELEMENT_TYPE_REAL,
			       eval_uminus, eval_uminus,
			       0, NULL);
    BIND(NXP_OP_STAR, times);
    BIND(NXP_OP_SLASH, divide);
    BIND(NXP_OP_HAT, eval_real_hat);
    BIND(NXP_OP_SLASH_BSLASH, nxp_real_and_or_real);
    BIND(NXP_OP_BSLASH_SLASH, nxp_real_and_or_real);

    // TODO -- IMPLEMENT AND BIND NXP_OP_DOT3

    nxp_bind_operator_function(NXP_OP_PLUS2_LEFT, 0, NXP_ELEMENT_TYPE_REAL,
			       inc_dec_error, inc_dec_error, 
			       0, NULL);
    nxp_bind_operator_function(NXP_OP_PLUS2_RIGHT, NXP_ELEMENT_TYPE_REAL, 0,
			       inc_dec_error, inc_dec_error,
			        0, NULL);
    nxp_bind_operator_function(NXP_OP_MINUS2_LEFT, 0, NXP_ELEMENT_TYPE_REAL,
			       inc_dec_error, inc_dec_error, 0, NULL);
    nxp_bind_operator_function(NXP_OP_MINUS2_RIGHT, NXP_ELEMENT_TYPE_REAL, 0,
			       inc_dec_error, inc_dec_error, 0, NULL);

    BIND(NXP_OP_GT, eval_gt);
    BIND(NXP_OP_GT_EQ, eval_gte);
    BIND(NXP_OP_LT, eval_lt);
    BIND(NXP_OP_LT_EQ, eval_lte);
    BIND(NXP_OP_EQ2, eval_eq);
    BIND(NXP_OP_EXCL_EQ, eval_neq);

    nxp_bind_operator_function(NXP_OP_TILDE_LEFT, 0, NXP_ELEMENT_TYPE_REAL,
			       eval_not, eval_not, 0, NULL);
    nxp_parser_register_func(NXP_ATOM_CLASS_REAL,
			     nxp_element_real_new_from_string, 50);
}
