/* -*- Mode: C; c-file-style: "bsd"; -*- */

/* Variable number relation */

#include <math.h>

#include "mpoly.h"
#include "tree.h"
#include "variable.h"
#include "variables.h"
#include "number.h"
#include "frac.h"
#include "errors.h"
#include "op_bindings.h"
#include "types_misc.h"
#include "i18n.h"
#include "container.h"

static NxpElement* divide_by_var(NxpOp op, NxpElement *a, NxpElement *b, 
				 NxpEvalContext *context, GError **error) {
    if (nxp_element_is_true(a, context, error)) {
	NxpElementVector *vector;
	NxpElement *mon, *args, *retval;
	vector = NXP_ELEMENT_VECTOR(nxp_element_vector_new(1, NULL));
	_element_vector_index(vector, 0) = nxp_element_int_new(-1);
	mon = nxp_monomial_new(a, NXP_ELEMENT(vector));
	NXP_ELEMENT_UNREF(vector);
	args = nxp_element_list_new_simple(b);
	retval = nxp_element_mpoly_new_simple(mon, args);
	NXP_ELEMENT_UNREF(mon);
	NXP_ELEMENT_UNREF(args);
	return retval;
    } else {
	return NXP_ELEMENT_REF(a);
    }
}

static NxpElement* divide_by_number(NxpOp op, NxpElement *a, NxpElement *b, 
				    NxpEvalContext *context, GError **error) {
    NxpElement *one;
    NxpElement *retval, *args, *mon, *coeff, *vector;

    g_return_val_if_fail(NXP_ELEMENT_IS_NUMBER(b), NULL);

    if (nxp_element_is_true(b, context, error)) {
	args = nxp_element_list_new_simple(a);
	vector = nxp_element_vector_new(1, NULL);
	one = nxp_element_int_new(1);
	_element_vector_set(NXP_ELEMENT_VECTOR(vector), 0, one);
	coeff = nxp_element_tree_simplify_op(op, one, b,
					    context, error);
	mon = nxp_monomial_new(coeff, vector);
	retval = nxp_element_mpoly_new_simple(mon, args);
	NXP_ELEMENT_UNREF(one);
	NXP_ELEMENT_UNREF(args);
	NXP_ELEMENT_UNREF(vector);
	NXP_ELEMENT_UNREF(coeff);
	NXP_ELEMENT_UNREF(mon);
	return retval;
    } else {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DIVIDE_BY_ZERO, _("divide by zero"));
	return NULL;
    }
}

static NxpElement* simplify_hat(NxpOp op, NxpElement *a, NxpElement *b, 
				 NxpEvalContext *context, GError **error) {
    NxpElement *pol;
    NxpElement *args, *mon, *coeff, *vector;

    g_return_val_if_fail(NXP_ELEMENT_IS_NUMBER(b), NULL);

    if (nxp_element_is_true(b, context, error)) {
	if (nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(b), 1) == 0) {
	    return NXP_ELEMENT_REF(a);
	} else if (nxp_element_number_is_integer(NXP_ELEMENT_NUMBER(b)) /*&&
		nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(b), 1) > 0*/) {
	    args = nxp_element_list_new_simple(a);
	    vector = nxp_element_vector_new(1, NULL);
	    _element_vector_set(NXP_ELEMENT_VECTOR(vector), 0, b);
	    coeff = nxp_element_int_new(1);
	    mon = nxp_monomial_new(coeff, vector);
	    pol = nxp_element_mpoly_new_simple(mon, args);
	    NXP_ELEMENT_UNREF(args);
	    NXP_ELEMENT_UNREF(vector);
	    NXP_ELEMENT_UNREF(coeff);
	    NXP_ELEMENT_UNREF(mon);
	    return pol;
	}
	else
	    return nxp_element_tree_new(NXP_OP_HAT, a, b);
    } else
	return nxp_element_int_new(1);
}

static NxpElement* sub_v_n(NxpOp op, NxpElement *a, NxpElement *b, 
			   NxpEvalContext *context, GError **error) {
    if (nxp_element_is_true(b, context, error)) {
	NxpElement *retval, *args, *vector, *one, *mon, *coeff;

	args = nxp_element_list_new_simple(a);
	vector = nxp_element_vector_new(1, NULL);
	one = nxp_element_int_new(1);
	_element_vector_index(NXP_ELEMENT_VECTOR(vector), 0) = nxp_element_int_new(0);
	coeff = nxp_element_tree_simplify_op(NXP_OP_MINUS_LEFT, NULL,
					     b, context, error);
	mon = nxp_monomial_new(coeff, vector);
	retval = nxp_element_mpoly_new_simple(mon, args);

	NXP_ELEMENT_UNREF(args);
	NXP_ELEMENT_UNREF(vector);
	NXP_ELEMENT_UNREF(coeff);
	NXP_ELEMENT_UNREF(mon);
	vector = nxp_element_vector_new(1, NULL);
	_element_vector_set(NXP_ELEMENT_VECTOR(vector), 0, one);
	coeff = one;
	mon = nxp_monomial_new(coeff, vector);
	NXP_ELEMENT_UNREF(vector);
	NXP_ELEMENT_UNREF(coeff);
	nxp_element_list_append(NXP_ELEMENT(NXP_ELEMENT_MPOLY(retval)->mon), mon);
	NXP_ELEMENT_UNREF(mon);
	return retval;
    } else
	return NXP_ELEMENT_REF(a);
}

static NxpElement* sub_n_v(NxpOp op, NxpElement *a, NxpElement *b, 
			   NxpEvalContext *context, GError **error) {
    NxpElement *retval;
    NxpElement *args, *vector, *mon, *coeff;

    args = nxp_element_list_new_simple(b);
    vector = nxp_element_vector_new(1, NULL);
    _element_vector_index(NXP_ELEMENT_VECTOR(vector), 0) = nxp_element_int_new(0);
    mon = nxp_monomial_new(a, vector);
    retval = nxp_element_mpoly_new_simple(mon, args);

    NXP_ELEMENT_UNREF(args);
    NXP_ELEMENT_UNREF(vector);
    NXP_ELEMENT_UNREF(mon);
    vector = nxp_element_vector_new(1, NULL);
    _element_vector_index(NXP_ELEMENT_VECTOR(vector), 0) = nxp_element_int_new(1);
    coeff = nxp_element_int_new(-1);
    mon = nxp_monomial_new(coeff, vector);
    NXP_ELEMENT_UNREF(vector);
    NXP_ELEMENT_UNREF(coeff);
    nxp_element_list_append(NXP_ELEMENT(NXP_ELEMENT_MPOLY(retval)->mon), mon);
    NXP_ELEMENT_UNREF(mon);
    return retval;
}

static NxpElement* add(NxpOp op, NxpElement *a, NxpElement *b, 
			NxpEvalContext *context, GError **error) {
    if (nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(a), 0)) {
	NxpElement *retval;
	NxpElement *args, *vector, *mon, *coeff;

	args = nxp_element_list_new_simple(b);
	vector = nxp_element_vector_new(1, NULL);
	_element_vector_index(NXP_ELEMENT_VECTOR(vector), 0) = nxp_element_int_new(0);
	mon = nxp_monomial_new(a, vector);
	retval = nxp_element_mpoly_new_simple(mon, args);

	NXP_ELEMENT_UNREF(args);
	NXP_ELEMENT_UNREF(vector);
	NXP_ELEMENT_UNREF(mon);
	vector = nxp_element_vector_new(1, NULL);
	coeff = nxp_element_int_new(1);
	_element_vector_index(NXP_ELEMENT_VECTOR(vector), 0) = NXP_ELEMENT_REF(coeff);
	mon = nxp_monomial_new(coeff, vector);
	NXP_ELEMENT_UNREF(vector);
	NXP_ELEMENT_UNREF(coeff);
	nxp_element_list_append(NXP_ELEMENT(NXP_ELEMENT_MPOLY(retval)->mon), mon);
	NXP_ELEMENT_UNREF(mon);
	return retval;
    } else {
	return NXP_ELEMENT_REF(b);
    }
}

static NxpElement* eq(NxpOp op, NxpElement *a, NxpElement *b,
		      NxpEvalContext *context, GError **error)
{
    NxpElement *content = nxp_variable_get(NXP_ELEMENT_VARIABLE(b)->ident, context);
    gboolean retval;

    if (content) {
	NxpElement *simp = nxp_element_tree_simplify_op(NXP_OP_MINUS, content, a, context, error);
	retval = nxp_element_is_true(simp, context, error);
	NXP_ELEMENT_UNREF(simp);
    }
    else
	return nxp_element_tree_new(op, a, b);

    return nxp_element_int_new(op == NXP_OP_EQ2 ? !retval : retval);
}

static NxpElement* mult(NxpOp op, NxpElement *a, NxpElement *b, 
			NxpEvalContext *context, GError **error) {
    g_return_val_if_fail(NXP_ELEMENT_NUMBER(a), NULL);

    if (nxp_element_is_true(a, context, error)) {
	if (nxp_element_number_compare_int(NXP_ELEMENT_NUMBER(a), 1) == 0)
	    return NXP_ELEMENT_REF(b);
	else {
	    NxpElement *retval;
	    NxpElement *args, *vector, *mon;

	    args = nxp_element_list_new_simple(b);
	    vector = nxp_element_vector_new(1, NULL);
	    _element_vector_index(NXP_ELEMENT_VECTOR(vector), 0) = nxp_element_int_new(1);
	    mon = nxp_monomial_new(a, vector);
	    retval = nxp_element_mpoly_new_simple(mon, args);

	    NXP_ELEMENT_UNREF(args);
	    NXP_ELEMENT_UNREF(vector);
	    NXP_ELEMENT_UNREF(mon);
	    return retval;
	}
    } else
	return nxp_element_int_new(0);
}


#define BIND_V(op,simp)    nxp_bind_operator_function (op, 		\
  				NXP_ELEMENT_TYPE_NUMBER, NXP_ELEMENT_TYPE_VARIABLE, \
  				NULL,  			\
  				simp, 			\
  				NXP_OP_BIND_SYMMETRIC, \
				NULL); 

#define BIND_V_N(op,simp)    nxp_bind_operator_function (op, 			\
  				NXP_ELEMENT_TYPE_VARIABLE, NXP_ELEMENT_TYPE_NUMBER, \
  				NULL,  			\
  				simp, 			\
  				0, \
				NULL); 

#define BIND_N_V(op,simp)    nxp_bind_operator_function (op, 			\
  				NXP_ELEMENT_TYPE_NUMBER, NXP_ELEMENT_TYPE_VARIABLE, \
  				NULL,  			\
  				simp, 			\
  				0, \
				NULL); 

#define BIND_T(op,simp)    nxp_bind_operator_function (op, 		\
  				NXP_ELEMENT_TYPE_NUMBER, NXP_ELEMENT_TYPE_TREE, \
  				NULL,  			\
  				simp, 			\
  				NXP_OP_BIND_SYMMETRIC, \
				NULL);

#define BIND_T_N(op,simp)    nxp_bind_operator_function (op, 			\
  				NXP_ELEMENT_TYPE_TREE, NXP_ELEMENT_TYPE_NUMBER, \
  				NULL,  			\
  				simp, 			\
  				0, \
				NULL);

#define BIND_N_T(op,simp)    nxp_bind_operator_function (op, 			\
  				NXP_ELEMENT_TYPE_NUMBER, NXP_ELEMENT_TYPE_TREE, \
  				NULL,  			\
  				simp, 			\
  				0, \
				NULL);

void nxp_element_variable_number_operator_init(void)
{
    BIND_V(NXP_OP_PLUS,add);
    BIND_V_N(NXP_OP_MINUS,sub_v_n);
    BIND_N_V(NXP_OP_MINUS,sub_n_v);

    BIND_V(NXP_OP_STAR,mult);
    //NXP_OP_MINUS
    BIND_V_N(NXP_OP_SLASH, divide_by_number);
    BIND_N_V(NXP_OP_SLASH, divide_by_var);
    BIND_V_N(NXP_OP_HAT, simplify_hat);
    BIND_V(NXP_OP_EQ2, eq);
    BIND_V(NXP_OP_EXCL_EQ, eq);

    //NXP_OP_MINUS_LEFT
    //NXP_OP_MINUS
    //NXP_OP_DOT3
    //NXP_OP_PLUS2_LEFT
    //NXP_OP_MINUS2_LEFT
    //NXP_OP_PLUS2_RIGHT
    //NXP_OP_MINUS2_RIGHT
    //NXP_OP_GT
    //NXP_OP_LT
    //NXP_OP_GE_EQ
    //NXP_OP_LT_EQ
    //NXP_OP_ExCL_EQ
    //NXP_OP_SLASH_BSLASH
    //NXP_OP_BSLASH_SLASH
    //NXP_TILDE_LEFT
    //NXP_EXCL_RIGHT
    //NXP_OP_EQ
    //NXP_OP_COLON_EQ
}

