/* -*- c -*- */
#include <gsl/gsl_interp.h>
#include "nxp_kernel.h"
#include "gsl.h"
#include "i18n.h"


typedef struct _NxpInterpolator NxpInterpolator;

struct _NxpInterpolator {
    double *x, *y;
    gint size;
    gsl_interp *interp;
    gsl_interp_accel *accel;
};

NxpElement *nxp_interpolator_eval(NxpElementReal *x, 
				  NxpInterpolator *interp, ETC) {
    return nxp_element_real_new(gsl_interp_eval(interp->interp, 
						interp->x, interp->y,
						nxp_element_r_get_value_as_double(x), interp->accel));
}

void nxp_interpolator_free(gpointer data) {
    NxpInterpolator *interp = (NxpInterpolator*)data;
    gsl_interp_accel_free(interp->accel);
    gsl_interp_free(interp->interp);
    g_free(interp->x);
    g_free(interp->y);
    g_free(interp);
}


static int
compare_double_pair(const void *a, const void *b)
{
    return *((double *) a) - *((double *) b);
}

static void
nxp_vector_of_pairs_to_two_double_arrays(NxpElementVector * vec,
					 double **X, double **Y, GError **error)
{
    NxpElement     *pair,
                   *el_x,
                   *el_y;
    int             size = vec->size;
    int             index;
    double         *x;
    double         *y;
    double         *xy;
    double	    tmp;

    *X = NULL;
    *Y = NULL;

    xy = g_new(double, 2 * size);
    if (!xy) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, _("Not enough memory."));
	return;
    }

    for (index = 0; index < size; ++index) {
	pair = NXP_ELEMENT_VECTOR_ELEM(vec, index);
	if (pair && NXP_ELEMENT_IS_VECTOR(pair)
	    && NXP_ELEMENT_VECTOR(pair)->size == 2) {
	    el_x = NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(pair), 0);
	    el_y = NXP_ELEMENT_VECTOR_ELEM(NXP_ELEMENT_VECTOR(pair), 1);

	    if (NXP_ELEMENT_IS_R(el_x) && NXP_ELEMENT_IS_R(el_y)) {
		xy[2 * index] =
		    nxp_element_r_get_value_as_double(NXP_ELEMENT_R(el_x));
		xy[2 * index + 1] =
		    nxp_element_r_get_value_as_double(NXP_ELEMENT_R(el_y));
	    } else {
		g_free(xy);
		g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK, _("Pairs are represented by a vector of two real numbers."));
		return;
	    }
	} else {
	    g_free(xy);
	    g_set_error(error, nxp_error_domain, NXP_ERROR_TYPE_CHECK, _("The data vector must contain pairs of real numbers. Those pairs are represented by vectors too."));
	    return;
	}
    }

    qsort(xy, size, sizeof(double) * 2, &compare_double_pair);

    x = g_new(double, size);
    y = g_new(double, size);

    if (!x || !y) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_ERROR, _("Not enough memory."));
	g_free(xy);
	if (y)
	    g_free(y);
	if (x)
	    g_free(x);
	return;
    }

    tmp = NAN;
    for (index = 0; index < size; ++index) {
	if (tmp != xy[index << 1]) {
	    tmp = x[index] = xy[index << 1];
	    y[index] = xy[(index << 1) + 1];
	}
	else {
	    g_free(xy);
	    g_free(x);
	    g_free(y);
	    g_set_error(error, nxp_error_domain, NXP_ERROR_INVALID_ARGUMENT, _("Some pairs in the vector have the same X-coordinate"));
	    return;
	}
    }

    g_free(xy);

    *X = x;
    *Y = y;

    return;
}


static NxpElement* nxp_gsl_interp_generic(const gsl_interp_type *type,
					  NxpElementVariable *id, 
					  NxpElementVector *data, ETC) {
    NxpInterpolator *interp;
    GError *err = NULL;

    interp = g_new(NxpInterpolator, 1);
    interp->size = data->size;
    g_print ("data->size = %d\n", data->size);
    interp->interp = gsl_interp_alloc(type, interp->size);
    nxp_vector_of_pairs_to_two_double_arrays(data, &interp->x, &interp->y, &err);
    if (err) {
	g_propagate_error(error, err);
	return NULL;
    }
    gsl_interp_init(interp->interp, interp->x, interp->y, interp->size);
    interp->accel = gsl_interp_accel_alloc();

    nxp_bifunc_define_detailed(id->ident, &nxp_prototype__REAL, nxp_interpolator_eval,
			       NULL, nxp_interpolator_eval, interp, nxp_interpolator_free, 
			       context, NXP_PROTO_INSTANCE_VOLATILE, &err);
    if (err) {
	nxp_interpolator_free(interp);
	g_propagate_error(error, err); 
	return NULL; 
    }
    nxp_output_message(context, _("Function defined"));
    return nxp_element_void_new();
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
static NxpElement* nxp_gsl_interp_linear(NxpElementVariable *id, 
					 NxpElementVector *data, ETC) {
    if (data->size < 2) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("The data vector must contain at least %d pairs."), 2);
	return NULL;
    }
    return nxp_gsl_interp_generic(gsl_interp_linear, id,
				  data, named_args, context, error);
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
static NxpElement* nxp_gsl_interp_polynomial(NxpElementVariable *id,
					 NxpElementVector *data, ETC) {
    if (data->size < 3) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("The data vector must contain at least %d pairs."), 3);
	return NULL;
    }
    return nxp_gsl_interp_generic(gsl_interp_polynomial, id, data,
				  named_args, context, error);
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
static NxpElement* nxp_gsl_interp_cspline(NxpElementVariable *id,
					 NxpElementVector *data, ETC) {
    if (data->size < 3) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("The data vector must contain at least %d pairs."), 3);
	return NULL;
    }
    return nxp_gsl_interp_generic(gsl_interp_cspline, id, data, named_args,
				  context, error);
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
static NxpElement* nxp_gsl_interp_cspline_periodic(NxpElementVariable *id,
					 NxpElementVector *data, ETC) {
    if (data->size < 3) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("The data vector must contain at least %d pairs."), 3);
	return NULL;
    }
    return nxp_gsl_interp_generic(gsl_interp_cspline_periodic, id, data, named_args, context, error);
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
static NxpElement* nxp_gsl_interp_akima(NxpElementVariable *id,
					 NxpElementVector *data, ETC) {
    if (data->size < 5) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("The data vector must contain at least %d pairs."), 5);
	return NULL;
    }
    return nxp_gsl_interp_generic(gsl_interp_akima, id, data, named_args, context, error);
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
static NxpElement* nxp_gsl_interp_akima_periodic(NxpElementVariable *id,
					 NxpElementVector *data, ETC) {
    if (data->size < 5) {
	g_set_error(error, nxp_error_domain, NXP_ERROR_DOMAIN, _("The data vector must contain at least %d pairs."), 5);
	return NULL;
    }
    return nxp_gsl_interp_generic(gsl_interp_akima_periodic, id, data, named_args, context, error);
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */




void
nxp_gsl_interp_init(NxpEvalContext * context)
{
    NxpIdentifier  *id;
nxp_name_space_new(context->cwd,"gsl");
nxp_name_space_new(nxp_name_space_find_child(context->cwd,"gsl"),"interpolation");
id = nxp_identifier_new_with_full_path(3,"gsl","interpolation", "linear");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_VECTOR, nxp_gsl_interp_linear, NULL, nxp_gsl_interp_linear, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","interpolation", "polynomial");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_VECTOR, nxp_gsl_interp_polynomial, NULL, nxp_gsl_interp_polynomial, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","interpolation", "cspline");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_VECTOR, nxp_gsl_interp_cspline, NULL, nxp_gsl_interp_cspline, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","interpolation", "csplinePeriodic");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_VECTOR, nxp_gsl_interp_cspline_periodic, NULL, nxp_gsl_interp_cspline_periodic, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","interpolation", "akima");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_VECTOR, nxp_gsl_interp_akima, NULL, nxp_gsl_interp_akima, context, NULL); nxp_identifier_unref(id);
id = nxp_identifier_new_with_full_path(3,"gsl","interpolation", "akimaPeriodic");
 nxp_bifunc_define(id, &nxp_prototype__VARIABLE_VECTOR, nxp_gsl_interp_akima_periodic, NULL, nxp_gsl_interp_akima_periodic, context, NULL); nxp_identifier_unref(id);
}
