#include "latex.h"
#include "nxp_kernel.h"
#include <string.h>

static GString* string2latex(NxpElement *elem, int *prec, GError **error)
{
    NxpElementString *string = NXP_ELEMENT_STRING(elem);
    GString *gstr;
    if (prec) *prec = 0;
    gstr = g_string_new("\\hbox{``");
    gstr = g_string_append(gstr, string->string);
    gstr = g_string_append(gstr, "''}");
    return gstr;
}


static GString* matrix2latex(NxpElement *elem, int *precedence, GError **error) {
    GString *str = g_string_new("{\\left[\\matrix{");

    int i, j;
    NxpElement *el;
    NxpElementMatrix* m;
    GString* old;
    
    if (precedence) *precedence = 0;

    m = NXP_ELEMENT_MATRIX(elem);
    for (i = 0; i < m->height; i++) {
      for (j = 0; j < m->width; j++) {
	el = _element_matrix_get(m, i, j);

	old = nxp_latex(el, NULL, NULL);

	NXP_ELEMENT_UNREF(el);
	str = g_string_append(str, old->str);
	str = g_string_append_c(str, '&');
	g_string_free(old, TRUE);
      }
      str = g_string_truncate(str, (str)->len - 1);
      str = g_string_append(str, "\\cr");
    }
    str = g_string_truncate(str, (str)->len - 3);
    
    str = g_string_append(str, "}\\right]}");

    return str;
}


static GString* vector2latex(NxpElement *elem, int *precedence, GError **error)
{
    GString *string = g_string_new("{<");
    int i;
    NxpElement *el;
    NxpElementVector* v;
    GString* old;

    if (precedence) *precedence = 0;

    v = NXP_ELEMENT_VECTOR(elem);
    for (i = 0; i < v->size; i++) {
	el = _element_vector_get(v, i);
	old = nxp_latex(el, NULL, NULL); 
	NXP_ELEMENT_UNREF(el);
	string = g_string_append(string, old->str);
	string = g_string_append(string, ", ");
	g_string_free(old, TRUE);
    }
    if (v->size) string = g_string_truncate(string, (string)->len - 2);
    string = g_string_append(string, ">}");

    return string;
}

static GString *func2latex(NxpElement *elem, int *prec, GError **error)
{
    GString *str;
    GSList *arg;
    NxpElementFunction *func = NXP_ELEMENT_FUNCTION(elem);

    // treat sqrt as a special case
    if (strcmp(nxp_identifier_to_string(NXP_ELEMENT_VARIABLE(func)->ident),"sqrt") == 0) {
	GString *str1;

	str = g_string_new("\\sqrt");
	str1 = nxp_latex(NXP_ELEMENT(NXP_ELEMENT_LIST(func->simple_args)->list->data), NULL, NULL);
	str = g_string_append(str, str1->str);
	g_string_free(str1, TRUE);
	
    } else {
	str = nxp_identifier_to_gstring(NXP_ELEMENT_VARIABLE(func)->ident);

	str = g_string_prepend(str, "{{\\rm ");
	str = g_string_append(str, "}\\left( ");
	for (arg = NXP_ELEMENT_LIST(func->simple_args)->list; arg; arg = arg->next)
	{
	    GString *str1;
	    NxpElement *el;
	    el = NXP_ELEMENT(arg->data);
	    str1 = nxp_latex(el, NULL, NULL);
	    str = g_string_append(str, str1->str);
	    g_string_free(str1, TRUE);
	    if (arg->next)
		str = g_string_append(str, ", ");
	}
	str = g_string_append(str, "\\right)}");
    }
    if (prec) *prec = 0;
    return str;
}


static GString* direct2latex(NxpElement* el, int *precedence, GError **error) {
    GString* repr;
    if (precedence) *precedence = 0;
    repr = nxp_element_repr(el);
    repr = g_string_prepend_c(repr,'{');
    repr = g_string_append(repr,"}\n");
    return repr;
}

static GString* tree2latex(NxpElement* el, int *precedence, GError **error) {
    NxpElementTree *tree = NXP_ELEMENT_TREE(el);
    const NxpOpInfo *opinfo;
    GString *result, *tmp = NULL;
    gint left = 0, right = 0, myprecedence;

    // Get operator information and precedence
    opinfo = nxp_op_get_info(tree->op);
    myprecedence = opinfo->precedence;
    
    // Initialize result string
    result = g_string_new("{");

    if (tree->left) {
	tmp = nxp_latex(tree->left, &left, NULL);
	if (myprecedence < left) {
	    tmp = g_string_prepend(tmp, "{\\left(");
	    tmp = g_string_append(tmp, "\\right)}");
	}
	result = g_string_append(result, tmp->str);
	g_string_free(tmp, TRUE);
    }
    switch(tree->op) {
	// NXP_OP_PLUS: OK
	// NXP_OP_MINUS: OK
	// NXP_OP_MINUS_LEFT: OK
    case NXP_OP_STAR:  // OK
	result = g_string_append(result, "\\times ");
	break;
    case NXP_OP_SLASH:  // OK
	result = g_string_append(result, "\\over ");
	break;
    case NXP_OP_DOT3:	// OK
	result = g_string_append(result, "\\dots ");
	break;
	// NXP_OP_HAT: OK
	// NXP_OP_DOT3: OK
	// NXP_OP_PLUS2_LEFT: --------- CHECK WHAT TO DO
	// NXP_OP_MINUS2_LEFT: -------- CHECK WHAT TO DO
	// NXP_OP_PLUS2_RIGHT: -------- CHECK WHAT TO DO
	// NXP_OP_MINUS2_RIGHT: ------- CHECK WHAT TO DO
	// NXP_OP_GT: OK
    case NXP_OP_GT_EQ: // OK
	result = g_string_append(result, "\\leq ");
	break;
	// NXP_OP_LT: // OK
    case NXP_OP_LT_EQ: // OK
	result = g_string_append(result, "\\geq ");
	break;
    case NXP_OP_EQ2: // OK
	result = g_string_append(result, "=");
	break;
    case NXP_OP_EXCL_EQ: // OK
	result = g_string_append(result, "\\not=");
	break;
    case NXP_OP_SLASH_BSLASH: // OK
	result = g_string_append(result, "\\wedge ");
	break;
    case NXP_OP_BSLASH_SLASH: // OK
	result = g_string_append(result, "\\vee ");
	break;
    case NXP_OP_TILDE_LEFT: // OK
	result = g_string_append(result, "\\sim ");
	break;
	// NXP_OP_EXCL_RIGHT: ------- CHECK WHAT TO DO
    case NXP_OP_EQ: // OK
	result = g_string_append(result, "\\leftarrow ");
	break;
	// NXP_OP_COLON_EQ: --------- CHECK WHAT TO DO
	// NXP_OP_PERCENT: ---------- CHECK WHAT TO DO
    default:
	result = g_string_append(result, opinfo->token);
    }
    if (tree->right) {
	tmp = nxp_latex(tree->right, &right, NULL);
	if (myprecedence < right) {
	    tmp = g_string_prepend(tmp, "{\\left(");
	    tmp = g_string_append(tmp, "\\right)}");
	}
	result = g_string_append(result, tmp->str);
	g_string_free(tmp, TRUE);
    } 
    result = g_string_append_c(result,'}');

    if (precedence) *precedence = myprecedence;
    return result;
}

static NxpElement* latex(NxpElement* element, ETC)
{
    gint precedence;
    GString *latex = nxp_latex(element, &precedence, NULL);
    NxpElement *rr = nxp_element_string_new(latex->str);
    g_string_free(latex, TRUE);
    return rr;
}

GString* nxp_latex(NxpElement* element, gint *precedence, GError **error) {
    NxpElementTypeInfo* info;
    NxpLaTeXrepr func;

    info = nxp_element_type_get_info(NXP_ELEMENT_GET_TYPE(element));
    func = (NxpLaTeXrepr)nxp_element_type_get_data(info, "latex");

    if (func) return (*func)(element, precedence, NULL);
    return g_string_new("??");
}

static void nxp_latex_init(NxpEvalContext *context) {
    nxp_bifunc_define_global("LaTeX", &nxp_prototype__ANY,
			     latex, NULL, NULL, context, NULL);

#define REGISTER(type,func)  nxp_element_type_set_data(nxp_element_type_get_info(type),\
                                                      "latex", func);
    REGISTER(NXP_ELEMENT_TYPE_INT,      direct2latex);
    REGISTER(NXP_ELEMENT_TYPE_REAL,     direct2latex);
    REGISTER(NXP_ELEMENT_TYPE_TREE,       tree2latex);
    REGISTER(NXP_ELEMENT_TYPE_VARIABLE, direct2latex);
    REGISTER(NXP_ELEMENT_TYPE_FUNCTION,   func2latex);
    REGISTER(NXP_ELEMENT_TYPE_COMPLEX,  direct2latex);
    REGISTER(NXP_ELEMENT_TYPE_VECTOR,   vector2latex);
    REGISTER(NXP_ELEMENT_TYPE_MATRIX,   matrix2latex);
    REGISTER(NXP_ELEMENT_TYPE_STRING,   string2latex);
#undef REGISTER
}


void nxp_module_init(NxpEvalContext *context, GError **error)
{
    nxp_latex_init(context);
}

